package heasarc.util;

/** This comprises a set  of static methods derived from
 *  the Fortran SLA library written by Patrick Wallace.
 *  Only a few of the extensive SLA library routines
 *  are included, those required for simple transformations
 *  among celestial coordinate systems.
 * 
 *  This code is a relatively literal translation of the
 *  original Fortran.  Significant semantic changes
 *  include:
 *  - all methods are functions rather than
 *    subroutines.  The calculated values are
 *    returned as scalar or vector double values.
 *    This accommodates Java's pass by value arguments.
 * 
 *  Note that in a few cases the original SLA library
 *  used in/out arguments.  Otherwise
 * <ul>
 *   <li> Comments have been changed to use javadoc syntax (typically
 *     the parameter descriptions have been moved to the end
 *     of the comments so as not to break up the
 *     textual description of the method.
 *   <li> Constant values use an 'e' in the exponent
 *   <li> Do loops are replaced by for loops.
 *   <li> The indices used in two-dimensional arrays are reversed.
 *   <li> Some constants are moved out of routines and into
 *        the class as a whole.
 *   <li> Declarations and definitions are changed to
 *        match Java syntax.  Note that we typically declare
 *        a variable at the point of first use rather than
 *        at the top of the method.
 *   <li> Case has been changed to match Java conventions.
 *   <li> Semicolons have been added after statements.
 *   <li> References to the Math library need to be
 *        made explicit in Java. (i.e., sin -> Math.sin)
 *   <li> The modulo function is an standard operator in Java.
 *   <li> Statements like x = x+y -> x += y; 
 *   <li> Strings are handled using Java methods rather than
 *        as character arrays.
 * </ul>
 *  
 *  Errors introduced in the transcription are the responsibility
 *  of T. McGlynn.  NASA/GSFC.
 * 
 */

public class Sla {
    
    /** The transformation matrix from equatorial
     *  to galactic coordinates.
     *  Note that in the original Fortran code
     *  the data is declared element by element
     *  with rows varying first.
     */
    public static final double[][] eqgal_rmat =
      {
      { -0.054875539726e0,
            -0.873437108010e0,
            -0.483834985808e0
      },
      { +0.494109453312e0,
	    -0.444829589425e0,
	    +0.746982251810e0
      },
      { -0.867666135858e0,
	    -0.198076386122e0,
	    +0.455983795705e0
      }
    };

    public static final double D2PI= 2*Math.PI;
    public static final double PMF=100e0*60e0*60e0*360e0/D2PI;
    public static final double TINY = 1.e-30;
    public static final double AS2R=0.484813681109535994e-5;
      
	
/** 
*  Transformation from J2000.0 equatorial coordinates to
*  IAU 1958 galactic coordinates
*
* <p>
*  <strong>Note:</strong>
*     The equatorial coordinates are J2000.0.  Use the routine
*     sla_EG50 if conversion from B1950.0 'FK4' coordinates is
*     required.
* <p>
*  Reference:
*     Blaauw et al, Mon.Not.R.Astron.Soc.,121,123 (1960)
*  <p>
*  copyright 1998 Rutherford Appleton Laboratory.
*  
*  @author P.T.Wallace   Starlink   21 September 1998
*
*  @param dr      J2000.0 RA.
*  @param dd      J2000.0 Dec.
*
*  @return double[2] array of galactic longitude and latitude.
*
*/
    public static double[] eqgal(double dr, double dd) {


      double[] v1 = new double[3], v2=new double[3];

/*
*  L2,B2 system of galactic coordinates
*
*  P = 192.25       RA of galactic north pole (mean B1950.0)
*  Q =  62.6        inclination of galactic to mean B1950.0 equator
*  R =  33          longitude of ascending node
*
*  P,Q,R are degrees
*
*  Equatorial to galactic rotation matrix (J2000.0), obtained by
*  applying the standard FK4 to FK5 transformation, for zero proper
*  motion in FK5, to the columns of the B1950 equatorial to
*  galactic rotation matrix:
*/
      v1 = dcs2c(dr,dd);

//  Equatorial to galactic *
      v2 = dmxv(eqgal_rmat,v1);

      double[] res = dcc2s(v2);

//  Express in conventional ranges
      res[0] = dranrm(res[0]);
      res[1] = drange(res[1]);

      return res;
    }
    
    
/**
*  Transformation from IAU 1958 galactic coordinates to
*  J2000.0 equatorial coordinates
*  <p>
*     The equatorial coordinates are J2000.0.  Use the routine
*     sla_GE50 if conversion to B1950.0 'FK4' coordinates is
*     required.
*  <p>
*  Reference:
*     Blaauw et al, Mon.Not.R.Astron.Soc.,121,123 (1960)
*
*  <p>
*  Copyright (C) 1998 Rutherford Appleton Laboratory.
*  @authorP.T.Wallace   Starlink   21 September 1998
*
*  @param   dl      galactic longitude L2.
*  @param   db      galactic longitude B2.
* 
*
*  @return     double[2] J2000.0 RA,Dec.
*/
    public static double[] galeq (double dl, double db) {

      double[] v1= new double[3], v2= new double[3];

/*
*  L2,B2 system of galactic coordinates
*
*  P = 192.25       RA of galactic north pole (mean B1950.0)
*  Q =  62.6        inclination of galactic to mean B1950.0 equator
*  R =  33          longitude of ascending node
*
*  P,Q,R are degrees
*
*  Equatorial to galactic rotation matrix (J2000.0), obtained by
*  applying the standard FK4 to FK5 transformation, for zero proper
*  motion in FK5, to the columns of the B1950 equatorial to
*  galactic rotation matrix:
*/

//  Spherical to Cartesian
      v1 = dcs2c(dl,db);

//  Galactic to equatorial
      v2 = dimxv(eqgal_rmat,v1);

//  Cartesian to spherical
      double[] res = dcc2s(v2);

//  Express in conventional ranges
      res[0] = dranrm(res[0]);
      res[1] = drange(res[1]);
      return res;
    }

/*
*  Transformation from J2000.0 equatorial coordinates to
*  ecliptic coordinates
*
*<p>
* Copyright 1995 Rutherford Appleton Laboratory
*
* @author P.T.Wallace   Starlink   March 1986
* @param   DR    J2000.0 mean RA
* @param   DD    J2000.0 mean Dec
* @param   DATE  TDB (loosely ET) as Modified Julian Date (JD-2400000.5)
* @return        double[2]  ecliptic longitude and latitude (mean of date, 
*                IAU 1980 theory, radians)
*
*/
    public static double[] eqecl (double dr, double dd, double date) {

      double[] v1 =dcs2c(dr,dd);

//  Mean J2000 to mean of date
      double[][] rmat = prec(2000e0, epj(date));
      double[] v2 = dmxv(rmat,v1);

//  Equatorial to ecliptic
      rmat = ecmat(date);
      double[] v3 = dmxv(rmat,v2);

//  Cartesian to spherical
      double[] res = dcc2s(v3);

//  Express in conventional ranges
      res[0]=dranrm(res[0]);
      res[1]=drange(res[1]);
      
      return res;
    }
    
    

/**
*  Transformation from ecliptic coordinates to
*  J2000.0 equatorial coordinates
*
*  <p>
*  Copyright 1995 Rutherford Appleton Laboratory.
*
*  @author P.T.Wallace   Starlink   March 1986
*  @param  dl      ecliptic longitude
*  @param  db      ecliptic latitude
*  @param  data    TDB (loosely ET) as Modified Julian Date
*                  (JD-2400000.5)
*  @return         double[2] J2000.0 mean RA,Dec
*
*/
   public static double[] ecleq (double dl, double db, double date) {


//  Spherical to Cartesian
      double[] v1 = dcs2c(dl,db);

//  Ecliptic to equatorial
      double[][] rmat = ecmat(date);
      double[] v2     = dimxv(rmat,v1);

//  Mean of date to J2000
      rmat = prec(2000e0, epj(date));
      double[] v3 = dimxv(rmat,v2);

//  Cartesian to spherical
      double[] res = dcc2s(v3);

//  Express in conventional ranges
      res[0]=dranrm(res[0]);
      res[1]=drange(res[1]);
       
      return res;
   }

/**
*  Convert B1950.0 FK4 star data to J2000.0 FK5 assuming zero
*  proper motion in the FK5 frame
*
*  This routine converts stars from the old, Bessel-Newcomb, FK4
*  system to the new, IAU 1976, FK5, Fricke system, in such a
*  way that the FK5 proper motion is zero.  Because such a star
*  has, in general, a non-zero proper motion in the FK4 system,
*  the routine requires the epoch at which the position in the
*  FK4 system was determined.
*
*  The method is from Appendix 2 of Ref 1, but using the constants
*  of Ref 4.
*
*  Notes:
*  <ol>
*
*  <li>  The epoch BEPOCH is strictly speaking Besselian, but
*      if a Julian epoch is supplied the result will be
*      affected only to a negligible extent.
*
*  <li>  Conversion from Besselian epoch 1950.0 to Julian epoch
*      2000.0 only is provided for.  Conversions involving other
*      epochs will require use of the appropriate precession,
*      proper motion, and E-terms routines before and/or
*      after FK45Z is called.
*
*  <li>  In the FK4 catalogue the proper motions of stars within
*      10 degrees of the poles do not embody the differential
*      E-term effect and should, strictly speaking, be handled
*      in a different manner from stars outside these regions.
*      However, given the general lack of homogeneity of the star
*      data available for routine astrometry, the difficulties of
*      handling positions that may have been determined from
*      astrometric fields spanning the polar and non-polar regions,
*      the likelihood that the differential E-terms effect was not
*      taken into account when allowing for proper motion in past
*      astrometry, and the undesirability of a discontinuity in
*      the algorithm, the decision has been made in this routine to
*      include the effect of differential E-terms on the proper
*      motions for all stars, whether polar or not.  At epoch 2000,
*      and measuring on the sky rather than in terms of dRA, the
*      errors resulting from this simplification are less than
*      1 milliarcsecond in position and 1 milliarcsecond per
*      century in proper motion.
*  </ol>
* 
*  References:
*  <ol>
*
*     <li>  Aoki,S., et al, 1983.  Astron.Astrophys., 128, 263.
*
*     <li>  Smith, C.A. et al, 1989.  "The transformation of astrometric
*        catalog systems to the equinox J2000.0".  Astron.J. 97, 265.
*
*     <li>  Yallop, B.D. et al, 1989.  "Transformation of mean star places
*        from FK4 B1950.0 to FK5 J2000.0 using matrices in 6-space".
*        Astron.J. 97, 274.
*
*     <li>  Seidelmann, P.K. (ed), 1992.  "Explanatory Supplement to
*        the Astronomical Almanac", ISBN 0-935702-68-7.
*
*  </ol>
*  Copyright (C) 1998 Rutherford Appleton Laboratory
*
*  @author P.T.Wallace   Starlink   21 September 1998
*     @param r1950	B1950.0 FK4 RA at epoch (rad)
*     @param d1950	B1950.0 FK4 Dec at epoch (rad)
*     @param bepoch     Besselian epoch (e.g. 1979.3e0)
*
*     @return		double[2] J2000.0 FK5 RA,Dec (rad)
*
*
*/


public static double[] fk45z (double r1950, double d1950, double bepoch) {


      double[] a  = new double[]{-1.62557e-6,  -0.31919e-6, -0.13843e-6};
      double[] ad = new double[]{+1.245e-3,    -1.580e-3,   -0.659e-3};
      double[][] em = new double[][] 
      {    
    
         {  
	     +0.9999256782e0,
             -0.0111820611e0,
             -0.0000271765e0
	 },{
	    
             +0.0111820610e0,
             +0.9999374784e0,
             -0.0048579477e0
	 },{

	     +0.0048579479e0,
             -0.0000271474e0,
             +0.9999881997e0
	 },{

	     -0.000551e0,
             -0.238565e0,
             +0.435739e0
	 },{
	    
             +0.238514e0,
             -0.002667e0,
             -0.008541e0
	 },{
	    
             -0.435623e0
             +0.012254e0
             +0.002117e0
	 }
      };

//  Spherical to Cartesian
      double[] r0= dcs2c(r1950,d1950);
      

//  Adjust vector A to give zero proper motion in FK5
      double w = (bepoch-1950e0)/PMF;
    
      double[] a1 = new double[3];
      for (int i=0; i<3; i += 1) {
	  a1[i]=a[i]+w*ad[i];
      }
    
//  Remove e-terms
      w = r0[0]*a1[0] + r0[1]*a1[1] + r0[2]*a1[2];
    
      double[] v1 = new double[3];
      for (int i=0; i<3; i += 1) {
         v1[i] = r0[i] - a1[i]+w*r0[i];
      }

//  Convert position vector to Fricke system
      double[] v2 = new double[6];
      for (int i=0; i<6; i += 1) {
         w = 0e0;
         for (int j=0; j<3; j += 1) {
            w = w + em[i][j]*v1[j];
         }
         v2[i] = w;
      }

//  Allow for fictitious proper motion in FK4
      w = (epj(epb2d(bepoch))-2000e0)/PMF;
      for (int i=0; i<3; i += 1) {
         v2[i] = v2[i] + w*v2[i+3];
      }

//  Revert to spherical coordinates
      double[] res = dcc2s(v2);
      res[0] = dranrm(res[0]);
      return res;
    }



/**
*  Convert a J2000.0 FK5 star position to B1950.0 FK4 assuming
*  zero proper motion and parallax
*  <p>
*  This routine converts star positions from the new, IAU 1976,
*  FK5, Fricke system to the old, Bessel-Newcomb, FK4 system.
*
*  <p>
*  Notes:
*  <ol>
*  <li>  The proper motion in RA is dRA/dt rather than cos(Dec)*dRA/dt.
*
*  <li>  Conversion from Julian epoch 2000.0 to Besselian epoch 1950.0
*      only is provided for.  Conversions involving other epochs will
*      require use of the appropriate precession routines before and
*      after this routine is called.
*
*  <li>  Unlike in the sla_FK524 routine, the FK5 proper motions, the
*      parallax and the radial velocity are presumed zero.
*
*  <li>  It is the intention that FK5 should be a close approximation
*      to an inertial frame, so that distant objects have zero proper
*      motion;  such objects have (in general) non-zero proper motion
*      in FK4, and this routine returns those fictitious proper
*      motions.
*
*  <li>  The position returned by this routine is in the B1950
*      reference frame but at Besselian epoch BEPOCH.  For
*      comparison with catalogues the BEPOCH argument will
*      frequently be 1950e0.
*
*  </ol>
* 
*  Copyright (C) 1995 Rutherford Appleton Laboratory
*  @author P.T.Wallace   Starlink   10 April 1990
*  @param r2000  J2000.0 FK5 RA  (rad).
*  @param d2000  J2000.0 FK5 Dec (rad).
*  @param bepoch Besselian epoch (e.g. 1950e0)
*
*  @return	 double[4] FK4 RA,Dec (rad) at epoch BEPOCH,
*                           FK4 proper motions (rad/trop.yr).
*/
   public static double[] fk54z (double r2000, double d2000, double bepoch) {

//  FK5 equinox J2000 (any epoch) to FK4 equinox B1950 epoch B1950
      double[] pj   = fk524(r2000,d2000,0e0,0e0,0e0,0e0);
      double r      = pj[0];
      double d      = pj[1];
      double dr1950 = pj[2];
      double dd1950 = pj[3];
      double px     = pj[4];
      double py     = pj[5];

//  Fictitious proper motion to epoch BEPOCH
      double[] res = pm(r,d,dr1950,dd1950,0e0,0e0,1950e0,bepoch);
       
      return new double[] {res[0], res[1], dr1950, dd1950};
       
   }


/**
*  Convert J2000.0 FK5 star data to B1950.0 FK4
*  <p>
*  This routine converts stars from the new, IAU 1976, FK5, Fricke
*  system, to the old, Bessel-Newcomb, FK4 system.  The precepts
*  of Smith et al (Ref 1) are followed, using the implementation
*  by Yallop et al (Ref 2) of a matrix method due to Standish.
*  Kinoshita's development of Andoyer's post-Newcomb precession is
*  used.  The numerical constants from Seidelmann et al (Ref 3) are
*  used canonically.
*
*
*  Notes:
*  <ol>
*  <li>  The proper motions in RA are dRA/dt rather than
*      cos(Dec)*dRA/dt, and are per year rather than per century.
*
*  <li>  Note that conversion from Julian epoch 2000.0 to Besselian
*      epoch 1950.0 only is provided for.  Conversions involving
*      other epochs will require use of the appropriate precession,
*      proper motion, and E-terms routines before and/or after
*      FK524 is called.
*
*  <li>  In the FK4 catalogue the proper motions of stars within
*      10 degrees of the poles do not embody the differential
*      E-term effect and should, strictly speaking, be handled
*      in a different manner from stars outside these regions.
*      However, given the general lack of homogeneity of the star
*      data available for routine astrometry, the difficulties of
*      handling positions that may have been determined from
*      astrometric fields spanning the polar and non-polar regions,
*      the likelihood that the differential E-terms effect was not
*      taken into account when allowing for proper motion in past
*      astrometry, and the undesirability of a discontinuity in
*      the algorithm, the decision has been made in this routine to
*      include the effect of differential E-terms on the proper
*      motions for all stars, whether polar or not.  At epoch 2000,
*      and measuring on the sky rather than in terms of dRA, the
*      errors resulting from this simplification are less than
*      1 milliarcsecond in position and 1 milliarcsecond per
*      century in proper motion.
*  </ol>
*  References:
*  <ol>
*     <li>  Smith, C.A. et al, 1989.  "The transformation of astrometric
*        catalog systems to the equinox J2000.0".  Astron.J. 97, 265.
*
*     <li>  Yallop, B.D. et al, 1989.  "Transformation of mean star places
*        from FK4 B1950.0 to FK5 J2000.0 using matrices in 6-space".
*        Astron.J. 97, 274.
*
*     <li>  Seidelmann, P.K. (ed), 1992.  "Explanatory Supplement to
*        the Astronomical Almanac", ISBN 0-935702-68-7.
*  </ol>
*
*  Copyright  1995 Rutherford Appleton Laboratory
*  
*  @author P.T.Wallace   Starlink   19 December 1993
*     @param r2000    	J2000.0 RA  (rad)
*     @param d2000    	J2000.0 Dec (rad)
*     @param dr2000     J2000.0 proper motion (rad/Jul.yr)
*     @param dd2000     J2000.0 proper motion (rad/Jul.yr)
*     @param p2000      parallax (arcsec)
*     @param v2000      radial velocity (km/s, +ve = moving away)
*
*     @return	        double[6] B1950.0 RA,Dec (rad),
*                                 B1950.0 proper motions (rad/trop.yr),
*                                 parallax (arcsec),
*                                 radial velocity (km/s, +ve = moving away)
*/
      public static double[] fk524 (double r2000, double d2000,
				    double dr2000,double dd2000,
				    double p2000, double v2000) {


/*
*  CANONICAL CONSTANTS  (see references)
*/

//  Km per sec to AU per tropical century
//  = 86400 * 36524.2198782 / 149597870
      double  vf=21.095e0;

      double[] a  = new double[]{-1.62557e-6,  -0.31919e-6, -0.13843e-6, 
                                 +1.245e-3,    -1.580e-3,   -0.659e-3};

      double[][] emi = new double[][]
	  {
	      {
                             +0.9999256795e0,
                             +0.0111814828e0,
                             +0.0048590039e0,
                             -0.00000242389840e0,
                             -0.00000002710544e0,
                             -0.00000001177742e0
	    }, {
                             -0.0111814828e0,
                             +0.9999374849e0,
                             -0.0000271771e0,
                             +0.00000002710544e0,
                             -0.00000242392702e0,
                             +0.00000000006585e0
	    }, {
                             -0.0048590040e0,
                             -0.0000271557e0,
                             +0.9999881946e0,
                             +0.00000001177742e0,
                             +0.00000000006585e0,
                             -0.00000242404995e0
	    }, {
                             -0.000551e0,
                              +0.238509e0,
                             -0.435614e0,
                             +0.99990432e0,
                             +0.01118145e0,
                             +0.00485852e0
	    }, {
		    
                             -0.238560e0,
                             -0.002667e0,
                             +0.012254e0,
                             -0.01118145e0,
                             +0.99991613e0,
                             -0.00002717e0
	    }, {
                             +0.435730e0,
                             -0.008541e0,
                             +0.002117e0 ,
                             -0.00485852e0,
                             -0.00002716e0,
                             +0.99996684e0
	    }
      };


     

//  Pick up J2000 data (units radians and arcsec/JC)
      double r  = r2000;
      double d  = d2000;
      double ur = dr2000*PMF;
      double ud = dd2000*PMF;
      double px = p2000;
      double rv = v2000;

//  Spherical to Cartesian
      double sr = Math.sin(r);
      double cr = Math.cos(r);
      double sd = Math.sin(d);
      double cd = Math.cos(d);

      double x = cr*cd;
      double y = sr*cd;
      double z =    sd;

      double w = vf*rv*px;

      double[] v1= new double[] {
	           x,y,z,
                   -ur*y - cr*sd*ud + w*x,
                    ur*x - sr*sd*ud + w*y,
                    cd*ud + w*z
      };

//  Convert position+velocity vector to BN system
      double[] v2 = new double[6];
      for (int i=0; i<6; i += 1) {
         w= 0e0;
         for (int j=0; j<6; j += 1) {
             w += emi[i][j]*v1[j];
         }
         v2[i]=w;
      }

//  Position vector components and magnitude
      x = v2[0];
      y = v2[1];
      z = v2[2];
      double rxyz = Math.sqrt(x*x + y*y + z*z);

//  Apply E-terms to position
      
	  
      w = x*a[0] + y*a[1] + z*a[2];
      x = x + a[0]*rxyz - w*x;
      y = y + a[1]*rxyz - w*y;
      z = z + a[2]*rxyz - w*z;

//  Recompute magnitude
	    
      rxyz = Math.sqrt(x*x + y*y + z*z);
	  

//  Apply E-terms to both position and velocity

      double wd, xd, yd, zd;
	  
      x  = v2[0];
      y  = v2[1];
      z  = v2[2];
      w  = x*a[0] + y*a[1] + z*a[2];
      wd = x*a[3] + y*a[4] + z*a[5];
      x  = x + a[0]*rxyz - w*x;
      y  = y + a[1]*rxyz - w*y;
      z  = z + a[2]*rxyz - w*z;
      xd = v2[3] + a[3]*rxyz - wd*x;
      yd = v2[4] + a[4]*rxyz - wd*y;
      zd = v2[5] + a[5]*rxyz - wd*z;
	  
//  Convert to spherical
	    
      double rxysq = x*x + y*y;
      double rxy   = Math.sqrt(rxysq);

      if (x  == 0e0 && y == 0e0) {
         r = 0e0;
      } else {
         r = Math.atan2(y,x);
         if (r  < 0.0e0) r += D2PI;
      }
      d = Math.atan2(z,rxy);

      if (rxy > TINY) {
	  
         ur = (x*yd - y*xd) / rxysq;
         ud = (zd*rxysq - z*(x*xd+y*yd)) / ((rxysq+z*z)*rxy);
	      
      }

//  Radial velocity and parallax

      if (px > TINY) {
         rv = (x*xd+y*yd+z*zd) / (px*vf*rxyz);
         px = px/rxyz;
      }

//  Return results
      double[] res = new double[] {r, d, ur/PMF, ud/PMF, px, rv};
      return res;
    }


/**
*  Precession - either FK4 (Bessel-Newcomb, pre IAU 1976) or
*  FK5 (Fricke, post IAU 1976) as required.
*
*  <p>
*  Notes:
*  <ol>
*     <li>  Lowercase characters in SYSTEM are acceptable.
*
*     <li>  The epochs are Besselian if SYSTEM='FK4' and Julian if 'FK5'.
*         For example, to precess coordinates in the old system from
*         equinox 1900.0 to 1950.0 the call would be:
*             CALL sla_PRECES ('FK4', 1900e0, 1950e0, RA, DC)
*
*     <li>  This routine will NOT correctly convert between the old and
*         the new systems - for example conversion from B1950 to J2000.
*         For these purposes see sla_FK425, sla_FK524, sla_FK45Z and
*         sla_FK54Z.
*
*     <li>  If an invalid SYSTEM is supplied, values of -99e0,-99e0 will
*         be returned for both RA and DC.
* </ol>
*
*  Copyright (C) 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   20 April 1990
*     @param system     precession to be applied: "FK4" or "FK5"
*     @param ep0        starting epoch
*     @param ep1        ending epoch
*     @param ra         RA mean equator & equinox of epoch EP0
*     @param dc         Dec mean equator & equinox of epoch EP0
*
*     @return           double[2] RA,Dec, mean equator & equinox of epoch EP1
*/

      public static double[] preces (String system, double ep0, double ep1, double ra, double dc) {
     
         double[][] pm;
    
         system = system.toUpperCase();
         if (!system.equals("FK4") && ! system.equals("FK5")) {
	     return null;
         }

//       Generate appropriate precession matrix
         if (system.equals("FK4")) {
            pm = prebn(ep0,ep1);
         } else {
            pm = prec(ep0,ep1);
         }
      
//       Convert RA,Dec to x,y,z
         double[] v1 = dcs2c(ra,dc);

//       Precess
         double[] v2 = dmxv(pm,v1);
      
//       Back to RA,Dec
         double[] res = dcc2s(v2);
         res[0] = dranrm(res[0]);
	  
	 return res;
      }

/**
*  Generate the matrix of precession between two epochs,
*  using the old, pre-IAU1976, Bessel-Newcomb model, using
*  Kinoshita's formulation
*
*  <p>
*  The matrix is in the sense   V(BEP1)  =  RMATP * V(BEP0)
*
*  Reference:
*  <p>
*     Kinoshita, H. (1975) 'Formulas for precession', SAO Special
*     Report No. 364, Smithsonian Institution Astrophysical
*     Observatory, Cambridge, Massachusetts.
* <p>
*  Copyright 1996 Rutherford Appleton Laboratory
*
*  @author P.T.Wallace   Starlink   23 August 1996
*
*  @param  bep0    beginning Besselian epoch
*  @param  bep1    ending Besselian epoch
*  @return         double[][]  precession matrix
*/
    public static double[][] prebn (double bep0, double bep1) {



//  Interval between basic epoch B1850.0 and beginning epoch in TC
      double bigt  = (bep0-1850e0)/100e0;

//  Interval over which precession required, in tropical centuries
      double t = (bep1-bep0)/100e0;

//  Euler angles
      double tas2r = t*AS2R;
      double w = 2303.5548e0+(1.39720e0+0.000059e0*bigt)*bigt;

      double zeta = (w+(0.30242e0-0.000269e0*bigt+0.017996e0*t)*t)*tas2r;
      double z = (w+(1.09478e0+0.000387e0*bigt+0.018324e0*t)*t)*tas2r;
      double theta = (2005.1125e0+(-0.85294e0-0.000365e0*bigt)*bigt+
                     (-0.42647e0-0.000365e0*bigt-0.041802e0*t)*t)*tas2r;

//  Rotation matrix
      return deuler("zyz", -zeta, theta, -z);
    }

/**
*  Form the matrix of precession between two epochs (IAU 1976, FK5)
*
*  Notes:
*  <ol>
*
*     <li>  The epochs are TDB (loosely ET) Julian epochs.
*
*     <li>  The matrix is in the sense   V(EP1)  =  RMATP * V(EP0)
*
*     <li>  Though the matrix method itself is rigorous, the precession
*         angles are expressed through canonical polynomials which are
*         valid only for a limited time span.  There are also known
*         errors in the IAU precession rate.  The absolute accuracy
*         of the present formulation is better than 0.1 arcsec from
*         1960AD to 2040AD, better than 1 arcsec from 1640AD to 2360AD,
*         and remains below 3 arcsec for the whole of the period
*         500BC to 3000AD.  The errors exceed 10 arcsec outside the
*         range 1200BC to 3900AD, exceed 100 arcsec outside 4200BC to
*         5600AD and exceed 1000 arcsec outside 6800BC to 8200AD.
*         The SLALIB routine sla_PRECL implements a more elaborate
*         model which is suitable for problems spanning several
*         thousand years.
* </ol>
*  References:
*    <p>
*     Lieske,J.H., 1979. Astron.Astrophys.,73,282.
*      equations (6) & (7), p283.
*     <p>
*     Kaplan,G.H., 1981. USNO circular no. 163, pA2.
*
*  <p>
* 
*  Copyright 1996 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   23 August 1996
*  @param   ep0		beginning epoch
*  @param   ep1         ending epoch
*  @return		precession matrix
*/
    public static double[][] prec (double ep0, double ep1) {

//  Interval between basic epoch J2000.0 and beginning epoch (JC)
      double t0 = (ep0-2000e0)/100e0;

//  Interval over which precession required (JC)
      double t = (ep1-ep0)/100e0;

//  Euler angles
      double tas2r = t*AS2R;
      double w = 2306.2181e0+(1.39656e0-0.000139e0*t0)*t0;

      double zeta  = (w+((0.30188e0-0.000344e0*t0)+0.017998e0*t)*t)*tas2r;
      double z     = (w+((1.09468e0+0.000066e0*t0)+0.018203e0*t)*t)*tas2r;
      double theta = ((2004.3109e0+(-0.85330e0-0.000217e0*t0)*t0)
                     +((-0.42665e0-0.000217e0*t0)-0.041833e0*t)*t)*tas2r;

//  Rotation matrix
      return deuler("zyz", -zeta, theta, -z);
    }

/**
*  Form the equatorial to ecliptic rotation matrix - IAU 1980 theory
*  <p>
*  Reference:
*  <br>
*     Murray,C.A., Vectorial Astrometry, section 4.3.
* <p>
*  Note:
* <br>
*    The matrix is in the sense   V(ecl)  =  RMAT * V(equ);  the
*    equator, equinox and ecliptic are mean of date.
*
* <p> 
*  Copyright 1996 Rutherford Appleton Laboratory
*  @author P.T.Wallace   Starlink   23 August 1996
*
*     @param date	TDB (loosely ET) as Modified Julian Date
*                                            (JD-2400000.5)
*     @return		double[][] matrix
*
*/
   public static double[][] ecmat (double date) {

//  Interval between basic epoch J2000.0 and current epoch (JC)
      double t = (date-51544.5e0)/36525e0;

//  Mean obliquity
      double eps0 = AS2R*
         (84381.448e0+(-46.8150e0+(-0.00059e0+0.001813e0*t)*t)*t);

//  Matrix
      return deuler("X", eps0, 0e0, 0e0);
   }

/**
*  Conversion of Modified Julian Date to Julian Epoch
*
* <p>
*  Reference:
*  <br>
*     Lieske,J.H., 1979. Astron.Astrophys.,73,282.
*
*
*  Copyright 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   February 1984
*  @param   date	Modified Julian Date (JD - 2400000.5)
*  @return		Julian epoch
*/
    public static double epj (double date){

      return  2000e0 + (date-51544.5e0)/365.25e0;
    }

/**
*  Conversion of Besselian Epoch to Modified Julian Date
*  <p>
*  Reference:
* <br> 
*     Lieske,J.H., 1979. Astron.Astrophys.,73,282.
*
* <p>
*  Copyright 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   February 1984
*  @param   epb		Besselian Epoch
*  @return  		Modified Julian Date.
*
*/
    public static double epb2d (double epb) {

      return 15019.81352e0 + (epb-1900e0)*365.242198781e0;
    }

/**
*
*  Spherical coordinates to direction cosines
*  <p>
*  The spherical coordinates are longitude (+ve anticlockwise
*  looking from the +ve latitude pole) and latitude.  The
*  Cartesian coordinates are right handed, with the x axis
*  at zero longitude and latitude, and the z axis at the
*  +ve latitude pole.
*  <p>
*
*  Copyright 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   October 1984
*  @param	a	spherical coordinates in radians
*  @param	b	spherical coordinates in radians
*  @return		x,y,z unit vector
*
*/
    public static double[] dcs2c (double a, double b) {

      double[] v = new double[3];

      double cosb = Math.cos(b);

      v[0] = Math.cos(a)*cosb;
      v[1] = Math.sin(a)*cosb;
      v[2] = Math.sin(b);
	
      return v;
    }


/**
*  Direction cosines to spherical coordinates
*  <p>
*  The spherical coordinates are longitude (+ve anticlockwise
*  looking from the +ve latitude pole) and latitude.  The
*  Cartesian coordinates are right handed, with the x axis
*  at zero longitude and latitude, and the z axis at the
*  +ve latitude pole.
*  <p>
*  If V is null, zero A and B are returned.
*  At either pole, zero A is returned.
*  <p>
*
*  Copyright (C) 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   July 1989
*  @param     v     x,y,z vector
*
*  @return    	    spherical coordinates in radians
*/
    public static double[] dcc2s (double[] v){

      double x = v[0];
      double y = v[1];
      double z = v[2];
      double r = Math.sqrt(x*x+y*y);

      double[] res = new double[2];
	
      if (r == 0e0) {
         res[0] = 0e0;
      } else {
         res[0] = Math.atan2(y,x);
      }

      if (z == 0e0) {
         res[1] = 0e0;
      } else {
         res[1] = Math.atan2(z,r);
      }
	
      return res;
    }
				   
/**
*  Performs the 3-D forward unitary transformation:
*  <p>
*     vector VB = matrix DM * vector VA
*  <p>
*
*  Copyright (C) 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   March 1986
*     @param dm       matrix
*     @param va       vector
*
*     @return	      result vector
*/
    public static double[] dmxv (double[][] dm, double[] va) {

      double[] vw = new double[3];


//  Matrix DM * vector VA -> vector VW
//
      for (int j=0; j<3; j += 1) {
	  
         double w=0e0;
	  
	 for (int i=0; i<3; i += 1) {
            w += dm[j][i]*va[i];
         }
         vw[j] = w;
      }
      return vw;
    }

/**
*  Performs the 3-D backward unitary transformation:
*  <p>
*     vector VB = (inverse of matrix DM) * vector VA
*  <p>
*
*  (n.b.  the matrix must be unitary, as this routine assumes that
*   the inverse and transpose are identical)
*  <p>
*
*  Copyright 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   March 1986
*     @param dm       matrix
*     @param va       vector
*
*     @return	      result vector
*
*/
    public static double[] dimxv (double [][] dm, double[] va) {

      double[] vw = new double[3];

//  Inverse of matrix DM * vector VA -> vector VW
      for (int j=0; j<3; j += 1) {
         double w=0e0;
         for (int i=0; i<3; i += 1) {
            w += dm[i][j]*va[i];
         }
         vw[j] = w;
      }
      return vw;
    }

/**
*  Apply corrections for proper motion to a star RA,Dec
*  <p>
*  References:
*  <br>
*     1984 Astronomical Almanac, pp B39-B41.
*  <br>
*     (also Lederle & Schwan, Astron. Astrophys. 134,
*      1-6, 1984)
*  <p>
*  Note:
*  <br>
*     The proper motions in RA are dRA/dt rather than
*     cos(Dec)*dRA/dt, and are in the same coordinate
*     system as R0,e0.
*
*  <p>
* 
*  Copyright 1996 Rutherford Appleton Laboratory
*  @author P.T.Wallace   Starlink   23 August 1996
*
*  @param	r0	ra at epoch ep0 (rad)
*  @param	d0	dec at epoch ep0 (rad)
*  @param       pr	proper motions:  RA changes per year of epoch
*  @param       pd	proper motions:  Dec changes per year of epoch
*  @param       px      parallax (arcsec)
*  @param       rv      radial velocity (km/sec, +ve if receding)
*  @param       ep0     start epoch in years (e.g. Julian epoch)
*  @param       ep1     end epoch in years (same system as EP0)
*
*  @return              RA,Dec at epoch EP1 (rad)
*
*/
    public static double[] pm (double r0, double d0,  
			       double pr, double pd, 
			       double px, double rv, 
			       double ep0, double ep1) {

      double vfr=0.21094502e0*0.484813681109535994e-5;

//  Spherical to Cartesian
      double[] p = dcs2c(r0,d0);

//  Space motion (radians per year)
      double[] em =  new double[3];
      double    w =  vfr*rv*px;
      em[0]       = -pr*p[1]-pd*Math.cos(r0)*Math.sin(d0)+w*p[0];
      em[1]       =  pr*p[0]-pd*Math.sin(r0)*Math.sin(d0)+w*p[1];
      em[2]       =          pd*Math.cos(d0)             +w*p[2];

//  Apply the motion
      double t=ep1-ep0;
      for (int i=0; i<3; i += 1) {
         p[i]=p[i]+t*em[i];
      }

//  Cartesian to spherical

      double[] res = dcc2s(p);
      res[0] = dranrm(res[0]);
      return res;
    }

/**
*  Form a rotation matrix from the Euler angles - three successive
*  rotations about specified Cartesian axes
*  <p>
*  A rotation is positive when the reference frame rotates
*  anticlockwise as seen looking towards the origin from the
*  positive region of the specified axis.
*  <p>
*  The characters of ORDER define which axes the three successive
*  rotations are about.  A typical value is 'ZXZ', indicating that
*  RMAT is to become the direction cosine matrix corresponding to
*  rotations of the reference frame through PHI radians about the
*  old Z-axis, followed by THETA radians about the resulting X-axis,
*  then PSI radians about the resulting Z-axis.
*  <p>
*  The axis names can be any of the following, in any order or
*  combination:  X, Y, Z, uppercase or lowercase, 1, 2, 3.  Normal
*  axis labelling/numbering conventions apply;  the xyz (=123)
*  triad is right-handed.  Thus, the 'ZXZ' example given above
*  could be written 'zxz' or '313' (or even 'ZxZ' or '3xZ').  ORDER
*  is terminated by length or by the first unrecognized character.
*  <p>
*  Fewer than three rotations are acceptable, in which case the later
*  angle arguments are ignored.  If all rotations are zero, the
*  identity matrix is produced.
*  <p>
*  Copyright 1997 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   23 May 1997
* 
*  @param 		order   specifies about which axes the rotations occur
*  @param  		phi     1st rotation (radians)
*  @param               theta   2nd rotation (   "   )
*  @param               psi     3rd rotation (   "   )
*
*  @return              	double[][] rotation matrix
*/
    public static double[][] deuler (String order, double phi, double theta, double psi) {


      double[][] result = new double[3][3];
      double[][] rotn   = new double[3][3];
      char axis;



//    Initialize result matrix
      for (int j=0; j<3; j += 1) {
	  for (int i=0; i<3; i += 1) {
	      if (i != j) {
		  result[i][j] = 0;
	      } else {
		  result[i][j] = 1;
	      }
	  }
      }

//    Establish length of axis string
      int l = order.length();

//    Look at each character of axis string until finished
      for (int n=0; n < 3 && n < l; n += 1) {

//          Initialize rotation matrix for the current rotation
            for (int j=0; j<3; j += 1) {
	        for (int i=0; i<3; i += 1) {
	            if (i != j) {
		      rotn[i][j] = 0;
	            } else {
		        rotn[i][j] = 1;
	            }
		    
	        }
            }

//          Pick up the appropriate Euler angle and take sine & cosine
            double angle;
            if (n == 0) {
               angle = phi;
            } else if (n == 1) {
               angle = theta;
            } else {
               angle = psi;
	    }
            double s = Math.sin(angle);
            double c = Math.cos(angle);

//          Identify the axis
            axis = order.charAt(n);
            if (axis == 'x' ||
               axis == 'X'  ||
               axis == '1') {

//             Matrix for x-rotation
               rotn[1][1] =  c;
               rotn[1][2] =  s;
               rotn[2][1] = -s;
               rotn[2][2] =  c;
	    } else if (axis == 'y' ||
                       axis == 'Y' ||
                       axis == '2') {

//             Matrix for y-rotation
               rotn[0][0] =  c;
               rotn[0][2] = -s;
               rotn[2][0] =  s;
               rotn[2][2] =  c;

            } else if (axis == 'z' ||
                       axis == 'Z' ||
                       axis == '3') {

//             Matrix for z-rotation
               rotn[0][0] =  c;
               rotn[0][1] =  s;
               rotn[1][0] = -s;
               rotn[1][1] =  c;

            }

//          Apply the current rotation (matrix ROTN x matrix RESULT)
            double[][] wm = new double[3][3];
		
            for (int i=0; i<3; i += 1) {
	       for (int j=0; j<3; j += 1) {
		   
                  double w = 0e0;
                  for (int k=0; k<3; k += 1) {
                     w += rotn[i][k]*result[k][j];
                  }
                  wm[i][j] = w;
	       }
	    }
	    
	    for (int j=0; j<3; j += 1) {
	       for (int i=0; i<3; i += 1) {
                  result[i][j] = wm[i][j];
	       }
	    }
      }
      return result;
    }
/**
*  Normalize angle into range 0-2 pi
*  <p>
*  Copyright 1995 Rutherford Appleton Laboratory
*  @author P.T.Wallace   Starlink   23 November 1995
*
*  @param	angle     angle in radians
*  @return   angle expressed in the range 0-2 pi
*/
    public static double dranrm (double angle) {

      angle = angle % D2PI;
      if (angle < 0) {
	  angle += D2PI;
      }
      return angle;
    }
    
/**
*  Normalize angle into range +/- pi
*
*  <p>
*  Copyright 1995 Rutherford Appleton Laboratory
*
*  @author  P.T.Wallace   Starlink   23 November 1995
*  @param     angle 	the angle in radians
*  @return              angle expressed in the range +/- pi.
*
*/
    public static double drange (double angle) {


      angle = angle % D2PI;
      if (angle > Math.PI) {
	  angle -= Math.PI;
      } else if (angle < -Math.PI) {
	  angle += Math.PI;
      }
      return angle;
    }
    
/**
*  Product of two 3x3 matrices:
*  <p>
*
*      matrix C  =  matrix A  x  matrix B
*
*  <p>
*  Copyright 1995 Rutherford Appleton Laboratory
* 
*  @author P.T.Wallace   Starlink   5 April 1990
*      @param 	A      matrix
*      @param   B      matrix
*
*      @return	       matrix result
*/
    public static double[][] dmxm (double[][] a, double[][] b) {


      double[][] wm = new double[3][3];

//  Multiply into scratch matrix

      for (int i=0; i<3; i += 1) {
	 double w;
	 for (int j=0; j<3; j += 1) {
            w=0e0;
	     
            for (int k=0; k<3; k += 1) {
               w = w + a[i][k]*b[k][j];
            }
            wm[i][j] = w;
	 }
      }
      return wm;

    }

    /** Convert from J2000 to any other coordinate system.
     * @author T. McGlynn
     * 
     * @param coords	Input coordinates (rad)
     * @param inSys	Input system.  Only the
     *                  first letter of the string is significant.
     *                  J=FK5, B=FK4, G=Galactic, E=Ecliptic
     * @param epoch	The epoch of the output coordinates.
     *                  Ignored for Galactic and Ecliptic transformations.
     * @return		The J2000 coordinates. (rad)
     */
    
    public static double[] toJ2000(double[] coords, String inSys, double epoch) {
	
	inSys = inSys.toUpperCase();
	
	if (inSys.charAt(0) == 'J') {
	    
	    if (epoch !=  2000) {
		coords =  preces("FK5", epoch, 2000.e0, coords[0], coords[1]);
	    }
	} else if (inSys.charAt(0) == 'B') {
	    
	    if (epoch != 1950) {
		coords = preces("FK4", epoch, 1950.e0, coords[0], coords[1]);
	    }
	    
	    coords =  fk45z(coords[0], coords[1], epoch);
	    
	} else if (inSys.charAt(0) == 'G') {
	    coords =  galeq(coords[0], coords[1]);
	} else if (inSys.charAt(0) == 'E') { 
	    coords =  ecleq(coords[0], coords[1], epb2d(2000.e0));
	} else {
	    coords = null;
	}
	return coords;
    }
 
    /** Convert from J2000 to any other coordinate system.
     * 
     * @author T. McGlynn
     * @param coords	J2000 coordinates (rad)
     * @param outSys	Output system.  Only the
     *                  first letter of the string is significant.
     *                  J=FK5, B=FK4, G=Galactic, E=Ecliptic
     * @param epoch	The epoch of the output coordinates.
     *                  Ignored for Galactic and Ecliptic transformations.
     * @return		The transformed coordinates. (rad)
     */
    
    public static double[] fromJ2000(double[] coords, String outSys, double epoch) {
	
	outSys = outSys.toUpperCase();
	if (outSys.charAt(0) == 'J') {
	    if (epoch != 2000) {
		coords =  preces("FK5", 2000.e0, epoch, coords[0], coords[1]);
	    }
	} else if (outSys.charAt(0) == 'B') {
	    coords = fk54z(coords[0], coords[1], 1950.e0);
	    if (epoch != 1950) {
		coords = preces("FK4", 1950.e0, epoch, coords[0], coords[1]);
	    }
	} else if (outSys.charAt(0) == 'G') {
	    coords = eqgal(coords[0], coords[1]);
	} else if (outSys.charAt(0) == 'E') {
	    coords = eqecl(coords[0], coords[1], epb2d(2000.e0));
	} else {
	    coords = null;
	}
	return coords;
    }
    
    /** Test the SLA routines.  The routine reads the input coordinate values in
     *  decimal degrees and writes the transformed values.
     *
     * @author T. McGlynn
     *  @param args  Usage: java skyview.geometry.sla.Sla ra0 dec0 inSys outSys epoch0 epoch
     */
    public static void main(String[] args) throws Exception {
	
	double[] coords = new double[2];
	coords[0] = new Double(args[0]).doubleValue()*Math.PI/180;
	coords[1] = new Double(args[1]).doubleValue()*Math.PI/180;
	String inSys  = args[2];
	String outSys = args[3];
	double ep0 = new Double(args[4]).doubleValue();
	double ep1 = new Double(args[5]).doubleValue();
	
	coords = toJ2000(coords, inSys, ep0);
	coords = fromJ2000(coords, outSys, ep1);
	
	System.out.println("RA/L: "+(coords[0]*180/Math.PI));
	System.out.println("Dec/B:"+(coords[1]*180/Math.PI));
    }
}
