package heasarc.w3browse.plot;

/** This class stores W3Browse data by column.
 */
public class Column {
    
    /** String containing name of field */
    protected  String fieldname;

    /** The data type of the column */
    protected  String type;

    /** Number of rows of data */
    protected  int nrows = 0;

    /** Indicates if column is numeric */
    public     boolean isNumeric; 

    /** The column data. If the column is numeric, data is a double[]
      * else column is String[] */
    public     Object data;

    /** Minimum data[] element */
    protected  double min = 0.0;

    /** Maximum data[] element */
    protected  double max = 0.0;

  /** Constructor for Column 
   * @param name name of field
   * @param type of data in column
   * @param data object array containing field data 
   * @param isNumeric indicates if column contains numeric data
   * @param nrows number of data rows for a field
   */
  public Column(String name, String type, Object data, 
                boolean isNumeric){
      
    this.fieldname = name;
    this.data      = data;
    this.type      = type;
    this.isNumeric = isNumeric;
      
    setNrows();
    if (nrows > 0) {
	this.setMinMax();
    }
      
  }
   
  private void setNrows() {
      
      if (isNumeric) {
	  nrows = ((double[])data).length;
      } else {
	  nrows = ((String[])data).length;
      }
  }
  
  public int getNrows() {
      return nrows;
  }
    
    
  /** Create a new column
   *  which is the logarithm of the input.
   * @param col from which the log column is constructed 
   */
  public static Column logColumn (Column col) {
    
    Column lcol = new Column(col.fieldname, col.type, 
			     col.data, col.isNumeric);
      
    if (!col.isNumeric) {
	return lcol;
    }

    lcol.fieldname = "Log " + col.fieldname;
    
    int baddata = 0;
      
    double [] cdata = (double []) lcol.data;
      
    for(int i=0; i<col.nrows; i++) {
	
       double p = cdata[i];
       if (Double.isNaN(p) || p <= 0) {
	   cdata[i] = Double.NaN;
       } else {
	   cdata[i] = Math.log(p)/Math.log(10);
       System.out.println("logCol: " +p + " log=" + cdata[i]);
       }
    }
      
    lcol.setMinMax();
      
    return lcol;
  }


  /** Constructor which creates a Column with new data array
   *  The rest of the data comes from the passed in Column 
   *  This constructor is used by the Aitoff class
   */
  public Column(double indata[], Column col) {
      
    this.fieldname = col.fieldname;
    this.data      = new double[indata.length];
    this.nrows     = indata.length;
    this.type      = col.type;
    this.isNumeric = true;
      
    System.arraycopy(indata, 0, this.data, 0, indata.length);
      
    this.setMinMax();
  }

  /** Sets the min and max variables of class, if min==max
   *  it sets the max to a little greater and min to a little lesser
   */
  private void setMinMax() {

    if (!isNumeric) {
        this.min = Double.NaN;
        this.max = Double.NaN;
        return;
    }
      
    double [] ndata = (double []) data;
      
    double min = Double.NaN;
    double max = Double.NaN;

    for (int i=0;i<nrows;i++) {
	
         if (Double.isNaN(ndata[i])) {
             continue;
	     
	 } else if (Double.isNaN(min)) {
             min  = ndata[i];
	     max  = ndata[i];
	     
	 } else {
	     

             if (ndata[i] < min) {
                 min = ndata[i];
	     }
             if (ndata[i] > max) {
                 max = ndata[i];
             }
	 }
    }
      
    this.min = min;
    this.max = max;
  }
    
  /** Returns min value for column data
   * @return minimum value in data[] array 
   */
  public double getMin() {
      return this.min;
  }
    
  /** Returns data type  of column data
   * @return data type  of column data
   */
  public String getType() {
      return this.type;
  }

  /** Returns max value for column data
   * @return maximum value in data[] array 
   */
  public double getMax() {
      return this.max;
  }
    
  /** Returns field name for Column
   * @return field name for Column 
   */
  public String getField() {
      return this.fieldname;
  }
    
  /** Returns data array for Column object.
   *  Throws an exception is called for a String value.
   * @return data[] array in Column 
   */
  public double[] getData() throws ClassCastException {
      return (double[]) ((double[]) data).clone();
  }
    
    
  /** Called when min == max
   */
  private double getFactor(double value){
      
      value = Math.abs(value);
      if (value == 0) {
	  return 0.1;
      } else {
	  return 2*value/10;
      }
  }
}

