# $Id: CachedQuery.pm,v 1.13 2003/12/02 19:41:21 lmm Exp $
# $Log: CachedQuery.pm,v $
# Revision 1.13  2003/12/02 19:41:21  lmm
# Small fix to a string substitute command
#
# Revision 1.12  2002/03/15 23:37:19  cwh
# Added '-' to the list of chars that must be transformed to underscores.
#
# Revision 1.11  2002/01/22 22:41:01  cwh
# Changed fixFields() so that fieldnames that include a '.', or a parentheses,
# or a space, have those characters changed to underscore.  There are no doubt
# other characters that should also be swapped out.
# Also added another call to fixFields() to catch these field names.
#
# Revision 1.10  2001/11/07 23:34:16  cwh
# Moved everything except DBD/ into lib/Browse/, so as not to pollute the module
# namespace with our Sort, Search, CGI, etc.  This required putting Browse:: in
# front of every module call and package declaration.  Also any mod::new() had
# to be changed to mod->new().
#
# Revision 1.9  2001/06/20 22:57:31  cwh
# Got rid of $::writeHeader in calls to ::Error.
#
# Revision 1.8  2001/06/15 03:23:26  tam
# Delete query data from object structures prior
# to serialization to minimize size of serialized string.
#
# Revision 1.7  2001/06/15 01:32:12  tam
# Added getPrefix method to return temporary table name.
#
# Revision 1.6  2001/06/12 16:10:12  tam
# Fixed updates to metadata for CachedQuery
#     - Added query key to tableinfo
#     - Proper handling of prefixes
#     - RA/Dec entries in tableinfo.
#
# Revision 1.5  2001/06/08 19:46:36  cwh
# Now passing $::writeHeader to ::Error calls, so we don't get repeating headers.
#
# Revision 1.4  2001/05/31 18:21:26  cwh
# Added $::pcount to the temporary table name so that successive xcorrs with
# the same process id don't complain that the table name is already in use.
# (Affects only mod_perl.)
#
# Revision 1.3  2001/05/22 17:52:02  cwh
# Added Id and Log to files without them.
#

use strict;

# This package allows a query to be saved into
# the HEASARC database and then queried from there.
# This is used to support cross-correlations for
# tables not initially in the HEASARC database.
# It is fairly primitive and does not fully support
# the HeasarcQuery and QueryRequest interfaces.  It
# is tightly coupled to the CrossRequest object.

package Browse::CachedQuery;

use vars qw( @ISA );
@ISA = qw( Browse::HeasarcQuery );

sub new {

    my ($self,$oldRequest) = @_;
    my $obj = {};
    bless $obj;
    
    $$obj{oldrequest} = $oldRequest;
    $obj->store($oldRequest);
    if ($oldRequest->getPrefix()) {
        $obj->setPrefix();
    }
    
    return $obj;
}

sub localQuery {
    return 1;
}


sub makeQuery {

    my ($self) = @_;
    unless ($$self{"tableinfo"}{"table_name"} =~ / $$self{"prefix"}\b/) {   
        $$self{"tableinfo"}{"table_name"} .= " " . $$self{"prefix"};
    }
    unless ($$self{"table"} =~ / $$self{"prefix"}\b/) {   
        $$self{"table"} .= " " . $$self{"prefix"};
    }
}

# This method allows the user to set the prefix associated
# with this subquery.

sub setPrefix {

    my ($self, $prefix) = @_;
    $$self{prefix} = $prefix;
   
    my $tinfo = $$self{tableinfo};
    my $par;
    
    # Update the metadata column names.
    
    foreach $par (("right_ascension", "declination","start_time",
                   "sequence_number")) {
        my $val = $$tinfo{$par};
	if ($val) {
	    $$tinfo{$par} =~ s/^(\@)?/$1$prefix\_/;
	}
    }
    
    my $fields = $$self{"fields"};
    my $fld;
    
    # Update the columns in the table.
    
    foreach $fld (@$fields) {
    
        # Don't put in the prefix if it is already there
        if ($fld =~ /^$prefix\./) {
	    last;
	}
	
	# Handle the w3_xxx as xxx case
	if ($fld =~ / as /) {
	    my ($beg, $end) = split(" as ", $fld);
	    my ($xfix, $xend) = split('\_', $end, 2);
	    $fld = "$prefix.$beg as $xfix\_$prefix\_$xend";
	} else {
	    $fld = "$prefix.$fld as $prefix\_$fld";
	}
    }

    # Update the parameter info.
    
    my $pinfo = $$self{"paraminfo"};
    foreach $fld(keys(%$pinfo)) {
        if (substr($fld,0,1) eq "#") {
	    next;
	}
	if ($fld =~ /^$prefix\./) {
	    last;
	}
	my $row = $$pinfo{$fld};
	my $nfld = $prefix."_".$fld;
	$$row{"parameter_name"} = $nfld;
	$$pinfo{$nfld} = $row;
	delete($$pinfo{$fld});
    }
#Browse::Utils::DumpRef($pinfo);
}
    
    

# Copy data from the external query into the local database

sub store {

    # Request should be an object of the "QueryRequest" class.
    my ($self, $request) = @_;

    # Probably this should be greater!
    $request->setMaxRows(9999);
    $$request{"status"} = "initiate";
    my $meta = $request->getResult();
    
    if ($$meta{"type"} ne "meta") {
        Browse::Utils::Error("Unable to cache query.  First result not metadata.");
    }
    
    $$self{meta} = $meta;
    
    
    my $paraminfo = $$meta{"paraminfo"};
    my $tableinfo = $$meta{"tableinfo"};
    
    $$self{"table"}     = $$tableinfo{"table_name"};
    $$self{"paraminfo"} = $paraminfo;
    $$self{"tableinfo"} = $tableinfo;


    $$self{"fields"}    = $request->getFields();
     
    
    my $nrow  = 0;
    my $nerr  = 0;
    my $nfld  = 0;
    my $nskip = 0;
    
    my $table;
    
    my $block;
    my @quotes;
    
    # Now get the data row by row.
    
    while ($block = $request->getResult()) {
    
	if (!defined($$block{"type"})) {
	    last;
	}
    
        # Only process the first table.
        if ($$block{"type"} ne "data") {
	    last;
	}
	
        if ($nrow == 0) {
	    
	    # Generate the table.
	    my $fld;
	    my $typ;
	    my $names = $$block{"names"};
	    my $types = $$block{"types"};
	    my @defs;
	    my $i;
	    for ($i=0; $i <= $#$names; $i += 1) {
	        $fld = $$names[$i];
		$typ = $$types[$i];
		
		$fld = $self->fixFields($fld, $$self{"fields"});
	    
	        my ($prefix, $suffix) = split(" as ", $fld);
		if (defined($suffix)) {
		    $fld = $suffix;
		}
		
		if ($fld =~ "^w3_") {
		    $nskip += 1;
		    next;
		}
		
		my $form;
		
		if ($typ == 1) {
		
		     $form = $$paraminfo{$fld}{"parameter_format"};
		     $form =~ s/\:.*//;
		     $quotes[$i] = "'";
		     $form =~ s/char(.*)/char\($1\)/;
		     
		} elsif ($typ == -1) {
		     $form = "text";
		     $quotes[$i] = "'";
		     
		} elsif ($typ == 6) {
		
		     $form = "double precision";
		     $quotes[$i] = "";
		     
		} elsif ($typ == 7) {
		     $form = "real";
		     $quotes[$i] = "";
		     
		} else {
		     $form = "int";
		     $quotes[$i] = "";
		}
		
		push(@defs, "$fld $form null");
		$nfld += 1;
	    }
	    
	    # Note that use of # to create a temporary
	    # table name is a Sybase convention.
	    
	    $table = "#t$$".$::pcount;
	    $$self{"ttable"} = $table;
	    
	    my $sql = "create table $table (".join(",", @defs).")";
	    
	    Browse::DBQ::execute($sql);
	    if ($DBI::errstr) {
	        Browse::Utils::Error("Error caching table: $DBI::errstr. SQL was: $sql");
	    }
	    
	    my $dec = $$tableinfo{"declination"};
	    
	    if (defined($dec)) {
	    
	        $dec =~ s/^\@//;
		
		if (grep($_ eq $dec, @$names)) {
		
		    $sql = "create clustered index $table"."x on $table($dec) ".
		           " with allow_dup_row";
			   
		    Browse::DBQ::execute($sql);
		    if ($DBI::errstr) {
	                Browse::Utils::Error("Error indexing cache table: $DBI::errstr");
		    }
		}
	    }
	    
	    my $flds = $$self{fields};
	    if ($#$flds < 0) {
	        my $ff = [];
		my $b = $$block{"names"};
	        my @f = @$b;
		
		while ($f[0] =~ /^w3\_/) {
		    shift(@f);
		}
		foreach (@f) {
		    $_ = $self->fixFields($_);
		}
		
	        $$self{fields} = \@f;
	    }
	}
	
	# Now insert the current rows data.
	my $sql = "insert into $table values (";
	my $data = $$block{"data"};
	
	my $i;
	my $comma;
	my $names = $$block{"names"};
	for ($i=$nskip; $i <= $#$names; $i += 1) {
	    my $fld = $$data[$i];
	    $fld =~ s/ *$//;
	    $fld = $quotes[$i].$fld.$quotes[$i];
	    if (length($fld)) {
	        $sql .= $comma.$fld;
	    } else {
	        $sql .= $comma."null";
	    }
	    $comma = ",";
	}
	$sql .= ")";
	
	Browse::DBQ::execute($sql);
	if ($DBI::errstr) {
	    $nerr += 1;
	}
	
	$nrow += 1;
    }
    
    $$self{"temptable"} = $table;

    # Add in the query key to the metadata.
    
    my $qkey = $$tableinfo{"query_key"};
	
    if ($qkey) {
	    
	my @qk = split(",", $qkey);
	my $k;
	my $f = $$self{fields};
	foreach $k(@qk) {
	    $k = $self->fixFields($k);
	    unshift(@$f, "$k as w3_$k");
	}
    }
    
    #Get rid of this for efficiency.
    delete($$request{"tree"});
    $request->qclose();
		
	        
}

# Return the temporary table name.
sub getTable {
    my ($self) = @_;
    
    return  $$self{"ttable"}." ".$self->getPrefix();
}

# Check for illegal field names.

sub fixFields {

    my ($self, $fld, $fields) = @_;
    
    if ($fld =~ /^count$|^\d|[-.(){}\s]/i ) {
    
        my $xfld = $fld;
        $xfld =~ s/[-.(){}\/\s]/_/g;
        $xfld = "_$fld" if $fld =~ /^count$|^\d/i;
        my $pin = $$self{paraminfo};
	my $line = $$pin{$fld};
	$$pin{$xfld} = $line;
	delete($$pin{$fld});
	
	my $afld;
	foreach $afld (@$fields) {
	    if ($afld eq $fld) {
	        $afld =~ s/^/\_/;
	    }
	}
	
	
	return $xfld;
	
    } else {

    
        return $fld;
    }
}

sub getFields {
    return undef;
}

sub getWhere {
    # We've already filtered the data.
    return undef;
}

sub getSource {
    my ($self) = @_;
    
    return " $$self{temptable} $$self{prefix} "
}

sub getOrder {
    return undef;
}

1;
