# $Id: Parameter.pm,v 1.88 2009/02/25 14:44:18 lmm Exp $
# $Log: Parameter.pm,v $
# Revision 1.88  2009/02/25 14:44:18  lmm
# Number of rows returned from each table query is now written to sql log
#   file.
#
# Revision 1.87  2008/12/15 18:28:46  lmm
# Added constraint processing for multiple coordinates (other_ra_dec) in
#   heasarc tables.
#
# Revision 1.86  2007/10/16 20:38:52  lmm
# Added zzdp table specific code to make queries selecting dp_tag
#    case-sensitive to decrease run time.
#
# Revision 1.85  2007/10/09 14:33:20  lmm
# Made MakeConeQuery() a class method.  Broke apart other code to make new
#   methods for Tabbed interface and hopefully better organization.
#
# Revision 1.84  2007/03/26 19:22:10  lmm
# Allow parameter queries on _unique_id (change provided by TAM)
#
# Revision 1.83  2007/01/08 16:28:26  lmm
# Improved sql query build for related links.
# Improved error message displayed to user for incorrectly formatting
#   parameter values.
#
# Revision 1.82  2006/11/29 15:59:37  lmm
# Added logging of Keyword Search origin (Archive page, HEASARC page, etc)
#
# Revision 1.81  2006/10/24 18:58:15  lmm
# Made changes (provided by TAM) to parameter parsing to better handle
#   user entered SQL.
#
# Revision 1.80  2006/07/03 13:59:24  lmm
# Added class method setDisplayMode to allow display to be set ouside
#   of class.
#
# Revision 1.79  2006/06/23 15:08:00  lmm
# Changed module to class (for Tabbed Display)
# Removed routines no longer used
# Added GRB name resolver
#
# Revision 1.78  2006/01/31 15:14:43  lmm
# Defined a query type for the new Keyword Search for logging.
#
# Revision 1.77  2006/01/12 20:04:56  lmm
# Further changes for better table parameter checking - skip table links
#
# Revision 1.76  2006/01/12 17:14:31  lmm
# Made changes to table parameter check for Browse Keyword Interface more
#   robust (code contributed by E. Sabol)
# Changed call to Display::resultsSummary() to reflect new argument
#
# Revision 1.75  2005/12/13 21:04:42  lmm
# Made changes to allow meta data table searches (changes contributed
#   by tam).
# Added table list to YourSearchWas() call.
#
# Revision 1.74  2005/08/19 16:53:14  lmm
# Fixed bug where resolving metaparameters in the sorts for one table query
#   was affecting sorts for all subsequent table queries
#
# Revision 1.73  2005/08/17 19:11:19  lmm
# Added FilterTableQuals() to address qualifiers that have table name
#   prefixes (code provided by Ed Sabol)
# Fixed bug where unit and format were not being defined for virtual
#   parameters (bparam_metaparam types)
#
# Revision 1.72  2005/06/23 20:52:36  lmm
# Using null as a parameter constraint for time and coordinates is now
#   allowed.
#
# Revision 1.71  2005/05/25 17:16:09  lmm
# Improved error message for coordinates not recognized by name resolvers.
#
# Revision 1.70  2005/03/07 17:34:50  lmm
# Fixed bug that prevented parameter searches (e.g., exposure = 0) from being
#   handled correctly
#
# Revision 1.69  2005/02/15 21:40:18  lmm
# Fixed cache determination bug in MakeConeQuery().
#
# Revision 1.68  2005/02/15 16:01:18  lmm
# Moved FreezeThaw.pm again to more appropriate library directory
#
# Revision 1.67  2005/02/01 16:10:30  lmm
# FreezeThaw.pm has been moved out the Browse directory to the
#   lib directory since it is an imported package
#
# Revision 1.66  2005/01/19 14:53:29  lmm
# Now calling Display method resultsSummary() method directly instead of
#   calling Directions().
#
# Revision 1.65  2004/12/20 17:31:57  lmm
# Set maxlimit to 0 for query if it has been set to the display string
#   "(no limit)"
#
# Revision 1.64  2004/12/15 20:17:32  lmm
# Added change to make sure sorting direction is "up" or "down" for requery
# Changed check for sorting direction to be more accurate in OrderClause()
# Cleaned up a bit by removing numerous debug statements
#
# Revision 1.63  2004/02/26 15:23:29  lmm
# Fixed bug in and improved code that checks differences between name
#   resolvers.  New code provided by Ed Sabol.
#
# Revision 1.62  2004/02/17 18:33:01  lmm
# Removed eval() of some user input due to problems with certain formats.
#
# Revision 1.61  2004/02/09 21:14:28  lmm
# Improved integer/float type checking on Additional Parameters page
#
# Revision 1.60  2004/01/15 20:53:28  lmm
# Change in method to save non-valid tables for requery
#
# Revision 1.59  2003/12/02 20:02:15  lmm
# Added serialization of query string to be used as download.
# Added storage for empty table information in case of requery.
#
# Revision 1.58  2003/09/03 20:57:33  lmm
# Table legend appears only once instead of before every table
#
# Revision 1.57  2003/08/29 16:43:12  lmm
# Defined text for page header
# Changes to MakeQuals made by cwh
#
# Revision 1.56  2003/07/31 18:49:01  cwh
# Added new call to displayer->legend(), which is broken out of tableHeader,
#   since not all display modes will want to display the legend.
#
# Revision 1.55  2003/07/31 14:08:06  cwh
# MakeConeQuery didn't need table_info.
# Added the ability to query by expressions by ignoring stuff in parens.
#
# Revision 1.54  2003/05/20 17:38:58  cwh
# Added new parameter handling.
# If name resolvers return different positions, the difference is reported with
#   one decimal place precision.
# Passing along the table delimiter if it exists.
#
# Revision 1.53  2002/12/13 22:04:02  cwh
# Replaced 'this object' with '<i>$entry</i>' in info statements regarding the
#   resolver.
# Added if ($object->resolver_used()) check before printing warning, since if
#   coordinates are specified instead of an object name, the resolver_used wont
#   exist.  We don't want a warning about the resolver in that case.
#
# Revision 1.52  2002/11/07 17:32:13  cwh
# Now adding a comment if a secondary name resolver fails to find an object,
# and a warning if the primary name resolver fails to find it.
#
# Revision 1.51  2002/09/13 20:33:55  cwh
# Removed a heasarc-specific library.
# Now handling param queries on time ranges for tables with both start and
#   end times.
# Commented out failure on first table if first table has no def radius.
# New name resolving routine - checks both NED and SIMBAD, reports difference
#   if significant.  Resolver message is passed to Search(), then to
#   displayer->YourSearchWas.
# Changed calls to displayer->setMaxRows and setInput.
# Passing less info to YourSearchWas.
#
# Revision 1.50  2002/05/14 18:12:14  cwh
# Changed the call to YourSearchWas to use display object.
# Added another display->closeTable, for the excel display method.
#
# Revision 1.49  2002/03/22 22:56:51  cwh
# Commented out heading for links to results tables.
#
# Revision 1.48  2002/03/17 18:33:21  cwh
# Added use Astro::Utils::SLAConvert;
# Changed Directions() to print a list of tables names with links to results
#    instead of the one-line 'Please select...'
# Not passing first table table-info to Search, Search now gets all table_info
#    hash at once.
#
# Revision 1.47  2002/02/12 20:53:39  cwh
# Changed the "if defined $w3min"s to "if $w3min =~ /\d/", since that includes
# all valid numbers and 0, but excludes defined nulls, which are valid but
# shouldn't be converted to HMS.  Same for w3max.
#
# Revision 1.46  2001/12/05 23:39:03  cwh
# If the table name given doesn't exist, call Display_tables_list() to give
# the batch user a list of available tables. (and state that it is unknown)
# Setting $$displayer{"offset_var"} for later.
#
# Revision 1.45  2001/11/08 20:34:33  cwh
# Added Browse:: to sub calls.
# Calling displayer->setInput to set input string for Heraform().
# Return from MakeConeQuery if no $in->{Entry}.
#
# Revision 1.44  2001/09/24 17:48:42  cwh
# Got rid of extra spaces around time fields - caused error in translation.
# Added default name resolver - NED-type queries were failing without it.
#
# Revision 1.43  2001/09/15 15:05:23  cwh
# Changed $batch_version to $batchflag - the version is now in w3config.pl.
# Always call MakeConeQuery.
# Now using Astro.pm instead of CoordConv.pm.
# Not doing iso2mjd for input already in MJD.
# Moved Time handling into MakeParamQuery where it belongs.
# Ra and dec field names no longer in position string.
# Got rid of all that ugly savequals/addquals stuff.
# Resolve metaparams that may be sort vars.
#
# Revision 1.42  2001/08/08 21:01:20  cwh
# Now passing Time to YourSearchWas.
# Handling Time entry from front page:  add it as a qualifier just before
#    querying each table.  Keep a copy of the original qualifiers and only
#    add the time qual unique to that table, since the time field can be
#    different.   Print search error if not searchable by time.
#
# Revision 1.41  2001/07/23 16:48:01  cwh
# Changed the part which calls Astro::Object::Resolver to handle using the
#   local caches.
# Added a bit to the positions line to preserve which service or cache
#   resolved the object.
# Calls the Results::YourSearchWas() subroutine to get the search criteria
#   box, prints the box by calling showText(), then prints Directions().
# Removed section that was preserving the list of missions.  Not using that
#   anymore.
#
# Revision 1.40  2001/07/17 18:40:53  cwh
# Added logic for logging Batch and Basic searches, and added $quals to logged
# info for the Where clause.
#
# Revision 1.39  2001/06/20 22:57:33  cwh
# Got rid of $::writeHeader in calls to ::Error.
#
# Revision 1.38  2001/06/14 22:54:37  cwh
# Now passing along missions to the request object for Vizier generic queries.
#
# Revision 1.37  2001/06/14 21:30:38  tam
# Deleted lowercasing of parameter names so that case sensitive
# VizieR names work properly.
#
# Revision 1.36  2001/05/30 21:22:48  cwh
# Added use lib's so that the Astro lib in /www/htdocs/cgi-bin/lib/heasarc/ will
#   be found.  Had to re-add the './lib' and '.' libs to keep the order right.
#
# Revision 1.35  2001/05/30 18:05:57  cwh
# Added Tom's real number regex that handles scientific notation.
#
# Revision 1.34  2001/05/24 17:05:55  cwh
# Now passing along max rows to self so that when user does a resort, only max
# rows are returned.
#
# Revision 1.33  2001/05/22 18:12:14  tam
# Fix problem with class searches.
#
# Revision 1.32  2001/05/17 17:52:18  cwh
# Added message when max rows = rows displayed.
#
# Revision 1.31  2001/05/03 18:58:59  cwh
# Added support for $in->{'Position'} as well as $in->{'Entry'}.
# Changed all ($w3min ne '') to (defined $w3min), same for w3max, because
#    var=0 wasn't being allowed as a valid param query.
# Changed the parsing order so that if Fields=All and some but not all field's
#    view checkboxes were on, the All is disregarded.
# Added support for multiple sort keys.
#
# Revision 1.30  2001/04/25 17:46:48  cwh
# Tom almost completely rewrote this.
#
# Revision 1.29  2001/04/03 16:22:38  cwh
# Commented out call to DisplayOptions, and changed call to DisplaySets so that
# it calls DisplayOptions.  This is so that the sets and actions boxes are displayed together.
#
# Revision 1.28  2001/01/22 20:42:14  cwh
# Changed the args for PrintTableName, and added call to PrintTableButtons.
#
# Revision 1.27  2001/01/02 23:16:49  cwh
# Added css span info to navtrail.
#
# Revision 1.26  2000/12/04 19:38:07  cwh
# Added navtrail variable which is passed from script to script to give users
# some navigation.
#
# Revision 1.25  2000/08/04 20:04:04  saima
# Made changes to the parameter searches so that all searches are
#   case-insensitive (=, != and wildcards) unless they involve
#   parameters with >255character length.
#
# Revision 1.24  2000/08/04 15:24:01  saima
# Added $textdisplay to list of parameters sent to
#    Results::PrintTableName() so that plot/xcorr icons dont get
#    displayed in textmode.
#
# Revision 1.23  2000/03/09 22:11:17  saima
# Default ResultMax rows displayed set to 1000 rows. Name change to Browse.
#
# Revision 1.22  2000/01/27 18:34:17  saima
# Modified so that if the flag UKeyDisplay is on, then unique key
#   columns are displayed.
#
# Revision 1.21  2000/01/10 19:24:07  saima
# Several changes include: parameter search now accepts empty query and
#   no where clause has to be specified; code can handle 'not range'
#   queries; code handles batch results; prior table information
#   about default, notdefault and mjd parameters is not required, a
#   database call gets this meta data.
#
# Revision 1.20  1999/10/07 20:13:51  saima
# Added aitoff_lii and aitoff_bii to $plotparam for displaying aitoff
#   projections in galactic coordinate system.
#
# Revision 1.19  1999/09/24 18:48:51  saima
# Sends $hits to Results::PrintTableName().
#
# Revision 1.18  1999/09/21 14:43:52  saima
# Fixed a minor bug in recognizing float numbers like '0.'.
#
# Revision 1.17  1999/09/20 20:24:43  saima
# Fixed a bug in code where plot icon was not getting displayed.
#
# Revision 1.16  1999/09/13 18:25:45  saima
# Added a more detailed message for 'Empty query not accepted'.
#
# Revision 1.15  1999/09/07 19:04:58  saima
# Plotting is not displayed on the Results page anymore. An icon
#   starts up another browser window which displays the plotting
#   utility.
#
# Revision 1.14  1999/08/31 19:03:50  saima
# Modified so that $numeric contains all parameter names for numeric
#   columns. This is used by server-side plotting.
#
# Revision 1.13  1999/06/01 17:59:00  tam
# Fixed bug searching on with val=0.
#   Added limits on number of rows to be retrieved from DB.
#
# Revision 1.12  1999/03/18 22:15:50  saima
# Added logging.
#
# Revision 1.10  1999/02/24 22:13:09  saima
# Fixed code so that sorts after a multiple class search work correctly.
#
# Revision 1.9  1999/02/24 16:21:04  saima
# Fixed bug where queries on 'text' type parameters were not
#   being treated differently. Sybase does not allow 'text' types (> 255
#   char) to be sorted, upper() 'ed. Also, '=', '!=' and 'range' queries
#   are not allowed on these parameters.
#
# Revision 1.8  1999/02/22 19:29:13  saima
# Fixed a problem where $in->{'Entry'} was not being checked properly.
#   No-conesearchable tables could not be searched...
#
# Revision 1.7  1999/02/17 21:34:02  saima
# Added extra checks to make sure user does not attempt to run Cone search
#   on a non-conesearchable tables.
#
# Revision 1.6  1999/02/08 20:08:20  saima
# Modified SQL query creation so that users can search on mulitple
#   class values.
#
# Revision 1.5  1999/02/03 20:27:11  saima
# For multiple table parameter searches, if a user sets a common
#   metabase parameter, the Search() routine looks for the
#   _metaparam_ string in the where clause and replaces the metabase
#   parameter name with the table parameter name.
#
# Revision 1.4  1999/01/28 19:53:27  saima
# If 10 *'s are in the min/max input areas they are replaced by
#   ''. This happens when user sets operator to '=' or '!='.
#
# Revision 1.3  1999/01/26 23:22:26  saima
# Added extra checks so that script looks if default_search_radius is
#   defined.
#
# Revision 1.2  1999/01/14 18:08:58  saima
# Modified code so that table name is lower-cased in SelectClause().
#
# Revision 1.1  1998/12/23 16:41:13  saima
# Initial Revision.
#
#
#=======================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name & address intact.
#=======================================================================#
# Module Name:    Parameter.pm
# Module Author:  Saima Zobair, Raytheon ITSS
# Date:           Nov, 98
# Project:        Browse : HEASARC
#=======================================================================#
# Module Description:
#
#
#=======================================================================#
# Outline:
#
#
#=======================================================================#

use strict;
use Astro;
use Astro::Utils::SLAConvert;
use Browse::Parse;
use Browse::Log;
use Browse::Display;
use Browse::QueryRequest;
use FreezeThaw qw(freeze thaw cmpStr safeFreeze cmpStrHard);

package Browse::Parameter;

# Create all Parameter objects.  The only argument is
# the class of the object to be created.

sub new {

    my ($self,$mode) = @_;
    
    my $obj = {};
    if ($mode) {
        $mode =~ /([\-\w]*)/; $mode=$1;   # untaint $mode

        #dynamically load module
        $mode = "Browse::$mode";
        eval "package Browse::_firesafe; require $mode";
        bless $obj, $mode;
    } else {
        bless $obj;
    }
    
    return $obj;
}

#-- Class method for setting display mode
sub setDisplayMode {
   my ($self, $mode) = @_;

  if (defined($mode)) {
     $self->{"displayer"} = Browse::Display->new($mode); 
  }
}


#------------------------------------------------------------------------
sub PrepareSearch {

  $0 = 'Parameter.pm'; # required for MailMessage
  
  my ($self,$in,$tables) = @_;

  print "in PrepareSearch ". scalar(@$tables) . "<br>" if $::debug; 
  my ($batchflag, @table) = Browse::Parse::GetTableInfo($tables);

  #--- Set displayer 
  my($textdisplay) = ($in->{'TextDisplay'} =~ /on/i) ? 1 : 0;

  my $mode = $$in{"displaymode"};
    
  if (!defined($mode)) {
     $mode = "TextDisplay" if $textdisplay;
     $mode= "BatchDisplay" if $batchflag;
  }

  $self->setDisplayMode($mode);

  #--- check for table for batch table display or web error
  my ($table_info) = { Browse::HDB::TableInfo($table[0]) };
 
  unless ($$table_info{table_name} =~ /\w/) {
    $table[0] =~ s/$::local_origin//ig;
    my $error = "Table ".$table[0] ." does not seem to exist!\n\n";

    if ($batchflag) { 
          print "Content-type:text/plain\n\nBatchStart\n"; 
          if ($table[0] =~ /\w/) { 
            print "$error\n\n"; 
          } else {
            print "Please specify a table to query.\n\n";
          }
          print "Available tables:\n\n";
          Browse::HDB::Display_tables_list("BatchDisplay"); 
          print "BatchEnd\n";
    } else { 
          Browse::Utils::Error($error); 
    }
    exit; 
  }
  my ($param_info) = Browse::Parse::ParamInfo($in);

  #my ($positions,$resmsg)  = MakeConeQuery($in);
  my ($positions,$resmsg)  = $self->MakeConeQuery($in);
  my ($qualifiers) = MakeParamQuery2($in, $table_info);
  
  $self->Search(\@table, $positions, $qualifiers, $in, $batchflag,$resmsg);

  my $query_type = "Parameter Search";
  $query_type = "Batch Search" if defined $batchflag;
  $query_type = "Basic Search" if defined $in->{'Basic'};
  $query_type = "Keyword Search" if defined $in->{'tableinfo'};
  $query_type = "Keyword Search:Archive" 
      if (defined $in->{'tableinfo'} && defined $in->{'archivepage'});
  $query_type = "Keyword Search:Heasarc Home" 
     if (defined $in->{'tableinfo'} && defined $in->{'mainpage'});

  Browse::Log::Query($0,$in,$query_type,$tables,join(',',@$qualifiers));
  return $batchflag;

} #eosub--Parameter::PrepareSearch
#------------------------------------------------------------------------

sub MakeParamQuery2 {
# gotta check for defined($var) instead of just $var since '0' is valid

  my($in, $table_info) = @_;
  my @allqualifiers;

  my $parameters = Browse::HDB::ParamInfo($$table_info{"table_name"});
  #Browse::Utils::DumpRef($in) if $::debug;
  my $tparameters = $$parameters{$$table_info{"table_name"}};
  #Browse::Utils::DumpRef($parameters) if $::debug;

  if ($in->{'Qualifier'}) {
      @allqualifiers = split('\0', $in->{'Qualifier'});
      #print "Parameter:: MPQ2 returning ". @allqualifiers ."<br>" if $::debug;
      return (\@allqualifiers);
  }
  my $end_time_done =0;

  #get param list
  my %paramlist;
  while ( my ($param,$pvalue) = each %$in) {
    #translate old style param info
    if ($param =~ /w3xparam/) {
        foreach my $p ( split("\0", Browse::CGI::Decode($pvalue)) ) {
           $p =~ s/(^name=|&.*$|)//gi;

           $$in{"w3op_$p"}=~s/(range|between)/../i;
           $$in{"w3op_$p"}=~s/(outside)/!../i;
           if ($$in{"w3max_$p"} =~ /\w/) {
               $pvalue = $$in{"w3min_$p"} . $$in{"w3op_$p"} . $$in{"w3max_$p"};
           } else {
               $pvalue = $$in{"w3op_$p"} . $$in{"w3min_$p"};
           }
           $pvalue = Browse::CGI::Decode($pvalue);
           $paramlist{$p} = $pvalue;

        }
      
      next; 
    }
    #next unless $param =~ /^bparam_|^Time/ and $pvalue;
    next unless $param =~ /^bparam_|^Time/ and ($pvalue =~ /\S/);
    next if $param =~ /::(format|unit)/;
    $paramlist{$param} = $pvalue;
    #print "Parameter:: MPQ2 setting $param to pvalue $pvalue<br>" if $::debug;
  }

  # process params
  while ( my ($param,$pvalue) = each %paramlist) {

    #--- This converts = to == for zzlink table links.  It will prevent
    #--- the ineffiecient upper() from being used in the sql that 
    #--- performs the link (help from ed sabol)
    if ($param =~ /bparam_/ && $in->{popupFrom} =~/Linked/ 
	&& $pvalue =~ /={1}/) {
       $pvalue =~ s/=/==/;
    }

    #print "Parameter: looking at $param<br>\n" if $::debug;
    my ($type,$unit);

    #--- Try to get units and format for virtual parameters
    if ($param =~ /^bparam__metaparam/) {
      $type=$in->{$param ."::format"} if $in->{$param ."::format"}; 
      $unit=$in->{$param ."::unit"}   if $in->{$param ."::unit"}; 
    } 

    $param =~ s/^bparam_//;
    if ($param =~ /^Time/) {
       $param = "_metaparam_start_time";
       $type = "char" unless $type;
       $unit = "mjd" unless $unit;
    } else {
       $type = $$tparameters{$param}{"parameter_format"} unless $type;
       $unit = $$tparameters{$param}{"parameter_unit"} unless $unit;
    }

    if ($type eq "class") {
      $pvalue =~ s/\s//g;
      $pvalue =  join(',',split("\0", $pvalue));
    }

    if ($type !~ /^char/) {
      #clean up any whitespace on input
      $pvalue =~ s/(^\s+|\s+$)//g;
    }
    
    # this handles ranges for tables with start and end times
    if ($param =~ /$$table_info{"start_time"}/ and $$table_info{"end_time"} 
         and !$end_time_done and $pvalue =~ /\.\./) {
       my ($pmin, $pmax) = split(/\s*\.\.\s*/,$pvalue,2);
       my $pquals2 = MakeQuals("_metaparam_start_time", "<=$pmax", $type, $unit, $table_info);
       #print "Parameter::  p=@$pquals2 <br>" if $::debug;
       push(@allqualifiers, @$pquals2);
       $pquals2 = MakeQuals("_metaparam_end_time", ">=$pmin", $type, $unit, $table_info);
       push(@allqualifiers, @$pquals2);
       #print "Parameter::  p=@$pquals2 <br>" if $::debug;
       $end_time_done =1;
       next;
    }

    my $pquals = MakeQuals($param, $pvalue, $type, $unit, $table_info);
    push(@allqualifiers, @$pquals);

  } #end each parameter
  return(\@allqualifiers);

} #eosub--Parameter::MakeParamQuery2
#--------------------------------------------------------------------
  sub MakeQuals {
    my @qualifiers;
    my ($param, $pvalue, $type, $unit, $table_info) = @_;
    #my (@pvals) = split(/(\bor\b|;|\|)/i,$pvalue); #'or', ';', or '|' for alts
    my($above255) = 0;

    # look for 'or', ';', or '|' to split expressions.  Only ; is value
    # when the user has entered an SQL expression (indicated by a '(' at the
    # beginning of the first parameter.
    
    my (@pvals); 
    if ($pvalue !~ /^\s*(<|>|=|>=|<=|!=)?\s*\(/ ) {
       @pvals = split(/\s*;\s*|\0|\s*\bor\b\s*|\s*\|\s*/i,$pvalue); 
    } else {
       @pvals = split(/\s*;\s*/i,$pvalue); 
    }
    
    foreach my $pval (@pvals) {
        # now can manipulate and form quals
        #print "Parameter:: pval => $pval<br>" if $::debug;

        my ($pmin,$pmax,$pop);
        if ($pval =~ /\.\./) {
           ($pmin,$pmax) = split(/\s*\!?\.\.\s*/,$pval);

           $pop = ($pval =~ /\!/? '!':'') . 'range';

           #print "init pmin:op-> $pop, min-> $pppmin, max-> $pppmax<br>\n" 
		#if $::debug;
        } else {
           ($pop, $pmin) = $pval =~ /^\s*([=<>!]*)\s*(.*)$/;
        
        }
        #if ($pvalue =~ /^bparam_/ && $pop =~/=/) {$pop = "==";}
        $pop ||= "=";

        #print "Parameter::param=$param pval:$pval pop:$pop<br>" if $::debug;
        # commented out due to eval problems -- lmm 2/2004
        #$pmin = eval{ $pmin };

        #--- find names of other ra/dec columns so they can be treated 
        #--- like standard ra and dec
        my ($oras, $odecs) = Browse::Display::otherRaDecs($table_info);

 
   # if the user has entered an expression with parens, don't evaluate it,
   # just pass it along unaltered.
   unless ($pmin =~ /[(]/ or $pmax=~/[(]/) {

        if ($param eq $table_info->{'right_ascension'} ) {
             # evaluate an expression that might be entered.  
             #2/2004 lmm commented out because of problems evaling HMS format
             #$pmin =~ s/(.*)/$1/ee;
             #$pmax =~ s/(.*)/$1/ee;
             #my ($accuracy) = $type =~ /:\d*\.(\d+)f/; 
             #Browse::Utils::DumpRef($accuracy) if $::debug;

             if ($pmin !~ /\bnull\b/) {
                $pmin = "error" unless ($pmin =~ /\d/ or $pmax =~ /\d/);
                $pmin = Astro::CoordConv::HMS2Decimal($pmin) if ($pmin =~ /\d/);
                $pmax = Astro::CoordConv::HMS2Decimal($pmax) if ($pmax =~ /\d/);
             }
             if (($pmin =~ /error/i) || ($pmax =~ /error/i)) {
                     Browse::Utils::Error("Error: Invalid right ascension ".
                     "coordinate input.<br>Please go back and enter an ".
                     "appropriate coordinate value.");
             }
        } elsif ( grep /^$param$/i, @$oras ) {
             my ($accuracy) = $type =~ /:\d*\.(\d+)f/; 
             #Browse::Utils::DumpRef($accuracy) if $::debug;
             if ($pmin !~ /\bnull\b/) {  #--- allow null search
                $pmin = "error" unless ($pmin =~ /\d/ or $pmax =~ /\d/);
                $pmin = Astro::CoordConv::HMS2Decimal($pmin, $accuracy) 
			if ($pmin =~ /\d/);
                $pmax = Astro::CoordConv::HMS2Decimal($pmax, $accuracy) 
			if ($pmax =~ /\d/);
             }
             #print "Parameter:  $type acc=$accuracy pmin=$pmin pmax=$pmax <br>"
		 #if $::debug;
             if (($pmin =~ /error/i) || ($pmax =~ /error/i)) {
                     Browse::Utils::Error("Error: Invalid right ascension ".
                     "coordinate input.<br>Please go back and enter an ".
                     "appropriate coordinate value.");
             }
        } elsif ($param eq $table_info->{'declination'}) {
             # evaluate an expression that might be entered
             #2/2004 lmm commented out because of problems evaling HMS format
             #$pmin =~ s/(.*)/$1/ee;
             #$pmax =~ s/(.*)/$1/ee;
             if ($pmin !~ /\bnull\b/) {  #--- allow null search
                $pmin = "error" unless ($pmin =~ /\d/ or $pmax =~ /\d/);
                $pmin = Astro::CoordConv::DMS2Decimal($pmin) if ($pmin =~ /\d/);
                $pmax = Astro::CoordConv::DMS2Decimal($pmax) if ($pmax =~ /\d/);
             }
             if (($pmin =~ /error/i) || ($pmax =~ /error/i)) {
                     Browse::Utils::Error("Error: Invalid declination ".
                     "coordinate input.<br>Please go back and enter an ".
                     "appropriate coordinate value.");
             }
        } elsif ( grep /^$param$/i, @$odecs ) {
             my ($accuracy) = $type =~ /:\d*\.(\d+)f/; 
             Browse::Utils::DumpRef($accuracy) if $::debug;
             if ($pmin !~ /\bnull\b/) {  #--- allow null search
                $pmin = "error" unless ($pmin =~ /\d/ or $pmax =~ /\d/);
                $pmin = Astro::CoordConv::DMS2Decimal($pmin,$accuracy) 
			if ($pmin =~ /\d/);
                $pmax = Astao::CoordConv::DMS2Decimal($pmax,$accuracy) 
			if ($pmax =~ /\d/);
             }
             print "Parameter:decs $type acc=$accuracy pmin=$pmin pmax=$pmax<br>"
		if $::debug;
             if (($pmin =~ /error/i) || ($pmax =~ /error/i)) {
                     Browse::Utils::Error("Error: Invalid declination ".
                     "coordinate input.<br>Please go back and enter an ".
                     "appropriate coordinate value.");
             }
        } elsif ($unit =~ /^mjd/) {
             if ($pmin !~ /\bnull\b/) {  #--- allow null search
                $pmin = "error" unless ($pmin =~ /\d/ or $pmax =~ /\d/);
                $pmin = DateTime::iso2mjd($pmin) if ($pmin =~ /\d/ and 
                        $pmin !~ /\d{5}\.?\d?/);
                $pmax = DateTime::iso2mjd($pmax) if ($pmax =~ /\d/ and 
                        $pmax !~ /\d{5}\.?\d?/);
             }
             if (($pmin =~ /error/i) || ($pmax =~ /error/i)) {
                 Browse::Utils::Error("Error: Date values must be entered in ".
                    "either the ISO standard format (YYYY-MM-DD hh:mm:ss) or ".
                    "in Modified Julian Date format (integer or real number). ".
                    "Please go back and enter an appropriate date value for ".
                    "parameter $param.");
             }
             if (!defined $pmax and $pop=~/^=$/ and $pmin !~ /\bnull\b/) { 
                # make +/- 1 day
                $pmax = $pmin + 1;
                $pmin -= 1;
                $pop = 'range';
             }
             $pmin  = int($pmin) if ((defined $pmin) && ($type =~ /^int/));
             $pmax  = int($pmax) if ((defined $pmax) && ($type =~ /^int/));

        } elsif ($type =~ /^int|^float/) {
      # Unclear whether one should support white space between
      # sign and number and between mantissa and exponent.
      my $REAL_MATCH=
      '^\s*'      .                   # Initial whitespace
      '([\+\-]?)' .                   # Optional sign
   #   '\s*'       .                  # White space between sign and mantissa
      '((\d+)|(\d+\.\d*)|(\.\d+))' .  # Mantissa
   #   '\s*'       .                  # White space b/w mantissa and exponent
      '('         .                   # Everything after is optional
        '[eEdD]'  .                   # Exponent identifier
        '[\+\-]?' .                   # Optional sign of exponent
        '(\d|\d\d|\d\d\d)'.           # Exponent has 1-3 digits
      ')?'        .                   # Make exponent optional
      '\s*$';                         # Terminal white space

      # check for non int/float characters first to avoid eval error 
      my $badTypeMatch=0;
      if ((length($pmin) && ($pmin !~ /$REAL_MATCH/) && ($pmin !~ /null/i)) ||
          (length($pmax) && ($pmax !~ /$REAL_MATCH/) && ($pmax !~ /null/i))
       ) {
          
          $badTypeMatch = 1;
      } else {

         # evaluate an expression that might be entered
         $pmin =~ s/(.*)/$1/ee;
         $pmax =~ s/(.*)/$1/ee;

         # SYBASE doesn't allow sci notation with 'd' or 'D' as separator.
         $pmin =~ s/[dD]/e/;
         $pmax =~ s/[dD]/e/;
      }

# commented out since we may want to allow arithmetic expressions in link_criteria
      if ((length($pmin) && ($pmin !~ /$REAL_MATCH/) && ($pmin !~ /null/i))
       || (length($pmax) && ($pmax !~ /$REAL_MATCH/) && ($pmax !~ /null/i))
       || $badTypeMatch) {
           #Browse::Utils::Error("Error: Parameter $param is of type $type, ".
            #"but one of the values you entered is not a valid " .
            #"$type. Please go back and enter an appropriate " .
            #"$type value.");
           Browse::Utils::Error("Error: Unable to apply constraint " .
		"to column $param.  This column is numeric but we cannot " .
                "convert '$pmin' to a number.<p> If you are trying to use " .
                "expressions involving other columns in a query you must " .
	 	"enclose the expression in parentheses. <br> E.g., <br>" .
		" > (3*exposure)<br> or <br> < (b_mag-v_mag+2)<p>" .
		"<a href='w3browse-help.html#Parameter_Search'>Help</a>");

      }

    } # end above type checking
  } # end if has parens 

    if ($type =~ /^char/) {

      if ($' > 255) {    #'
         $above255 = 1;
         $pmin .= '%' if(defined $pmin); 
         $pmax .= '%' if(defined $pmax); 
      } 

      #replace * with % wild card character
      $pmin =~ s/\*/\%/g;
      $pmax =~ s/\*/\%/g;
     ##testing quote

     # if ($pmin !~ /\(|^'/) {
     #    $pmin = "'" . $pmin . "'" if(defined $pmin and $unit !~ /mjd/); 
     #    $pmax = "'" . $pmax . "'" if(defined $pmax and $unit !~ /mjd/);
     # } else {
     #    $pmin =~ s/\(/\('/g;
     #    $pmin =~ s/\)/'\)/g;
     #    $pmax =~ s/\(/\('/g;
     #    $pmax =~ s/\)/'\)/g;
     # }
      if (!$above255) {
         $pmin =~ tr/a-z/A-Z/ if ($pmin !~ /\(/ && $param ne "dp_tag");
         $pmax =~ tr/a-z/A-Z/ if ($pmax !~ /\(/ && $param ne "dp_tag");;
      }
    } 
 
    # build query qualifiers.

    if ($type =~ /^class$/) {
        if (length($pmin) > 0) {  # class queries allows multiple classes
            push (@qualifiers, "$param|in|($pmin)");
	}

    } elsif ($pop !~ /range/ && defined $pmin) { # = or != (incl like/not like)

        push(@qualifiers, "$param|$pop|$pmin");

    } elsif ($pop =~ /range/i) {            # range queries

        my $qual;
        my $not;

        $not = "!" if $pop =~ /\!/;
        #print "MakeQuals; range $param|$pop|$pmin<br>" if $::debug;

        if ((defined $pmin) && (defined $pmax)) {
           $qual = "$param|$not"."between|$pmin|$pmax";
        } elsif (defined $pmin) {
           $qual = "$param|$not".">=|$pmin";
        } elsif (defined $pmax) {
           $qual = "$param|$not"."<=|$pmax";
        }
        push(@qualifiers, $qual);
    }

    } # end foreach pval

  return(\@qualifiers);

} #eosub--Parameter::MakeQuals
#------------------------------------------------------------------------
sub FilterTableQuals {
    my($self, $table_name,$qualifiers) = @_;

    my $new_qualifiers = [];
    foreach my $qual (@$qualifiers) {
        #print "Filter $qual<br>" if $::debug;
	my $new_qual = $qual;
	if ($qual =~ /^([a-z0-9_]+)\./i) {
	    next if (lc($table_name) ne lc($1));
	    $new_qual =~ s/^[a-z0-9_]+\.//i;
	}	    
        #print "pushing  $new_qual<br>" if $::debug;
	push(@$new_qualifiers, $new_qual)
    }
    return($new_qualifiers);

} #eosub--Parameter::FilterTableQuals
#------------------------------------------------------------------------
sub MakeConeQuery {

    my ($self,$in) = @_;
    my @positions;
    if ($in->{'Position'}) {
        @positions = split('\0', $in->{'Position'});
	return (\@positions);
    }

    return unless $in->{'Entry'} =~ /\w/;

    #get radius
    $in = Browse::Parse::PrepRadius($in);


    my ($radius) = defined($in->{"Radius"}) ? $in->{"Radius"} : "Default";
    $radius /= 60 unless $radius =~ /Default/i;

    my($coord_sys) = Astro::CoordSys->new('system' => $in->{'Coordinates'},
                                          'equinox' => $in->{'Equinox'});

    my $ncs;
    if ($in->{'Coordinates'} ne 'Equatorial' ||
        $in->{'Equinox'} ne '2000') {
        $ncs = Astro::CoordSys->new('system'    => "Equatorial",
                                            'equinox'  => 2000);
    }
    #$in->{'NR'} = "GRB/NED/SIMBAD";
    $in->{'NR'} = "CheckCaches/GRB/SIMBAD/NED" unless $in->{'NR'};
    my $NR = $in->{'NR'};
    my $use_cache = 1;
    if ($NR !~ /CheckCaches/i) {
        $use_cache = 0;
    }
    $NR =~ s/\/?CheckCaches\/?//i;

    my @res;
    my %seenres;
    foreach my $nr (split('/',$NR)) {
       push(@res,$nr) unless $seenres{$nr}++;
    }
    $Astro::Config::error_handler=\&{sub {}};

    my @resolvers;
    foreach my $res (@res) {
      @Astro::Object::Resolver::CheckCaches::default_caches_to_check=($res);
      my $resolver = Astro::Object::Resolver->new('type' => $res,
                                              'use_cache' => $use_cache);
      push(@resolvers,$resolver) if $resolver;
    }

    my $msg="";
    foreach my $entry ( split(/\s*;\s*/,$in->{'Entry'})) {
        #print "Parameter:: entry=$entry<br>" if $::debug;
        next unless $entry =~ /\S/;
        my @objects;
        foreach my $resolver (@resolvers) {
            my $obj = Astro::Object->new('string'=>$entry,
                                         'coordsys'=>$coord_sys,
                                         'resolver'=>$resolver);
            push(@objects,$obj) unless $obj->error();
            if ( $::debug && $obj->error()) {print "error: " . $obj->error() . "<br>";}
            #--- Add warning for GRB resolver only if entry is GRB
            if ($obj->error() && 
               ($resolver->{name} =~ /GRB/ && $entry =~ /^GRB/)) {
               $msg .= "<br>" if $msg;
               if ($resolver == $resolvers[0]) { 
                  $msg .= "WARNING!"; 
               } else {
                  $msg .= "Comment:";
               }
               $msg .= " <i>$entry</i> is not known to the ".
		  $resolver->name()." resolver.";
            } 
        }
        my $error;
        unless (scalar(@objects)) {
           #$error="$NR <i>$entry</i> is not recognized";
           #&$error_handler($error);
           Browse::Utils::Error("$NR <i>$entry</i> is not recognized by the ".
		join(" or ",@res)." name resolver".($#res>=1?"s":"").".<br>" .
                "If you are trying to enter a coordinate string " .
		"give at least minutes for sexagesimal coordinates or " .
		"use a decimal point in decimal coordinates.  Also separate " .
		"the longitude/RA and latitude/Dec by a comma or the sign " .
		"of the latitude/declination.  <br>E.g., use " .
    		"12.,14. or 12 00 +14 00 rather than 12,14");
        }
        my $object = $objects[0];
        if ($object->resolver_used() and $object->resolver_used() !~ /$res[0]/) {
           $msg .= "<br>" if $msg;
           $msg .= "Using the coordinates from the ".
		$object->resolver_used()." resolver for <i>$entry</i>.";
        }
        foreach my $obj (@objects[1 .. $#objects]) {
           my $angsep = $object->angular_separation($obj);
           my $checkrad = $::radius_tolerance;
           if (($radius !~ /Default/) and ($checkrad < ($radius*60/10))) {
              $checkrad = $radius*60/10; # 10% of radius in arcmin
           }
           if ($angsep > $checkrad) { 
              my $url = "w3query.pl?$$in{input}";
              my $coords = 
		Browse::CGI::Encode(join(',',@{$$obj{"coordinates"}}));
              my $eentry = quotemeta(Browse::CGI::Encode($entry));
              $url =~ s/$eentry/$coords/i;
              $msg .= "<br>" if $msg;
              $msg .= "WARNING! The positions returned by ".
		$object->resolver_used()." and ".  $obj->resolver_used().
		" for $entry differ by ".  sprintf("%.1f",$angsep).
		" arcminutes!<br><a href=\"$url\">" .
		"Requery with coordinates found by ".
		$obj->resolver_used()."</a>";
            }

        }
        my $res = $object->resolver_used();
        #print "Parameter::resolver=$res<br>" if $::debug;

        if ($use_cache) {
           my $cache = $object->resolver_cache_used();

           #print "Parameter::resolver=" .$object->is_name(). 
		#"cache=" . $cache ."<br>" if $::debug;

           # remove directory path, just want name
           $cache =~ s/^(.*\/)+// if $cache;  
           $res = "$cache (local cache)" if $cache;
        }

        $object = $object->convert("coordsys" => $ncs) if $ncs;
        my ($ra,$dec) = @{$$object{"coordinates"}};

        push(@positions, "$ra|$dec|$radius|$entry|$res");

    }
    return(\@positions,$msg);

} #eosub--Parameter::MakeConeQuery
#------------------------------------------------------------------------
sub Search {
    my ($self, $tables, $positions, $qualifiers, $in, $batchflag,$resmsg) = @_;
    
    # Serialize and Encode current query
    my($saveQuery) = FreezeThaw::freeze($tables, $positions, $qualifiers,
	 $in, $batchflag,$resmsg) ;  
    my($encSaveQuery)=Browse::CGI::Encode($saveQuery); 


    my($table,$table_info);
    my $allStr = join(",", @$tables);
    #print "Parameter $allStr<br>" if $::debug;
    #Browse::Utils::DumpRef($tables) if $::debug;


    #--- gather information from all tables
    if ($allStr !~ /\,/) {
        $table_info = {};
        $$table_info{$allStr} = {Browse::HDB::TableInfo($allStr)};
    } else {
        $table_info = Browse::HDB::TableInfo($allStr);
    }

    my $displayer = $self->{displayer};

    print "Parameter::I am a ".ref($displayer)."<br>" if $::debug;

    $$displayer{'delimiter'} = $$in{"delimiter"} if defined $$in{"delimiter"};

    $displayer->setCoordSys($$in{"Coordinates"}, $$in{"Equinox"}) 
	if $$in{"Coordinates"};

    #--- If max row limit is not defined set it to default
    #--- If max row limit is set to "no limit" reset it to 0 for query
    my($maxlimit) = defined($in->{'ResultMax'}) ? $in->{'ResultMax'}:1000;
    if ($maxlimit =~ /no limit/) {$maxlimit = 0;}
    $displayer->MaxRows($maxlimit);

    #print "Parameter:: input=$$in{input}<br>" if $::debug;
    $displayer->Input($$in{"input"});
    # pass on saved query to the displayer
    $displayer->setSavedQuery($encSaveQuery);

    if ($$in{'popupFrom'}) {
       $displayer->setPopupRef( $$in{"popupFrom"}); #lmm
    }

    my $hiddenMissions= $self->hiddenMissions($in,$displayer,$table_info);

    $displayer->pageHeader($$in{"navtrail"});

    #-- Get table sort column and direction of sort
    my($sorts, $dirs) = $self->OrderClause($in);
    #print print "sort = " .@$sorts . "<br>" if $::debug;
    
    #-- Added to use this method for requery. YourSearchWas expects 
    #--- direction as "up" or "down"
    my @searchDir;
    foreach my $dir (@$dirs) {
      my $x = ($dir =~ /1/) ? "up" : "down";
      push(@searchDir, $x);
    }

    $displayer->YourSearchWas($positions,$qualifiers,$sorts,\@searchDir,
	$resmsg,$tables);

    #$displayer->Directions($tables,$table_info);
    $displayer->resultsSummary($table_info,$tables);
    $self->beforeAllTables();
    $displayer->showText($$in{'msg'});
   
    print "Parameter::sorts:<br>@$sorts<br>\n" if ($sorts && $::debug); 
    print "Parameter::dirs:<br>@$dirs<br>\n" if ($dirs && $::debug); 
    print "Parameter::qualifiers:<br>@$qualifiers<br>\n" 
	if ($qualifiers && $::debug); 
    print "Parameter::positions:<br>@$positions<br>\n" 
	if ($positions && $::debug); 

    my ($first) = 1;	# for special 1st table headers
    my @nullTables;
    my @errTables;
    my $nullTablesHtml="";  #save unused tables as hidden params for requery
    my $allTablesCount = 0;

table_loop:
    foreach $table(@$tables) {

        print "----Parameter::table loop looking at table $table\n<br>" if $::debug;


        my $tablename=$displayer->chopTable($$table_info{$table}{table_name});
        my $tabletitle=$tablename;
        if ($tabletitle =~ /\//) {
           #--- remove "out" suffix for title
           $tabletitle =~ s/\/out//;
        }

        # check to see if we shouldn't do the query
        if (($in->{'Time'} or $in->{'bparam__metaparam_start_time'}) and 
               !defined($$table_info{$table}{'start_time'})) {
            push(@errTables, ["Table does not support searches by time:",
                              $table,$$table_info{$table}{"table_description"}]);
            $displayer->beforeNewTable($tablename,$tabletitle,
		$$table_info{$table}{table_description});
            $displayer->afterTable();

            next;
        }

        if ($in->{'Entry'} and 
               !defined($$table_info{$table}{'default_search_radius'}))  {
            push(@errTables, ["Table does not support coordinate searches:",
                              $table,$$table_info{$table}{"table_description"}]);
            #$self->beforeNewTable($table_info,$table);
            $displayer->beforeNewTable($tablename,$tabletitle,
		$$table_info{$table}{table_description});
            $displayer->afterTable();
            next;
        }
        
        #print "Parameter checking parameters<br>" if $::debug;
        my $filterQuals = $self->FilterTableQuals($table,$qualifiers); 
        if (@$filterQuals) {
           #print "in filter loop <br>" if $::debug;
	    my $paraminfo = Browse::HDB::ParamInfo($table);
	    $paraminfo = $$paraminfo{$table};
	    foreach (@$filterQuals) {
               #--- Extract parameter name from parameter expression
	       my ($param, $rest) = split('\|', $_);
               $param  =~ s/^\s+//;
               $param  =~ s/\s+$//;


               print "Parameter:  param $_" . "__*__" ."$param" . 
			"__*__" ."$table<br>" if $::debug;



	       #--- if parameter not defined for current table add
               #--- new error message for Errors section of results page
               #--- and go to next table
	       # Don't check metadata parameters or table links here.
	       if ($param !~ /meta/)  {
               print "Parameter:  link chk $_ <br>" if $::debug;
	       #if ($param !~ /meta/ && 
	           #$_ !~  /^(.*)([\=\>\<]|\s+in\s+|\s+between\s+|\s+not\s+in\s+)(.*)/)   {
                  #--- remove optional table name from parameter
                  if ($param =~ /^([a-z0-9_]+)\./i) {
           	     if (lc($table) eq lc($1)) {
               		$param =~ s/^[a-z0-9_]+\.//i;
                     }
                  }
                  #--- Allow parameter queries on _unique_id (TAM)
	          if (!defined($$paraminfo{$param}) && $param ne "_unique_id") {
	             push(@errTables, 
	             ["Table does not support search by parameter: $param",
	             $table, $$table_info{$table}{table_description}]);
	             next table_loop;
	          }

	       }
	    }
	}


        # resolve metaparams that may be sort vars
        my $sortvar;
        #--- First save original table sort column and direction of sort 
        #--- so changes aren't permanent
        my @currentsorts=@$sorts;
        foreach $sortvar (@currentsorts) {
          $sortvar = $$table_info{$table}{$sortvar} 
		if $$table_info{$table}{$sortvar} =~ /\w/;
        }

        my $request = Browse::QueryRequest->new($table);

        $request->setPositions($positions);
        $request->setQuals($filterQuals);
        $request->setSort(\@currentsorts, $dirs);
        $request->setFields( $self->SelectClause($in) );
        $request->setDefaultRadius(
		$$table_info{$table}{"default_search_radius"});
        $request->setMaxRows($maxlimit);
        

        my $datacount = 0;
        my $tdata     = 0;
        my $result;
        my $meta;
        # Do the request.
        $request->initiate();
        $request->storeState();
        while($result = $request->getResult()) {

            #print "Parameter::type=" . $$result{type} . 
		#"maxl=$maxlimit dcount=$datacount<br>" if $::debug;
            if ($$result{"type"} eq "meta") {

                if (!defined($meta) || $$result{"table"} ne $$meta{"table"}) {
                    $displayer->processMeta($result);
                    $meta = $result;
                    $datacount = 0;
                }

            } elsif (!$maxlimit || $datacount < $maxlimit) {

                my $xn = $$result{"names"};

                # Defer headers until we are sure we have data.
                if ($datacount == 0) {
                    $self->displayLegend($displayer,$first);
                    $first = 0;
                    $displayer->tableHeader();
                }

                $$displayer{"offset_var"} = $request->getOffsetVar();
                $$displayer{"sortcolumn"} = $request->getSortString();
                $displayer->processData($result);
                $datacount += 1;
                $tdata     += 1;
            }
        }   #--- end of rows processing
                #print "Parameter::out" if $::debug;
        my $msg="";
        $msg = "Warning: there may be more rows available: $maxlimit " .
		"rows was maximum rows requested." if $maxlimit == $datacount;

       print "parameter: table row count  $tdata<br>" if $::debug;
       if ($tdata != 0) {
            $allTablesCount += $tdata;
       }

       #--- Record number of rows in log file
       Browse::Log::WriteLog("[". DateTime::asctim()."] $::REMOTE_HOST :: " .
	   "$tablename: $tdata rows", $::sqlLog);
        

        $displayer->closeTable();
        $displayer->tableFooter($msg);
        if ($tdata == 0) {
            $displayer->beforeNewTable($tablename,$tabletitle,
		$$table_info{$table}{table_description});
            $displayer->afterTable();
            push(@nullTables, [$table, 
		$$table_info{$table}{"table_description"}]);

            #--- Keep info about empty tables around for requery
            my $results = $displayer->results();
            if ($results) {
              $nullTablesHtml .= $results->DisplayUnusedLinks($meta);
            }
        }
    } # end of foreach table

    $self->afterAllTables($displayer,$allTablesCount);

    $displayer->setNullTables($nullTablesHtml);
    $displayer->showEmpties(\@nullTables, \@errTables);

    $self->finishUp($hiddenMissions,$displayer);
    #$displayer->pageFooter();
#print "</body></html>\n";

} #eosub--Parameter::Search
#------------------------------------------------------------------------
sub SelectClause {

    my($self, $in) = @_;   # input values

    my($params);

    if ($in->{'Field'}) {
        my $fstr = $in->{'Field'};
        if (index(",", $fstr) > 0) {
            my @flds = split('\0', $fstr);
	    return \@flds;
	} else {
	}
    }

    if (defined($in->{'varon'})) {
        my(@varon)  = split('\0',$in->{'varon'});
        return \@varon;
    }

    if ($in->{'Fields'}) {
        return $in->{'Fields'};
    }

    return undef;

} #eosub--Parameter::SelectClause
#------------------------------------------------------------------------
sub OrderClause {

    my($self, $in) = @_;
    
    my (@sorts, @dirs);
    
    # Note that these have to be single arguments for
    # all sorting, since we have to ensure that
    # the right direction is associated with the right
    # sort field.  03/04 -lmm- need 'defined()' because  
    # $in->{Dirs} could be 0
    if ($in->{'Sorts'}  && defined($in->{'Dirs'})) {
        @sorts = split(",", $in->{'Sorts'});
        @dirs = split(",", $in->{'Dirs'});
        return (\@sorts, \@dirs);
    }

    my ($sortby) = $in->{'sortvar'};
    if (defined($sortby)){
        if ($sortby =~ /^-/) {
            $dirs[0] = -1;
            $sortby = $';      #'
        }

        if ($sortby =~ /\_metaparam\_(\w+)/) {
            $sortby = $1;
        }
        push(@sorts, $sortby);
    }
    return (\@sorts, \@dirs);
 
} #eosub--Parameter::OrderClause

#--- Class methods that can be overridden in subclasses
sub hiddenMissions {
   my($self,$in,$displayer,$table_info) = @_;
}
sub beforeAllTables {
   my ($self) = @_;
}
sub afterAllTables {
   my ($self,$displayer,$allTablesCount) = @_;
   
}
sub finishUp {
   my ($self,$hiddenMissions,$displayer) = @_; 
   $displayer->pageFooter();
}
sub displayLegend {
   #--- display legend only for first table
   my ($self, $displayer, $first) =@_; 
   if ($first == 1) {
      $displayer->legend();
   }
}
#------------------------------------------------------------------------
1;
#########
#  EOF  #
#########
