#$Id: Query.pm,v 1.4 2008/12/15 17:56:08 lmm Exp $
#=======================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name & address intact.
#=======================================================================#
# Module Name:    Query.pm
# Module Author:  Laura McDonald, ADNET
# Date:           November 2006
# Project:        Browse : HEASARC
#=======================================================================#
# Module Description:
# This module contains subroutines for processing Browse queries
# (replaces w3query.pl)
#=======================================================================#

use strict;
use lib './lib';
use Browse::Config;
use Browse::Log;
use Browse::Options;
use Browse::Parse;
use Browse::Parameter;
use Browse::TabbedParameter;
use Browse::ValidateInput;
use DateTime;
use Astro;
use Browse::Sort;
use Browse::Display;


package Browse::Query;

#------------------------------------------------------------------------
#--- Process CGI parameters before processing Brwose quesry
sub preprocess {
   my ($querystring) = @_;
   my($in,$input) = Browse::CGI::ReadParse(1);
   print Browse::CGI::Header,"****Query::preprocess: DEBUG<br>\n" if $::debug;

   process($in,$input);
}

#------------------------------------------------------------------------
#--- Process Browse query
sub process {
  my($in,$input) = @_;
   
  Browse::Config::AlwaysRun();

  print Browse::CGI::Header,"****Query::process: DEBUG<br>\n" if $::debug;
  
  $$in{"input"} = $input if $input;  #is this necessary?

  #check for overloading
  my $nprocs = Browse::Utils::checkprocs("w3query");
  if ($nprocs > $::maxprocesses) {
    my $msg = "Sorry, our system is overloaded with $nprocs simultaneous ".
              "Browse requests at this time.  Please keep trying.";
    if ($input && $input =~ /BATCHRETRIEVALCATALOG/) {
       print "Content-type: text/plain\n\n";
       print "BatchStart\n\nBATCH_RETRIEVAL_MSG ERROR: $msg\nBatchEnd\n";
       exit;
    }
    Browse::Utils::Error($msg);
  }


  # this GETTABLES is set by browse_extract when batch users request table list
  if ($in->{"GETTABLES"}) {
    print "Content-type:text/plain\n\nBatchStart\n";
    print "Available tables:\n";
    Browse::HDB::Display_tables_list("BatchDisplay");
    print "BatchEnd\n";
    exit;
  }

  $in = Browse::Parse::set_defaults($in);
  my ($tablehead, $missions, $tabletypes, $default_tables);
  ($tablehead, $missions, $tabletypes, $in, $default_tables) = 
	Browse::Parse::get_table_list($in);
  

  my @tables=();
  foreach (@{$tablehead}) { 
     push @tables, $_->{'name'}; 
   }

  if ($in->{'Action'} =~ /Parameter/i){  #Build Parameter Search
    print "<br>Building Parameter search<br>\n" if $::debug ;
    BuildParamQuery($in,\@tables,$tablehead);
    Browse::Log::Query($0,$in,"Build Parameter Search",\@tables);

  } else {                   #Search

    #--- Tabbed Display is default if undefined

    if (!$in->{displaymode} && $input && $input !~ /BATCHRETRIEVALCATALOG/) {
       if ($ENV{HTTP_USER_AGENT} =~ /MSIE/) {
          $in->{displaymode} = "Display";
       } else {
          $in->{displaymode} = "TabbedDisplay";
       }
    }
    
    my $parameterMode;
    if ($in->{displaymode} =~ /Tabbed/) {
       $parameterMode='TabbedParameter';
    } else { 
       $parameterMode='Parameter';
    }
    print "<br>w3query Preparing Parameter search " . 
	$in->{displaymode} ."  $parameterMode<br>\n" if $::debug;

    my $parameter = Browse::Parameter->new($parameterMode);
    $parameter->PrepareSearch($in,\@tables);
  }
   print "*** exiting w3query<br>" if $::debug;
  exit;
}# eosub--process
#--------------------------------------------------------------------------
sub BuildParamQuery {

  my($in,$tables,$tablehead) = @_;
  my(@missions)  = split("\0", $in->{'mission'});

  my (@tabletypes) = ("Observation","Object","Proposal","Physics");
  @tabletypes = split("\0", $in->{'table_type'}) if $in->{'table_type'};

  my(@tables)    = @ {$tables};
  my $selected = "selected" if $#tabletypes < 3;

  Browse::Options::Header( [ @missions ], $selected );
  my $formname="w3query";
  print "\n<form method='POST' name=\"$formname\" action=\"w3query.pl\">\n";

  my $navtrail;
  $navtrail = $in->{'navtrail'}; 
  unless ($navtrail) { 
      my $referrer = ($ENV{"HTTP_REFERER"} or "/".$::cgiurl."/w3browse.pl");

      # change " to ' to fix html problem with Catalog index search --lmm
      $navtrail = "<a class='navpast' href='$referrer'> Choose Tables</a>";}

  print "$navtrail &gt; <span class='navcurrent'>Parameter Search</span>" .
	" &gt; <span class='navfuture'>Search Results &gt; " .
	"Choose Data Products</span><br>";

  if ($ENV{REQUEST_METHOD} eq "GET") {
     $navtrail .= " &gt; &lt;a class='navpast' href=\'".$ENV{SCRIPT_NAME}.
	'?'.$ENV{QUERY_STRING}."\'&gt;Parameter Search&lt;/a&gt;";
  } else {
     $navtrail .= " &gt; Parameter Search";
  }
  $navtrail =~ s/\</&lt;/g;
  $navtrail =~ s/\>/&gt;/g;
  print "<input type='hidden' name='navtrail' value=\"$navtrail\">\n";

  # Create Header for Results page lmm
  my($popupRef) = "Query Results";
  print "<input type='hidden' name='popupFrom' value=\"$popupRef\">\n";  



  Browse::ValidateInput::checkParameterQuery($formname,\@tables);
  
  # Special kludge to kick out tables that don't
  # work in parameter query.  Must become metadata based!!!!
  
  foreach (@tables) {
      if ($_ eq "heasarc_usnoa2" || $_ =~ /^Z\//) {
          print "</table>";
          Browse::Utils::Error("Error:  Parameter query not available for table: $_");
	  #Browse::Decor::Footer();
	  exit();
      }
  }
  
  print "<script type=\"text/javascript\" language=\"JavaScript\">\n";
  Browse::ValidateInput::tickTable();
  print "</script>\n";

  print "<br>"; 
  Browse::Options::TableHeadings( "", $tablehead, 0);
  Browse::Options::DisplayTables( $tablehead, $tablehead );

my $num=1;
  print "<p><b>".$num++.". Enter any constraints on the query below. <a target='helpwin' href='$::w3bhelp#build_query'>Help on constraint syntax</a></b><br>\n";
  print "<span class='small'>(What about <a target='helpwin' href=\"$::w3bhelp#wildcards\">wildcards, spaces, and case sensitivity</a>?)</span> <br>";
  print "<b>" . $num++.". To change the fields that are returned, select the box in the 'View' column beside each field desired.<br>\n" unless scalar @tables >1;
  print $num++.". To sort the results by any field, select one box in the 'Sort' column beside the field to sort on.</b>\n";
  print "<select name='dummy'><option>Examples of query constraints:\n";
  print "<option><5 (less than 5)\n";
  print "<option>bar .. foo (between the strings bar and foo)\n";
  print "<option><5; 10 .. 20; >=40 (less than 5 or between 10 and 20 or more than 40)\n";
  print "<option>3c* (string field: string starting with 3c)\n";
  print "<option>10 !.. 15 (not between 10 and 15)\n";
  print "<option>>=1996-01-01 (in a time field: on or after Jan 1 1996)\n";
  print "<option>!=null (not null)\n";
  print "<option>positions can be specified in decimal or sexagesimal\n";
  print "<option>dates can be specified in MJD or YYYY-MM-DD hh:mm:ss\n";
  print "</select></p>\n";

  my($fields) = defined($in->{'Fields'}) ? $in->{'Fields'} : '';
  $fields     = 'standard' if (! $fields);

  my($parameters,$used_info);
  
  if (scalar ( @tables ) > 1) {
    ($used_info,$parameters) = MultipleTables( [ @tables ]);
    
  } else {
    my($table)  = $tables[0];
    print "w3query::table=$table<br>" if $::debug;
    $parameters = Browse::HDB::ParamInfo($table); 
    $used_info  = { % { $parameters->{$table} } };
  }

  if (! defined($used_info) && (scalar( @tables ) > 1)) {
     print "<br></center><b>No common parameters found for selected catalogs.<br>".
           "Please try reducing the number of catalogs selected.</b>";
  } else {
     ParameterHeading( scalar(@tables));
     $used_info = ListParameters( [ @tables ], $used_info, $fields,$in);
  }
  #print "</table></table>\n</table></center>\n";
  print "</table>\n";
  print "<p><b>".$num++.". Do you want to change your current query settings?</b><br>\n";
  Browse::Options::OptionBar($in);
  NewSubmitButtons();

  if(defined($used_info)) {
    &PrintHiddenParams($parameters, $used_info, [ @tables ]);
  }
  print "</form><br>\n";
  #Browse::Decor::Footer;
  Browse::Decor::EmailFooter();
  Browse::Decor::EndTags();



} # eosub--BuildParamQuery 
#--------------------------------------------------------------------------
sub SubmitSection {
  SubmitButtons();

} # eosub--SubmitSection
#--------------------------------------------------------------------------
sub NewSubmitButtons {
  print<<EOT;
     <p><b>5. <span class="large"><input type="submit" name="Action" value="Start Search" onClick="return(pageCheck())">
     <input type="reset" value="Reset"></span> 
EOT
}

sub SubmitButtons {
  print <<EOT;
  <tr>
  <td align='center' colspan='7'> 
     <span class="large"><input type='submit' name="Action" value="Submit" onClick="return(pageCheck())">
     <input type="reset" value="Reset"></span> 

  <script type="text/javascript" language="JavaScript1.2">
<!--
     document.write ("   <input type='checkbox' name='show' checked id='showranges' onClick='showRanges()'> <label for='showranges'>Show Ranges</label>");
// -->
  </script>

  </tr>
EOT

} #eosub SubmitButtons
#--------------------------------------------------------------------------
sub ParameterHeading {
  my($numtables) = @_;

#print "</td><tr><td>\n<table border='1'>\n";
print "<table class='othertable'>\n<tr class='othertablehead'>\n";

if ($numtables < 2) {
  print <<EOT;
  <th class='othertable' scope='col'><label for="varon"><a target='helpwin' href="$::w3bhelp#view"><b>View</b></a></label><br>
  <script type="text/javascript" language="javascript">
       <!-- 
       document.write("<input type='checkbox' name='Fieldson' value='All' id='all' onClick='tickVars(&quot;varon&quot;,this.checked);tickVars(&quot;Fields&quot;,this.checked)'> <label for='all'>All</label>")
       // -->
  </script>
</th> 
EOT
}

  print <<EOT2;
  <th class='othertable' scope='col'><label for="sortvar"><a target='helpwin' href="$::w3bhelp#sort"><b>Sort</b></a></label></th>
  <th class='othertable' scope='col'><a target='helpwin' href="$::w3bhelp#param"><b>Parameter (Unit)</b></a></th>
  <th class='othertable' scope='col'><label for="param"><a target='helpwin' href="$::w3bhelp#build_query"><b>Query Terms</b></a></label></th>
  <th class='othertable' scope='col'><b>Min Value</b></th>
  <th class='othertable' scope='col'><b>Max Value</b></th>
  <th class='othertable' scope='col'><b>Value Type</b></th>
  </tr>
EOT2

} # eosub--ParameterHeading
#--------------------------------------------------------------------------
sub ListParameters {
  my($tables,$param_info,$fields,$in) = @_; 

  my(@tables) = @{ $tables };
  my($param_html,$table);
  my($mtables) = 0;

  $table = $tables[0];
  my($table_info) = { Browse::HDB::TableInfo($table) };
 
  if (scalar(@tables) > 1) {
     $table_info->{'table_name'} = join(',', @tables);
     $table_info->{'table_document_url'} = '';
     $mtables = 1;
  }

  my($parameter);

  my($default) = join(', ',@{ $param_info->{'#default#'} });
  
  foreach $parameter(grep (!/^\#/, sort keys %{$param_info} )) {
  
    #print "Table Parameter:: working on param: $parameter <br>" if ($::debug);
    
    ($param_info->{$parameter},$param_html->{($parameter)}) = 
                 BuildParameter($parameter,$table_info,$mtables,
                                 $param_info->{ $parameter },$default,
                                 $fields, $in);
  }
  
  PrintParameters2($param_info->{"#all#"},$param_html,$table_info,$default);
  
  return($param_info);
  
} #eosub--ListParameters
#--------------------------------------------------------------------------
sub MultipleTables {

  my($tables) = @_;

  my($tablist)    = join(',',@{$tables});
  my($parameters) = Browse::HDB::ParamInfo($tablist); 

  my($first_table) = shift( @{$tables} );
  my($good_info)   = { % { $parameters->{$first_table} } };
  my(@all_arr)     = @ { $good_info->{'#all#'} };
  my($good_list)   = join(', ',@{ $good_info->{'#all#'} });

  my($param,$table);
  foreach $param (@all_arr) {
  
    foreach $table(@ {$tables} ) {
    
      my ($param_info) = { % { $parameters->{$table} } };
      my $all = $$param_info{"#all#"};
      my ($newtable_list) = join(', ',@{ $param_info->{'#all#'} });
      if ($newtable_list !~ /\s$param\,/ && $newtable_list !~ /\s$param$/
                && $newtable_list !~ /^$param\,/) { 
		
        $good_list =~ s/(\s|^)$param(\,|$)//;  
        $good_list =~ s/\,*\s\,/\,/g;
        last; # stop checking, go to the next parameter 
	
      } elsif ($param_info->{$param}{'parameter_unit'} ne 
                  $good_info->{$param}{'parameter_unit'}){
        $good_list =~ s/$param//;  
        $good_list =~ s/\,*\s\,/\,/g;
	
      } else {
        my($min) = $param_info->{$param}{'parameter_minval'};
        my($max) = $param_info->{$param}{'parameter_maxval'};
        $good_info->{$param}{'parameter_minval'} = $min
                    if(lc($good_info->{$param}{'parameter_minval'})> lc($min));
        $good_info->{$param}{'parameter_maxval'} = $max
                    if(lc($good_info->{$param}{'parameter_maxval'})< lc($max));
      } 
    }
  }

  my($joint_param_info) = {}; 
  ($joint_param_info) = &MultTabMetabaseParams($tablist,$parameters); 
my @joint_good_list = keys %$joint_param_info;

  $good_list =~ s/\s*\,\s*$//;  #chopping off the last comma
  $good_list =~ s/^\s*\,\s*//;  #chopping off the first comma
  $good_list =~ s/\s+//g;

  my(@good_list)  = split(',',$good_list);
  if (scalar(@good_list) < 1) {
    return (undef $joint_param_info);
  }

  foreach(@good_list) {
    $joint_param_info->{$_} = $good_info->{$_};
  }
  $joint_param_info->{'#default#'} = [ @joint_good_list, @good_list ];
  $joint_param_info->{'#all#'} = [ @joint_good_list, @good_list ];

  return($joint_param_info,$parameters);
} #eosub--MultipleTables 
#--------------------------------------------------------------------------
sub MultTabMetabaseParams {
  my($tablist,$parameters) = @_;

  my(@tables) = split(',',$tablist);
  my($table,$masterlist);
  my($alltable_info,$joint_param_info) = {};
  my($good_info)   = { % { $parameters->{$tables[0]} } };
  $alltable_info =  Browse::HDB::TableInfo($tablist); 

  my($metaparam) = {};
  my(@checkparms) = ('right_ascension', 'declination', 'start_time', 
                     'stop_time', 'target_name');

  my($cparam);
  foreach $cparam(@checkparms) {
     my($param_value);
     my($ok) = 1;
     my($diffname) = 0;
     foreach $table(@tables) {
       my($param_info) = { % { $parameters->{$table} } };
       my($params) = join(', ',keys %{$alltable_info->{$table}});
       if ($params =~ /$cparam/ && $ok) {
          my($pname)  =  $alltable_info->{$table}{$cparam};
          $param_value = $pname if (!$param_value); 
          if ( $pname =~ /^[0-9].*/) {
              $ok = 0;  # no param column name present 
              last;
          }
          if ($pname ne $param_value && (!$diffname)) {
              $diffname = 1;
          }

          my($min) = $param_info->{$pname}{'parameter_minval'};
          my($max) = $param_info->{$pname}{'parameter_maxval'};
          my($unit)= $param_info->{$pname}{'parameter_unit'};
          my($type)= $param_info->{$pname}{'parameter_format'};

          if(! defined($good_info->{$cparam}{'parameter_unit'})) {
              $good_info->{$cparam}{'parameter_unit'} = $unit; 
          } 
          if($good_info->{$cparam}{'parameter_unit'} ne $unit) {
              $ok = 0;  # units not same for parameters
              last;
          }
          if(! defined($good_info->{$cparam}{'parameter_format'})) {
              $good_info->{$cparam}{'parameter_format'} = $type; 
          } 
          if(! defined($good_info->{$cparam}{'parameter_minval'})) {
              $good_info->{$cparam}{'parameter_minval'} = $min; 
          } 
          if(! defined($good_info->{$cparam}{'parameter_maxval'})) {
              $good_info->{$cparam}{'parameter_maxval'} = $max; 
          } 
          $good_info->{$cparam}{'parameter_minval'} = $min
                     if(lc($good_info->{$cparam}{'parameter_minval'})> lc($min));
          $good_info->{$cparam}{'parameter_maxval'} = $max
                     if(lc($good_info->{$cparam}{'parameter_maxval'})< lc($max));
          $ok = 1;
       } else {
          $ok = 0;
          last;
       }
     }
     if ($ok && $diffname){
        $masterlist .= '_metaparam_' .$cparam . ','; 
        $joint_param_info->{'_metaparam_'.$cparam} = $good_info->{$cparam};
     } 
  }

  return($joint_param_info);

} #eosub--MultTabMetabaseParams
#--------------------------------------------------------------------------
sub BuildParameter {
  my($name,$table_info,$mtables,$param_info,$default_param,$fields,$in) = @_;

  my($above255)= 0;

  my($type)    = $param_info->{'parameter_format'};
  my($accuracy) = $type =~ /:([\d.\w]*)/;
  my($unit)    = $param_info->{'parameter_unit'} if $param_info->{'parameter_unit'};
  my($default) = $param_info->{'parameter_default'};
  my($min)     = $param_info->{'parameter_minval'};
  my($max)     = $param_info->{'parameter_maxval'};
  $min         = $$in{"w3min_$name"} if defined $$in{"w3min_$name"};
  $max         = $$in{"w3max_$name"} if defined $$in{"w3max_$name"};
  my($class)   = ($name =~ /^class$/i) ? 1 : 0;

  #-- check for other virtual ras and decs
  my ($oras, $odecs) = Browse::Display::otherRaDecs($table_info);

  $default = 1 if ($in->{'varon'} =~ /$name/i || $in->{'sortvar'} =~ /$name/i);

  my($url)     = (defined($table_info->{'table_document_url'})) ?
                 $table_info->{'table_document_url'} : '';

  my($disp_name) = $name;
  if( $disp_name =~ /\_metaparam\_/) { #italicize metabase common params
    $disp_name = "<i>" . $` . $' . "</i>";
  };


  my($urline)  = ($url ne '') ? 
                "<a target='helpwin' href=\"$url#$name\">$disp_name</a>" :
                "$disp_name";
  $urline .= " ($type) ($name) " if $::debug;   

  if ($in->{"w3op_$name"} =~ /=/) {
     $max = "*" x 10;
  }

  my ($def_check);
  if ($default or ($fields =~ /all/i)) {
      $def_check = " checked ";
  }

  my($new_min,$new_max,$html);
  if ($class) {
     return if ($min eq $max); # if class min=max dont display info!

     my($class_info)= Browse::HDB::GetClassinfo($table_info->{'table_name'},$name);
     #--- create list of class numbers and associated class names
     my %classes; 
     foreach ( @{$class_info} ){
         $classes{$_} = $HDBclass::name{$_};
     }

     $html = "<tr class='othertable'>";
     $html.= "<td align='right'><input type='checkbox' name='varon' id='varon' value='$name' $def_check ></td>\n" 
             if(!$mtables);
     $html.= "<td align='left'><input type='radio' name='sortvar' id='sortvar' value='$name'></td>\n<td>$urline</td>\n".
             "<td colspan='4'>\n".
             "<select name='bparam_class' id='param' multiple size='4'";
     $html .= " onChange='if (this.selectedIndex > -1) {tickTablestrict(this.form.varon,1,\"$name\")}'>\n";

     #foreach ( @{$class_info} ){
       #$html .= "  <option value=\"$_\">" . $HDBclass::name{$_} . "\n";
     #}
     #--- Sort class list by name and display as select list
     my @classSorted = sort { $classes{$a} cmp $classes{$b} } keys %classes;
     foreach  (@classSorted) {
       $html .= "  <option value=\"$_\">" . $classes{$_}  . "\n";
     }
     $html .= "</select></td></tr>\n";
     $min = $max = '';
     
  } elsif ($unit =~ /^mjd/i) {
     $type =~ /float\d*:\d*\.(\d+)f/;
     $new_min = DateTime::mjd2iso($min, $1);
     $new_max = DateTime::mjd2iso($max, $1);
     #$new_min = DateTime::mjd2iso($min);
     #$new_max = DateTime::mjd2iso($max);
     $new_min =~ s/UTC//g;
     $new_max =~ s/UTC//g;
     $min = $new_min unless (!defined($min) || $new_min =~ /error/i);
     $max = $new_max unless (!defined($max) || $new_max =~ /error/i);
     
  } elsif (grep /$name/i, @$oras) {
  
     $type =~ /float\d*:\d*\.(\d+)f/;
     $new_min = Astro::CoordConv::RA_Decimal2HMS($min,$1);
     $new_max = Astro::CoordConv::RA_Decimal2HMS($max,$1);
     #print "ORAS $name :: min, max=$min,$max->$new_min, $new_max acc=$1<br>" 
	#if $::debug;
     $min = $new_min unless (!defined($min) || $new_min =~ /error/i);
     $max = $new_max unless (!defined($max) || $new_max =~ /error/i);
     
  } elsif (grep /$name/i, @$odecs) {
  
     $type =~ /float\d*:\d*\.(\d+)f/;
     $new_min = Astro::CoordConv::Dec_Decimal2DMS($min,$1);
     $new_max = Astro::CoordConv::Dec_Decimal2DMS($max,$1);
     $min = $new_min unless (!defined($min) || $new_min =~ /error/i);
     $max = $new_max unless (!defined($max) || $new_max =~ /error/i);
     
  } elsif (lc($name) eq $table_info->{'right_ascension'}) {
  
     $type =~ /float\d*:\d*\.(\d+)f/;
     $new_min = Astro::CoordConv::RA_Decimal2HMS($min,$1);
     $new_max = Astro::CoordConv::RA_Decimal2HMS($max,$1);
     print " $name :: min, max=$min,$max->$new_min, $new_max acc=$1<br>" if $::debug;
     $min = $new_min unless (!defined($min) || $new_min =~ /error/i);
     $max = $new_max unless (!defined($max) || $new_max =~ /error/i);
     
  } elsif (lc($name) eq $table_info->{'declination'}) {
  
     $type =~ /float\d*:\d*\.(\d+)f/;
     $new_min = Astro::CoordConv::Dec_Decimal2DMS($min,$1);
     $new_max = Astro::CoordConv::Dec_Decimal2DMS($max,$1);
     $min = $new_min unless (!defined($min) || $new_min =~ /error/i);
     $max = $new_max unless (!defined($max) || $new_max =~ /error/i);
     
  } elsif ($type =~ /float/i and $max !~ /\*\*\*\*/) {
     my $mm='';
     foreach $mm ($min,$max) {
       #if ($mm == 0) {
           #$mm = "0";
       if ($accuracy) {
           $mm = sprintf("%-".$accuracy,$mm);
       } else {
           my $f='';
           ((abs($mm) > 1000000) || (abs($mm) < 0.0001)) ? $f = "e" : $f="f";
           $mm = sprintf("%-10.4".$f,$mm);
       }
     }

  } elsif ($type =~ /char/i) {
     $above255 = 1 if ($' > 255);         #'Editor match
     if($min =~ /^\s+$/) { $min = ''; }
     if($max =~ /^\s+$/) { $max = ''; }
  } elsif ($type=~/int/) {
     my ($length) = $type =~ /:(\d+)/;
     $min = sprintf("%-".$length."d",$min) if $length;
     $max = sprintf("%-".$length."d",$max) if $length;
  }

  # prevents min and max =0 on Vizier param forms
  if (($min==0) and ($max==0) and ($table_info->{"table_name"} =~ /\//)) {
     $min = $max = '';
  }

  $param_info->{'parameter_minval'} = $min;
  $param_info->{'parameter_maxval'} = $max;
  $min =~ s/\"//g;
  $max =~ s/\"//g;

  my($w3op)      = "w3op_$name";
  my($onChange)  = "onChange=\"setMax('$max','$w3op')\"";
  #my ($def_check);

  #if ($default or ($fields =~ /all/i)) {
      #$def_check = " checked ";
  #}

  my $checkedifsort = " checked " if $in->{'sortvar'} =~ /$name/i;


  if (!$html){  #if $html is not set elsewhere (eg class)
     $html = "<tr class='othertable'>";
     $html.= "<td align='right'><input type='checkbox' name='varon' id='varon' value='$name' $def_check></td>\n " if(!$mtables);

     my($range) = $above255 ? '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;' : 'between';  
     my($sort)  = $above255 ? '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;' : "<input type='radio' id='sortvar' name='sortvar' value='$name' $checkedifsort>";  

     $html.= "<td align='left'>$sort</td>\n<td>$urline ";
     my $ounit = $unit;
     $unit =~ s/_/ /g;
     $unit =~ s/micro/&micro;/g;

     #--- Define parameter page display type
     my $disptype='';
     $disptype = "float" if $type =~ /float/;
     $disptype = "integer" if $type =~ /int/;
     $disptype = "string" if $type =~ /char/;
     $disptype = "position" if $name =~ /^$table_info->{right_ascension}$|^$table_info->{declination}$/;
     $disptype = "position" if (grep /$name/i, @$oras); 
     $disptype = "position" if (grep /$name/i, @$odecs);  
     $disptype = "date" if $unit =~ /mjd/;
     
     #display unit  for some parameters
     $html .= " ($unit)" if ($unit =~ /\S/ && $unit !~ /^mjd/i && !($unit =~ /degree/i && ($name =~ /^$table_info->{right_ascension}$|^$table_info->{declination}$/ ) || grep /$name/i,@$oras || grep /$name/i, @$odecs));
     $html .= "</td> <td>\n";

     $html .= "<input name='bparam_${name}' size='30' id='param' onChange='if (this.value) {tickTablestrict(this.form.varon,1,\"$name\")}'>\n";
     $html .= "<input type='hidden' name='bparam_${name}::unit' value='$ounit'>\n" if $ounit; 
     $html .= "<input type='hidden' name='bparam_${name}::format' value='$type'>\n" if $type;
     $html .= "</td>\n<td>$min</td>\n<td>$max</td>\n<td>$disptype</td>\n</tr>\n";
     #<td><input name="bparam_${name}::${type}::${ounit}" size='30' id="param"></td>
##      $html .= "</td><td><input name=$w3op type=text size=30></td>";
##      $html .= "<td>$min</td><td>$max<//td></tr>";
  }
  return ($param_info,$html);

} #eosub--BuildParameter
#--------------------------------------------------------------------------
sub PrintParameters {

  my($param_html,$table_info,$default) = @_;

  my($ra)   = defined($table_info->{'right_ascension'}) ? 
              $table_info->{'right_ascension'} : '';
  my($dec)  = defined($table_info->{'declination'}) ?
              $table_info->{'declination'} : '' ;
  my($name) = defined($table_info->{'target_name'}) ?
              $table_info->{'target_name'} : '' ;

  # ordering display:
  print $param_html->{$name} if (defined($param_html->{$name}));
  print $param_html->{$ra}   if (defined($param_html->{$ra}));
  print $param_html->{$dec}  if (defined($param_html->{$dec}));
  
  undef $param_html->{$name};
  undef $param_html->{$ra};
  undef $param_html->{$dec};

  if (defined($param_html->{'lii'}) && defined($param_html->{'bii'})){
     print $param_html->{'lii'};
     print $param_html->{'bii'};
     undef $param_html->{'lii'};
     undef $param_html->{'bii'};
  }     
  
  if (defined($param_html-> {'ra_1950'})&& defined($param_html->{'dec_1950'})){
  
     print $param_html->{'ra_1950'};
     print $param_html->{'dec_1950'};
     undef $param_html->{'ra_1950'};
     undef $param_html->{'dec_1950'};
  }     

  my(@default) = split (', ',$default);
  foreach(@default) {
    print $param_html->{$_} if ($param_html->{$_} ne '');
    undef $param_html->{$_};
  }

  foreach(sort keys %{$param_html}) {
    next if (/^class$/);
    print $param_html->{$_};
  }

  print $param_html->{'class'}  if (defined($param_html->{'class'}));
  undef $param_html->{'class'};

} #eosub--PrintParameters
#--------------------------------------------------------------------------
sub PrintParameters2 {
  my($order,$param_html,$table_info,$default) = @_;

  my($ra)   = defined($table_info->{'right_ascension'}) ?
              $table_info->{'right_ascension'} : '';
  my($dec)  = defined($table_info->{'declination'}) ?
              $table_info->{'declination'} : '' ;
  my($name) = defined($table_info->{'target_name'}) ?
              $table_info->{'target_name'} : '' ;

  foreach (@$order) {
     print $param_html->{$_};
  }

} #eosub--PrintParameters2
#--------------------------------------------------------------------------
sub PrintHiddenParams {
  my($parameters,$used_info,$tables) = @_;

  my($param);
  foreach $param(grep (!/^\#/, sort keys %{$used_info} )) {
    my($min)   = $used_info->{$param}{'parameter_minval'};
    my($max)   = $used_info->{$param}{'parameter_maxval'};
    my($unit)  = $used_info->{$param}{'parameter_unit'};
    my($type)  = $used_info->{$param}{'parameter_format'};
    $min =~ s/\"//g;
    $max =~ s/\"//g;

    my($stuff) = "name=$param&min=$min&max=$max&unit=$unit&type=$type";
    #$stuff     = Browse::CGI::Encode($stuff);
    print "<input type=hidden name=\"w3xparam\" value=\"$stuff\">\n";
  }

  my(@tables) = @ {$tables};
  foreach (@tables) {
    if($_ !~ /_/){
       $_ =  $::local_origin . $_;
    }

    print "<input type=hidden name=\"table\" value=\"$_\">\n"; 
  }

} #eosub--PrintHiddenParams
1;
#--------------------------------------------------------------------------
#####
#EOF#
#####
#$Log: Query.pm,v $
#Revision 1.4  2008/12/15 17:56:08  lmm
#Added check for multiple coordinates in query.  If they exist they are
#  treated like ra, dec (convert to sexagesmal, display as "position", etc)
#
#Revision 1.3  2007/03/26 19:31:19  lmm
#Added accuracy parameter to calls to Astro::CoordConv::RA_Decimal2HMS()
#  and Astro::CoordConv::Dec_Decimal2HMS()
#
#Revision 1.2  2007/02/16 14:15:37  lmm
#Added new accuracy argument to revised function mjd2iso().
#
#Revision 1.1  2006/11/29 15:50:47  lmm
#initial commit
#
