# $Id: TableFormat.pm,v 1.23 2008/12/15 19:10:36 lmm Exp $
# $Log: TableFormat.pm,v $
# Revision 1.23  2008/12/15 19:10:36  lmm
# Made changes to accommodate multiple pairs of coordinates in heasarc tables.
#
# Revision 1.22  2007/04/02 18:51:06  lmm
# Added back the check for null value before formatting dates.
#
# Revision 1.21  2007/03/26 19:38:55  lmm
# Added ra and dec accuracy parameters to call to
#   Astro::CoordConv::Equ2Sexagesimal()
#
# Revision 1.20  2007/02/16 14:17:56  lmm
# Added new accuracy argument to revised function mjd2iso().
#
# Revision 1.19  2006/06/23 18:46:30  lmm
# Added changes to propogate precision information
#   (code provided by T. McGlynn)
#
# Revision 1.18  2004/12/21 14:13:20  lmm
# Changed TransformCoords() to check if ra and dec are defined
#   instead of checking for non-zero (coordinates may be 0,0)
#
# Revision 1.17  2003/08/04 21:11:00  cwh
# Fixed class column selection so it won't pick up colnames that just start
# with 'class'
#
# Revision 1.16  2002/09/13 20:52:12  cwh
# removed extra 'use lib'.
#
# Revision 1.15  2002/04/02 23:42:39  cwh
# Fixed bug introduced in last commit where B1950 coords were being returned in
#   decimal degrees instead of sexagesimal.
#
# Revision 1.14  2002/03/25 18:16:47  cwh
# Added default accuracy for galactic format.
#
# Revision 1.13  2002/03/17 19:09:38  cwh
# Added use Astro::Utils::SLAConvert;
# Receiving accuracy info from RowDisplay
#
# Revision 1.12  2002/02/12 20:31:58  cwh
# Return from fromJ2000() without doing anything if ra, dec don't exist.
#   Otherwise you'll get '00 00 00.0, 00 00 00.0' as the ra and dec when they
#   should be null.
#
# Revision 1.11  2001/11/07 23:34:20  cwh
# Moved everything except DBD/ into lib/Browse/, so as not to pollute the module
# namespace with our Sort, Search, CGI, etc.  This required putting Browse:: in
# front of every module call and package declaration.  Also any mod::new() had
# to be changed to mod->new().
#
# Revision 1.10  2001/08/28 19:35:09  cwh
# Now using Astro instead of CoordConv for coord transformations.
# Passing accuracy to Astro::CoordConv::Equ2Sexagesimal so we get ra and dec
#   reported with the correct accuracy.
# Instead of printing the long error that Astro.pm returns if the coords cannot
#   be transformed, now we are printing the original ra/dec, along with an
#   error flag.
# Now using accuracy in non-sexagesimal coord formats.
#
# Revision 1.9  2001/07/17 18:42:02  cwh
# Added if defined $$line[$tcol] so that if a field value is undefined, it
# won't be formatted by its formatting info to something like 0.0000.
#
# Revision 1.8  2001/07/01 01:42:45  cwh
# Just got rid of some debug stuff.
#
# Revision 1.7  2001/04/25 17:18:03  cwh
# Completely rewritten by Tom.  Much more modularized.
#
# Revision 1.6  2000/03/08 20:13:02  saima
# W3Browse to Browse name change.
#
# Revision 1.5  1999/10/28 17:42:08  saima
# Optimized so that the %HDBclass::name hash is used to convert class
#   ids to class names. The database is no longer called to obtain the
#   information.
#
# Revision 1.4  1999/10/07 20:16:31  saima
# Fixed a bug where if date was null, 'error' was being returned by the
#   DateTime::mjd2iso() converter.
#
# Revision 1.3  1999/09/24 18:50:59  saima
# Fixed bug in which ra and dec's were being truncated if they were in
#   lower precistion (e.g. usnoa-2, abell catalogs)
#
# Revision 1.2  1999/09/03 20:41:46  saima
# Some code cosmetic changes. Also, if the sybase type=7 (real),
#   the numbers are displayed in lower precision.
#
# Revision 1.1  1999/05/27 17:20:52  tam
# Initial release.
#
#=======================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=======================================================================#
# Package Name:   TableFormat.pm
# Package Author: Tom McGlynn, USRA STX
# Date:           09/99
# Project:        Browse : HEASARC
#=======================================================================#
# Module Description:
# Perform required transformation on data prior to ouput.
#
#=======================================================================#

package Browse::TableFormat;

use strict;

use Astro;
use Astro::Utils::SLAConvert;
use SLA;
use HDBclass;

my $dtor = 3.1415926535/180;

#-----------------------------------------------------------------------
#
# This routine transforms an entire table of results.
#
sub transformTable {

    my ($results, 	# Results of query
        $names, 	# Names of columns
	$types,         # Data types.
	$precisions,    # Precisions/lengths of columns  long strings tam
        $radecs, 	# RA,Dec pairs
	$mjds,		# MJD columns
	$truncs,	# Columns to reformat
	$coordsys, 	# Coordinate System
	$equinox,	# Equinox
	$dateformat,	# Desired date formats
	$coordformat	# Format desired for coordinates.
	
       ) = @_;
       
    #my  ($transHash) = setTransform($names, $types, $radecs, $mjds, $truncs, $coordsys,$equinox,$dateformat,$coordformat);
    # long strings tam
    my  ($transHash) = setTransform($names, $types, $precisions, $radecs, 
	$mjds, $truncs, $coordsys,$equinox,$dateformat,$coordformat);

    
    my $line;
    foreach $line(@$results) {
         transformLine($line, $transHash);
    }
}


#-----------------------------------------------------------------------
# This routine determines what transformations are appropriate
# for a given set of results.
# Returns a hash of transformation parameters.
#
sub setTransform {

    my ($names, 	# Names of columns
        $types, 	# Types of columns
	$precisions,    # Precisions/lengths of columns  long strings tam
	$radecs, 	# Array giving ra,dec column names
	$mjds, 		# Array of column names in MJD
	$truncs, 	# Array of columns with special formatting
	$coordsys,	# Coordinate system for output (input assumed J2000
	$equinox,	# equinox
	$dateformat,	# 
	$coordformat,	# 
	$accuracy	# accuracy of ra/dec
       )
       = @_;
    my ($racols, $deccols)    = getRaDecCols($radecs, $names);
    my ($mjdcols)             = getColNums($mjds, $names);
    my ($classcols)           = getClassCols($names);
    my ($trunccols, $formats, $mjdformats,$coordformats) = 
	getTruncs($truncs, $names, $racols, $deccols, $mjdcols);
    #my ($longcols)            = getLongs($types);
    my ($longcols)            = getLongs($types, $precisions); #long string tam
    my ($paircols)            = getPairs($names);
    return (
        { "names"       => $names,
	  "types"       => $types,
	  "precisions"  => $precisions,  #long strings tam
	  "coordsys"    => $coordsys,
	  "racols"      => $racols,
	  "deccols"     => $deccols,
	  "mjds"        => $mjdcols,
	  "mjdformats"  => $mjdformats,
	  "classes"     => $classcols,
	  "truncs"      => $trunccols,
	  "formats"     => $formats,
	  "longs"       => $longcols,
	  "pairs"       => $paircols,
          "equinox"     => $equinox,
          "dateformat"  => $dateformat,
          "coordformat" => $coordformat,
          "accuracy"    => $accuracy,
          "coordformats" => $coordformats
          #"raaccuracy"    => $raaccuracy,
          #"decaccuracy"    => $decaccuracy
	 });
}

#-----------------------------------------------------------------------
# Get the Ra and Dec column numbers given pairs of RA,Dec column names.
sub getRaDecCols {

    my ($radecs, $names) = @_;
    
    ## Coordinate conversions.
    # There may be multiple columns of RA/Dec pairs.
    
    
    my ($radec);
    my (@racols, @deccols);
    foreach $radec (@$radecs) {
        my ($ra,$dec) = split(",", $radec);
	my ($racol, $deccol);
	my $i;

	
        for ($i=0; $i <= $#$names; $i += 1) {
    
            if ($$names[$i] eq $ra) {
	        $racol = $i;
                #print "TF: $i, $radec, $ra, $dec, $$names[$i] <br>\n";
	    
	    } elsif ($$names[$i] eq $dec) {
	        $deccol = $i;
	    }
	}
	
	
	if (defined($racol) && defined($deccol)) {
	    push(@racols, $racol);
	    push(@deccols, $deccol);
	}
    }
    return (\@racols, \@deccols);
}

#-----------------------------------------------------------------------
# Get the column numbers given a list of column names.
sub getColNums {
     
    my ($list, $names) = @_;
    
    
    my $name;
    my $i;
    my @cols;
    
    foreach $name(@$list) {
    
	for ($i=0; $i <= $#$names; $i += 1) {
	    if ($name eq $$names[$i]) {
		push(@cols, $i);
		last;
	    }
	}
    }
    return \@cols;
}



#-----------------------------------------------------------------------
# Get any column that appears (by name) to be a 
# class column.  We need to handle multiple
# fields for cross-correlations.

sub getClassCols {
    
     my ($names) = @_;
     
     my $i;
     my @class;
     
     for ($i=0; $i <= $#$names; $i += 1) {
     
         if ($$names[$i] =~ /^(.*\.)?class$/) {
	     push(@class, $i);
	 }
     }
     
     return (\@class);
}
     
  
#-----------------------------------------------------------------------
# Column reformatting and truncation. Columns that are formatted elsewhere
# are ignored.  MJD get special hash for format storage
sub getTruncs {

    my ($truncs, $names, $racols, $deccols, $mjdcols) = @_;

    my $trunc;
    my $j;
    my @trunccols;
    my @formats;
    my %mjdformats;
    my %coordformats;
    #my ($raaccuracy, $decaccuracy);
    
    foreach $trunc (@$truncs) {
    
        my ($col, $format) = split(",", $trunc);
	if (!defined($format) ) {
	    $format = "%.3f";
	}
	
	for ($j=0; $j <= $#$names; $j += 1) {
	
	    if ($$names[$j] eq $col) {
	    
	        # Don't truncate RA/Dec
	        if (grep($_ == $j, @$racols)) {
                    #$raaccuracy=$format;
                     #print " TF::getTruncs RA $trunc: adding $j and " .
			#"$format to coordformats<br>" if $::debug;
                    $coordformats{"$j"}=$format;
		    next;
		}
	
	        if (grep($_ == $j, @$deccols)) {
                    #$decaccuracy=$format;
                     #print " TF::getTruncs DEC $trunc: adding $j and " .
			#"$format to coordformats<br>" if $::debug;
                    $coordformats{"$j"}=$format;
		    next;
		}
		
	        if (grep($_ == $j, @$mjdcols)) {
                    #--- key is the mjd column value is the date format 
                    $mjdformats{"$j"}=$format;
		    next;
		}
		
	    
	        push(@trunccols,  $j);
	        push(@formats, $format);
	    
	    }
	}
    }
    #return (\@trunccols, \@formats, \%mjdformats, $raaccuracy, $decaccuracy);
    return (\@trunccols, \@formats, \%mjdformats, \%coordformats);
}

#-----------------------------------------------------------------------
# Get long strings.  We use the data type value -1 to
# indicate long strings.  This seems to work
# with Sybase, and we set it ourselves for VizieR requests.
# However it may not work with other DBI compatible databases.

sub getLongs {

    #my ($types) = @_;
    my ($types, $precisions) = @_;  #long strings tam
    
    my @longs;
    my $i;
    
    for ($i = 0; $i <= $#$types; $i += 1) {
        # Not sure if -1 is hardwired for DBI or just what we
	# get in Sybase DBD.
        #if ($$types[$i] == -1) {
	    #push(@longs, $i);
	#}
	# Also treat very long character strings in this way
	# since Sybase now allows much longer char's.
        if ( $$types[$i] == -1  || 
	    ($$types[$i] == 1 && defined($precisions) && $$precisions[$i] > 255)
	   ) {
	    push(@longs, $i);
	}
    }
    return (\@longs);
}

#-----------------------------------------------------------------------
# Find column pairs of the form measurement, measurement_error
# The error column must be returned immediately after
# the measurement column
sub getPairs {

    my ($names) = @_;

    my $i;

    my @pairs;
    
    for ($i=0; $i <= $#$names-1; $i += 1) {
        if ($$names[$i+1] eq $$names[$i]."_error") {
	    push(@pairs, $i);
	    $i += 1;
	}
    }
    return (\@pairs);
}
    
#-----------------------------------------------------------------------
# This routine actually performs the transformations on the line.
# It checks to see if the transformation is needed and if
# so it calls the appropriate function.
#
sub transformLine {
    my ($line, $transHash) = @_;

    # Set up header
    if ($$transHash{"racols"}) {
        transformCoords($line, $$transHash{"racols"},   $$transHash{"deccols"},
                               $$transHash{"coordsys"}, $$transHash{"equinox"},
			       $$transHash{"coordformat"}, 
                               $$transHash{"accuracy"},
                               #$$transHash{"raaccuracy"},
                               #$$transHash{"decaccuracy"}, 
			       $$transHash{"coordformats"});
    }
			   
    if ($$transHash{"mjds"}) {
        transformDate($line, $$transHash{"mjds"}, $$transHash{"dateformat"},
	    $$transHash{"mjdformats"});
    }
    
    if ($$transHash{"classes"}) {
        transformClass($line, $$transHash{"classes"});
    }
    
    if ($$transHash{"truncs"}) {
        truncateFields($line, $$transHash{"truncs"},$$transHash{"formats"});
    }
    
    if ($$transHash{"longs"}) {
        prepareLongs($line, $$transHash{"longs"});
    }

} #eosub--TableFormat::transformLine

#-----------------------------------------------------------------------
## Coordinate conversions.
sub transformCoords  {

    my ($line, $racols, $deccols, $coordsys, $equinox, $coordformat,
	#$accuracy, $raaccuracy, $decaccuracy , $truncs, $formats) = @_;
	$accuracy, $coordformats) = @_;

    my $i;
    for($i=0; $i <= $#$racols; $i += 1) {
    
        my $racol  = $$racols[$i];
	my $deccol = $$deccols[$i];
        my $ratrunc = $$coordformats{$racol};
        my $dectrunc = $$coordformats{$deccol};
	
	my ($x,$y) = ($$line[$racol], $$line[$deccol]);
	
	#($$line[$racol], $$line[$deccol]) = fromJ2000($x, $y, $coordsys, 
	   #$equinox, $coordformat,$accuracy,$raaccuracy, $decaccuracy);
	($$line[$racol], $$line[$deccol]) = fromJ2000($x, $y, $coordsys, 
	   $equinox, $coordformat,$accuracy,$ratrunc, $dectrunc);

	print "----> fromJ2000: $i cols=$racol, $deccol rform=$ratrunc, " .
	" dform=$dectrunc x=$x, y=$y returns=$$line[$racol] , " .
	" $$line[$deccol]  <br>" if $::debug;

    }
}
    
#-----------------------------------------------------------------------
# Time Conversions
sub transformDate {

    my ($line, $mjdcols, $dateformat, $formats)  = @_;
    
    if (lc($dateformat) eq "mjd") {
        return;
    }

    #--- get mjd columns and check for format
    my $i;
    for ($i=0; $i<= $#$mjdcols; $i += 1) {
       my $dformat;
       my $dcol = $$mjdcols[$i];
         
       if ($$formats{"$dcol"}) {
          $$formats{"$dcol"} =~ /\d*\.(\d+)f/;
	  $dformat = $1;
       }
       $$line[$dcol] = DateTime::mjd2iso($$line[$dcol],$dformat) if $$line[$dcol];
    }
}

#-----------------------------------------------------------------------
# Class conversion...
sub transformClass {

    my ($line, $classcols) = @_;
    
    my $i;
    
    foreach $i(@$classcols) {
	$$line[$i] = $HDBclass::name{$$line[$i]};
    }
}

#-----------------------------------------------------------------------
# Column reformatting and truncation.
sub truncateFields {

    my ($line, $truncols, $forms) = @_;
    
    my $i;

    for ($i=0; $i<= $#$truncols; $i += 1) {
    
        my $tcol = $$truncols[$i];
	my $tfor = $$forms[$i];
    	$$line[$tcol] = sprintf($tfor, $$line[$tcol]) if defined $$line[$tcol];
    }
}

#-----------------------------------------------------------------------
# Prepare long fields.
sub prepareLongs {
    my ($line, $longs) = @_;
    
    my $i;
	
    foreach $i(@$longs) {
	$$line[$i] = presplit($$line[$i]);
    }
}

#-----------------------------------------------------------------------
# This function gets a long string ready
# to be split into multiple lines.  The first space
# after 55 characters is replaced with a '\0' and
# then the remainder of the string is presplit.
sub presplit {

    my ($string) = @_;
    
    my $pos = 55;
    
    while ($pos < length($string)) {
        my $next = index($string, " ", $pos);
	if ($next >= 0) {
	    substr($string, $next, 1) = "\0";
	    $pos = $next + 55;
	} else {
	    return $string;
	}
    }
    return $string;
}
        
    

#-----------------------------------------------------------------------
# Convert data from J2000 to desired format.
sub fromJ2000 {

   my ($ra, $dec, $sys, $eq, $format,$accuracy, $raaccuracy, $decaccuracy) = @_;
    return if !defined($ra) and !defined($dec);
    if (lc($sys) !~ /equatorial/i || $eq ne '2000') {
        ($ra, $dec) = fromJ2000Rad($ra*$dtor, $dec*$dtor, lc($sys), $eq);
	$ra = $ra/$dtor;
	if ($ra < 0) {
	    $ra += 360;
	}
	$dec = $dec/$dtor;
        #print "TF::$ra, $dec, $sys, $eq, $format,$accuracy <br>";
        if (lc($sys) !~ /equatorial/i) {
          $accuracy ||=4; # default accuracy for galactic format
          return (sprintf("%.${accuracy}f", $ra),
		sprintf("%.${accuracy}f", $dec));
        }
    }
    
    if (lc($sys) =~ /equatorial/i && lc($format) ne 'decimal') {

        #--- Parse for format that conversion program expects
        $raaccuracy =~ /\%\d*\.(\d+)f/;
        my $racc = $1;
        $decaccuracy =~ /\%\d*\.(\d+)f/;
        my $dacc = $1;

        #return (Astro::CoordConv::RA_Decimal2HMS($ra,$raaccuracy), 
	#        Astro::CoordConv::Dec_Decimal2DMS($dec,$decaccuracy));
        #my ($newra,$newdec) = Astro::CoordConv::Equ2Sexagesimal($ra,$dec,$accuracy); 
        my ($newra,$newdec) = Astro::CoordConv::Equ2Sexagesimal($ra,$dec,
		$racc, $dacc); 
        $newra = "$ra !!" if $newra =~ /error/i;
        $newdec = "$dec !!" if $newdec =~ /error/i;
        return ($newra,$newdec);
    } else {
        if ($ra < 0) {
	    $ra += 360;
	}
        $accuracy ||=6; # default accuracy for non-sexagesimal format
        return (sprintf("%.${accuracy}f", $ra),sprintf("%.${accuracy}f", $dec));
    }
} #eosub--TableFormat::fromJ2000

#-----------------------------------------------------------------------
# Call the appropriate SLA library transformations.
sub fromJ2000Rad {

    my ($ra, $dec, $sys, $eq) = @_;
    
    
    if ($sys =~ /galactic/i) {
        return SLA::slaEqgal($ra, $dec);
    }
    
    my $neq = $eq;
    if (uc(substr($neq, 0, 1)) eq "B"  || uc(substr($neq, 0, 1)) eq "J") {
        $neq = substr($neq, 1);
    }
    
    if ($neq < 1700) {
	return ($ra, $dec);
    }
    
    if ($sys =~ /ecliptic/i) {
        my $mjd;
        if ($neq - int($neq) > 0) {
	    my $t1 = DateTime::ymd2mjd(int($neq),1,1);
	    my $t2 = DateTime::ymd2mjd(int($neq)+1,1,1);
	    $mjd = $t1 + ($neq-int($neq)) * ($t2-$t1);
	} else {
	    $mjd = DateTime::ymd2mjd($neq, 1, 1);
	}
        ($ra, $dec) =  SLA::slaEqecl($ra, $dec, $mjd);
	
	return ($ra, $dec);
    }
    
    if ($sys =~ /equatorial/i) {
    
        if ($eq == 1950 || $eq eq "B1950") {
	    return SLA::slaFk54z($ra, $dec);
	} elsif (uc(substr($eq, 0, 1)) eq 'J' || $neq > 1975) {
            return SLA::slaPreces("FK5", 2000, $neq, $ra, $dec);
	} else {	
            ($ra,$dec) = SLA::slaFk54z($ra, $dec);
	    return SLA::slaPreces("FK4", 1950, $neq, $ra, $dec);
	}
    }
} #eosub--TableFormat::fromJ2000Rad


#-----------------------------------------------------------------------
# Add white-space to make fields align properly
sub neatenTable {

    my  ($results, 	# Table of results
         $names, 	# Column names
	 $types, 	# Column types
	 $overrides	# Fields where we want the reverse of the normal alignments.
	)
	= @_;
    
    my %reversePad;
    
    foreach (@$overrides) {
        $reversePad{$_} = 1;
    }
    
    my $ncol = $#$names + 1;
    
    my @lengths;
    my $i;
    
    my $nrow = $#$results + 1;
    my $j;
    
    my $row;
    
    my $maxlen = 0;
    
    # Find the width needed for each column.
    for ($i=0; $i < $ncol; $i += 1) {
    
        # Don't do anything about long strings.
	if ($$types[$i] == -1) {
	    next;
	}

        $lengths[$i] = length($$names[$i]);
	
        foreach $row (@$results) {

	    my $len = length($$row[$i]);
	    	    
	    if ($len > $lengths[$i]) {
	        $lengths[$i] = $len;
	    }
	    
	}
	
	if ($lengths[$i] > $maxlen) {
	    $maxlen = $lengths[$i];
	}
    }

    # Make a buffer of white space.
    my $space = " " x $maxlen;
    
    # Now add the white space back in.
    for ($i=0; $i < $ncol; $i += 1) {
       
        my $max = $lengths[$i];
	my $left = Browse::DBQ::padOnLeft($$types[$i]);
	
	if ($reversePad{$$names[$i]} ) {
	    $left = !$left;
	}

        # White space is always appended to column names.
	$$names[$i] .= substr($space, 0, $max-length($$names[$i]));
	
        foreach $row (@$results) {
	
	    # White space is prepended to 'numeric' types and
	    # appended to 'character' types.
	    
	    if ($left) {
	        $$row[$i] = substr($space, 0, $max-length($$row[$i])).$$row[$i];
	    } else {
	        $$row[$i] .= substr($space, 0, $max-length($$row[$i]));
	    }
	    
	}
	
    }
    
} #eosub--TableFormat::neatenTable


#-----------------------------------------------------------------------
# This generates an output string giving a full tables results.
#
sub render {

    my ($results, $names) = @_;
    
    my @rendering;
    
    my $divider = '+';
    foreach (@$names) {
        $divider .= '-' x length($_) . '+';
    }
    $divider .= "\n";
    
    
    push(@rendering, $divider, pushLine($names), $divider);
    foreach (@$results) {
        push(@rendering, pushLine($_));
    }
    
    push(@rendering, $divider);
    push(@rendering, "(".($#$results+1)." rows)\n");
    return \@rendering;
} #eosub--TableFormat::render


#-----------------------------------------------------------------------
sub pushLine {

    my ($line) = @_;
    
    return '|'.join('|', @$line)."|\n";
} #eosub--TableFormat::pushLine
#-----------------------------------------------------------------------

1; 
#########
#  EOF  #
#########
