#=============================================================================#
# Module Name:    Astro::Object
# Module Author:  Edward J. Sabol
# Creation Date:  1997/05
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Object-oriented container for astronomical names or coordinates.
#=============================================================================#
#
# $Id: Object.pm,v 1.13 2004/03/16 04:14:50 sabol Exp $
# $Log: Object.pm,v $
# Revision 1.13  2004/03/16 04:14:50  sabol
# Constructor returns an error if the name resolver returns multiple, ambiguous
#   results and $Astro::Config::ambiguous_match_is_error is true.
#
# Revision 1.12  2003/02/03 21:13:04  sabol
# Updated comments style. Added the following new methods: attribute(),
#   set_attribute(), list_attributes(), resolver_matches(), and
#   name_is_ambiguous(). The resolve() method can now check all resolvers for
#   multiple matches. The first match is still used, but the others can be
#   retrieved through the resolver_matches() methods. The name_is_ambiguous()
#   method allows the client to determine if there are multiple matches that
#   disagree by more than a specified tolerance. The attribute(),
#   set_attribute(), and list_attributes() methods add formal support for
#   client and run-time defined object attributes. The resolvers now use this
#   method to return additional information such as alternative object names,
#   object types, and bibliographic reference codes.
#
# Revision 1.11  2001/08/09 20:58:45  sabol
# Added support for 'accuracy' parameter to coordinates() method. The
#   coordinates() method also now supports a 'decimal' format, and any errors
#   reported by the coordinates() method are returned inside an array
#   reference for consistency with external usage.
#
# Revision 1.10  2001/08/08 18:51:52  sabol
# When parsing coordinates, change values formatted using Fortran "D"
#   scientific notation to the more commonly accepted "E" scientific notation.
#   Sybase doesn't grok "D" notation, for example.
#
# Revision 1.9  2001/05/11 20:40:42  sabol
# Decimal degrees in scientific notation are now valid inputs to the parse()
#   method.
#
# Revision 1.8  2001/05/03 18:33:22  sabol
# Revamped parse() method with better format checking and error reporting.
#
# Revision 1.7  2000/11/15 15:58:35  sabol
# Added 'default' format option to the coordinates() method that formats the
#   coordinates to the same number of decimal places as
#   $Astro::Config::decimal_accuracy.
#
# Revision 1.6  1999/11/10 18:39:07  sabol
# Due to a change in Perl 5.005, the result of an sprintf() call with a
#   floating point format is considered to be tainted; such values are now
#   detainted automatically.
#
# Revision 1.5  1999/08/11 11:46:27  sabol
# Moved all 'use' directives after the package declaration.
#
# Revision 1.4  1999/08/10 18:40:31  sabol
# Added object class overview for better documentation. Labeled parse() and
#   resolve() methods as private (not for external use). The 'coordinates'
#   argument for the new() constructor method can now be a string which will be
#   parsed. However, if any alphabetic characters are present, an error will
#   result. Added angular_separation() and rect_ranges() methods.
#
# Revision 1.3  1998/06/19 15:46:25  sabol
# The constructor, new(), is more robust in dealing with undefined argument
#   values. Some of the error messages have been tweaked. Added the error()
#   method for retrieving constructor error messages.
#
# Revision 1.2  1998/06/16 13:41:54  sabol
# Now uses Astro::Config instead of Astro::Utils. Replaced all references to
#   Astro::Utils with Astro::Config. Replaced all "<br>" and "\n" strings with
#   $Astro::Config::eol.
#
# Revision 1.1  1998/06/09 18:04:19  sabol
# The coordinates() method now takes an optional 'format' argument which can
#   either be the string 'sexagesimal' or a reference to a two-element array of
#   sprintf()-compatible formats. Added resolver_cache_used() method for
#   determining whether the cache was used when doing any name resolving.
#   Ecliptic sexagesimal coordinates are now correctly parsed in the parse()
#   method.
#
# Revision 1.0  1997/05/01 15:28:04  sabol
# Initial release.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Object;

use strict;
use Astro::Config;
use Astro::Utils;
use Astro::CoordConv;

#------------------------------------------------------------------------------
# Object class overview:
#
# See documentation for individual methods for more details.
#
# Public methods:
#	new			Creates a new object.
#	coordinates		Returns the coordinates.
#	coordsys		Returns the Astro::Coordsys object.
#	convert			Converts to another coordinate system.
#	is_name			Non-zero return value if object was a name.
#	name_is_ambiguous	Non-zero return value if object was a name and
#                                 the name resolver returned multiple matches.
#	name			Returns the name if one was specified.
#	entry			Returns the 'string' argument used.
#	error			Returns the error message if there was one.
#	attribute		Returns the value of client-defined attribute.
#	set_attribute		Assigns value to client-defined attribute.
#	list_attributes		Returns list of attributes that the object has.
#	resolver_used		Returns the name of the resolver object used.
#	resolver_cache_used	Returns the name of the cache used.
#	resolver_matches	Returns array of hashes containing
#                                 information on all of the possible matches
#                                 (if there was more than one) returned by the
#                                 name resolver.
#	angular_separation	Returns angular distance between two points.
#	rect_ranges		Returns ranges of rectangle encompassing circle.
#
# Private methods:
#	parse			Do not use this method. Internal use only.
#	resolve			Do not use this method. Internal use only.
#
# Also, do not access internal data members.
#
#------------------------------------------------------------------------------
sub new {
    # Creates blessed reference to astronomical name/coordinates object.
    #
    # Constructor parameters:
    #   a hash with the following keys and corresponding values
    #     coordinates : ref to anonymous array of coordinates
    #     name        : string consisting of the astronomical object's name
    #     coordsys    : an Astro::CoordSys object (required)
    #     string      : string consisting of either an object name or
    #                     comma-separated list of coordinates; parse() method
    #                     is automatically invoked by constructor
    #     resolver    : either a reference to an Astro::Object::Resolver object
    #                     or a string consisting of one of the following:
    #                     'SIMBAD', 'NED', 'CheckCaches', or any combination
    #                     of those separated by a '/' character
    #                     (required if 'name' is specified without
    #                     'coordinates' or if 'string' is specified)
    #
    # Note: A reference to a hash is returned is returned if a constructor
    #       error is encountered. One of the key values of the ref'd hash will
    #       be 'error' and the value will be a string that describes what the
    #       error was. You can also use the error() method, which is probably
    #       the preferred way of finding out what the constructor error was
    #       (if any).

    my($this) = shift;
    my($class) = ref($this) || $this;
    my($self) = {};
    bless $self, $class;
    my(%args) = @_;
    $self->{'coordsys'} = (defined($args{'coordsys'}) && ref($args{'coordsys'}))
      ? $args{'coordsys'} : Astro::CoordSys->new('system' => 'Equatorial',
						 'equinox' => 2000);
    my($error_handler) = ((defined($Astro::Config::error_handler)
			   && ref($Astro::Config::error_handler))
			  ? $Astro::Config::error_handler : sub { });
    my($error) = '';

    if (((exists($args{'name'}) && !defined($args{'coordinates'}))
	 || exists($args{'string'}))
	&& (!exists($args{'resolver'}) || !ref($args{'resolver'}))) {
	$error = 'Error: A valid name resolver object must be specified.';
	&$error_handler($error);
	$self->{'error'} = $error;
    }
    if (exists($args{'coordinates'})) {
	if ((ref($args{'coordinates'}) eq 'ARRAY')
	    && (scalar(@{ $args{'coordinates'} }) == 2)
	    && $args{'coordinates'}->[0] =~ /^[-+]?\d+(\.(\d+)?)?$/
	    && $args{'coordinates'}->[1] =~ /^[-+]?\d+(\.(\d+)?)?$/) {
	    $self->{'coordinates'} = $args{'coordinates'};
	    if (defined($args{'name'}) && ($args{'name'} !~ /^\s*$/)) {
		$self->{'name'} = &$Astro::Config::cleaner($args{'name'});
	    }
	} elsif (!ref($args{'coordinates'}) && length($args{'coordinates'})
		 && $args{'coordinates'} !~ /[a-z]/i) {
	    $self->{'entry'} = $args{'coordinates'};
	    $error = $self->parse();
	    if ($error) {
		&$error_handler($error);
		$self->{'error'} = $error;
	    }
	} else {
	    $error = "Error: Invalid 'coordinates' argument given to " .
	             "Astro::Object constructor.";
	    &$error_handler($error);
	    $self->{'error'} = $error;
	}
    } elsif (exists($args{'name'})) {
	if (defined($args{'name'}) && ($args{'name'} !~ /^\s*$/)) {
	    $self->{'name'} = &$Astro::Config::cleaner($args{'name'});
	    $error = $self->resolve($args{'resolver'});
	    if ($error) {
		&$error_handler($error);
		$self->{'error'} = $error;
	    } elsif ($Astro::Config::ambiguous_match_is_error
		     && $self->name_is_ambiguous) {
		$error = "Error: The object name \"$self->{'name'}\" is ambiguous.";
 		&$error_handler($error);
		$self->{'error'} = $error;
	    }
	} else {
	    $error = "Error: Invalid 'name' argument given to " .
	             "Astro::Object constructor.";
	    &$error_handler($error);
	    $self->{'error'} = $error;
	}
    } elsif (exists($args{'string'})) {
	$self->{'entry'} = $args{'string'};
	$error = $self->parse();
	if ($error) {
	    &$error_handler($error);
	    $self->{'error'} = $error;
	} elsif ($self->is_name()) {
	    $error = $self->resolve($args{'resolver'});
	    if ($error) {
		&$error_handler($error);
		$self->{'error'} = $error;
	    } elsif ($Astro::Config::ambiguous_match_is_error
		     && $self->name_is_ambiguous) {
		$error = "Error: The object name \"$self->{'name'}\" is ambiguous.";
 		&$error_handler($error);
		$self->{'error'} = $error;
	    }
	}
    } else {
	$error = "Error: Either a 'name', 'coordinates', or 'string' " .
	         "argument must be supplied to the Astro::Object constructor.";
	&$error_handler($error);
	$self->{'error'} = $error;
    }

    return($self);

} #eosub--Astro::Object::new
#------------------------------------------------------------------------------
sub coordinates {
    # Returns the object's coordinates (or error message) as an array ref.
    #
    # Method parameters:
    #      (optional) 'format'   => (a) 'sexagesimal'  --or--
    #                               (b) 'decimal'      --or--
    #                               (c) 'default'      --or--
    #                               (d) a reference to a two-element array of
    #                                   sprintf()-compatible formats
    #      (optional) 'accuracy' => number of significant digits to the right
    #                               of the decimal point (ignored if (c) or (d)
    #                               'format' options are specified)

    my($self) = shift(@_);
    my(%args) = @_;

    my($accuracy);
    if (defined($args{'accuracy'})) {
	return([ 'Error: Specified coordinates accuracy is not a valid non-negative number.' ]) unless ($args{'accuracy'} =~ /^(\d+(\.\d*)?|\.\d+)$/);
	$accuracy = $args{'accuracy'};
    }

    my($coord1,$coord2) = @{ $self->{'coordinates'} };

    if ($args{'format'}) {
	if ($args{'format'} eq 'sexagesimal') {
	    if ($self->{'coordsys'}{'short'} eq 'Eq') {
		($coord1,$coord2) =
		  &Astro::CoordConv::Equ2Sexagesimal($coord1,$coord2,$accuracy);
	    } else {
		($coord1,$coord2) =
		  &Astro::CoordConv::Gal2Sexagesimal($coord1,$coord2,$accuracy);
	    }
	} elsif ($args{'format'} eq 'decimal') {
	    $accuracy = $Astro::Config::decimal_accuracy
	      unless (defined($accuracy));
	    my($fmt) = '%.' . $accuracy . 'f';
	    $coord1 = sprintf($fmt,$coord1);
	    $coord2 = sprintf($fmt,$coord2);
	    ($coord1 =~ /^(.*)$/) && ($coord1 = $1);
	    ($coord2 =~ /^(.*)$/) && ($coord2 = $1);
	} elsif ($args{'format'} eq 'default') {
	    my($fmt) = '%.' . $Astro::Config::decimal_accuracy . 'f';
	    $coord1 = sprintf($fmt,$coord1);
	    $coord2 = sprintf($fmt,$coord2);
	    ($coord1 =~ /^(.*)$/) && ($coord1 = $1);
	    ($coord2 =~ /^(.*)$/) && ($coord2 = $1);
	} elsif (ref($args{'format'}) eq 'ARRAY') {
	    my($fmt1,$fmt2) = @{ $args{'format'} };
	    $coord1 = sprintf($fmt1,$coord1);
	    $coord2 = sprintf($fmt2,$coord2);
	    ($coord1 =~ /^(.*)$/) && ($coord1 = $1);
	    ($coord2 =~ /^(.*)$/) && ($coord2 = $1);
	} else {
	    return([ "Error: Invalid coordinates 'format' option." ]);
	}
    } elsif (defined($accuracy)) {
	my($fmt) = '%.' . $accuracy . 'f';
	$coord1 = sprintf($fmt,$coord1);
	$coord2 = sprintf($fmt,$coord2);
	($coord1 =~ /^(.*)$/) && ($coord1 = $1);
	($coord2 =~ /^(.*)$/) && ($coord2 = $1);
    }

    return([ $coord1, $coord2 ]);

} #eosub--Astro::Object::coordinates
#------------------------------------------------------------------------------
sub coordsys {
    # Returns the object's coordinate system (as an Astro::CoordSys object).
    #
    # Method parameters: none

    my($self) = shift(@_);

    return($self->{'coordsys'});

} #eosub--Astro::Object::coordsys
#------------------------------------------------------------------------------
sub is_name {
    # Returns 0 or non-zero value depending on whether the object is an
    #   astronomical object name or a pair of coordinates.
    #
    # Method parameters: none

    my($self) = shift(@_);

    return(defined($self->{'name'}));

} #eosub--Astro::Object::is_name
#------------------------------------------------------------------------------
sub name {
    # Returns the object's astronomical name if it has one; otherwise, it
    #   returns a zero-length string.
    #
    # Method parameters: none

    my($self) = shift(@_);

    return($self->is_name() ? $self->{'name'} : '');

} #eosub--Astro::Object::name
#------------------------------------------------------------------------------
sub name_is_ambiguous {
    # Returns non-zero if object name was ambiguously resolved to multiple
    #   possibilities.
    #
    # Method parameters: (optional) 'tolerance' => maximum offset in arcminutes

    my($self) = shift(@_);
    my(%args) = @_;
    my($tolerance) = ($args{'tolerance'}
		      || $Astro::Config::ambiguous_match_tolerance || 0);

    if (defined($self->{'resolver_matches'})
	&& (scalar(@{ $self->{'resolver_matches'} }) > 1)) {
	my @matches = @{ $self->{'resolver_matches'} };
	print 'Astro::Object::name_is_ambiguous: total number of matches = ',
	      scalar(@matches),$Astro::Config::eol if $::debug;
	my %matches_per_resolver = ();
	my $match;
	foreach $match (@matches) {
	    $matches_per_resolver{$match->{'resolver_used'}}++;
	}
#  	&Util::DumpHash({ 'matches_per_resolver' => \%matches_per_resolver });
	my $status = 0;
	my $resolver;
	foreach $resolver (keys(%matches_per_resolver)) {
	    $status ||= $matches_per_resolver{$resolver} > 1;
	}
	return($status) if $status;
	# Only one match per resolver... Compare the coordinates to see if
	# they are the same.
	my $max_offset = 0;
	shift(@matches);
	foreach $match (@matches) {
	    my $this_offset = $match->attribute('offset_from_primary_match');
	    if ($this_offset > $max_offset) {
		$max_offset = $this_offset;
	    }
	}
	return($max_offset > $tolerance);
    } else {
	return(0);
    }

} #eosub--Astro::Object::name_is_ambiguous
#------------------------------------------------------------------------------
sub set_attribute {
    # Allows for generic, client-defined, run-time attributes to be attached
    #   to an object. One could, but just assign values to the internal data
    #   hash directly, but this is a cleaner interface.
    #
    # Method parameters: hash specifying the attribute name and value you
    #   wish to attach to the object -- or -- just the attribute name without
    #   a value will delete that attribute

    my($self) = shift(@_);
    if (!scalar(@_)) {
	unless (ref($self->{'attributes'})) {
	    $self->{'attributes'} = {};
	}
    } elsif (scalar(@_) == 1) {
	my($attribute) = @_;
	if (ref($self->{'attributes'})) {
	    delete($self->{'attributes'}{$attribute});
	}
	unless (scalar(keys(%{ $self->{'attributes'} }))) {
	    delete($self->{'attributes'});
	}
    } else {
	my($attribute,$value) = @_;
	if ($attribute) {
	    unless (ref($self->{'attributes'})) {
		$self->{'attributes'} = {};
	    }
	    $self->{'attributes'}{$attribute} = $value;
	}
    }

} #eosub--Astro::Object::set_attribute
#------------------------------------------------------------------------------
sub attribute {
    # Returns the value of the attribute specified.
    #
    # Method parameters: the name of the attribute you want the value for

    my($self,$attribute) = @_;

    if (ref($self->{'attributes'})) {
	return($self->{'attributes'}{$attribute});
    } else {
	return(undef);
    }

} #eosub--Astro::Object::attribute
#------------------------------------------------------------------------------
sub list_attributes {
    # Returns the list of attributes that have been attached to the object.
    #
    # Method parameters: none

    my($self) = @_;

    if (ref($self->{'attributes'})) {
	return(keys(%{ $self->{'attributes'} }));
    } else {
	return();
    }

} #eosub--Astro::Object::list_attributes
#------------------------------------------------------------------------------
sub resolver_used {
    # Returns the name of the name resolver used to resolve an object's
    #   astronomical name into coordinates if such an action was necessary;
    #   otherwise, it returns an undefined reference.
    #
    # Method parameters: none

    my($self) = shift(@_);

    return($self->{'resolver_used'});

} #eosub--Astro::Object::resolver_used
#------------------------------------------------------------------------------
sub resolver_cache_used {
    # Returns the file name of the name resolver cache used to resolve an
    #   object's astronomical name into coordinates if such an action was
    #   necessary; otherwise, it returns an undefined reference.
    #
    # Method parameters: none

    my($self) = shift(@_);

    return($self->{'resolver_cache_used'});

} #eosub--Astro::Object::resolver_cache_used
#------------------------------------------------------------------------------
sub error {
    # Returns the error message, if there was an error in the constructor
    #   (and only constructor errors!); otherwise, it returns an
    #   undefined reference.
    #
    # Method parameters: none

    my($self) = shift(@_);

    return($self->{'error'});

} #eosub--Astro::Object::error
#------------------------------------------------------------------------------
sub entry {
    # Returns the Astro::Object constructor's 'string' argument that was
    #   parsed in order to populate the internal 'name' or 'coordinates'
    #   data members, if such an action was necessary;
    #   otherwise, it returns an undefined reference.
    #
    # Method parameters: none

    my($self) = shift(@_);

    return($self->{'entry'});

} #eosub--Astro::Object::entry
#------------------------------------------------------------------------------
sub resolver_matches {
    # Returns an array containing all of the matches obtained from querying
    #   the specified name resolvers, but only if the name resolver object
    #   returned multiple matches; otherwise, an empty list is returned.
    #
    # Method parameters: none

    my($self) = shift(@_);

    if (defined($self->{'resolver_matches'})) {
	return(@{ $self->{'resolver_matches'} });
    } else {
	return();
    }

} #eosub--Astro::Object::resolver_matches
#------------------------------------------------------------------------------
sub angular_separation {
    # Returns the angular separation (in arcminutes) between the coordinates
    #   of two Astro::Object objects.
    #
    # Method parameters: another Astro::Object

    my($self,$other) = @_;
    my($self_coordsys) = $self->coordsys();
    return('Error: Not a valid astronomical name/coordinates object.')
      unless (ref($other));
    my($other_coordsys) = $other->coordsys();
    my(@self_coords) = @{ $self->coordinates() };
    my(@other_coords);
    if ($self_coordsys->equals($other->coordsys())) {
	@other_coords = @{ $other->coordinates() };
    } else {
	my($other_same_coordsys) = $other->convert('coordsys'
						   => $self_coordsys);
	@other_coords = @{ $other_same_coordsys->coordinates() };
    }
    return(&Astro::Utils::angsep(@self_coords,@other_coords));

} #eosub--Astro::Object::angular_separation
#------------------------------------------------------------------------------
sub rect_ranges {
    # Given the radius of a circle around a point on a sphere, this method
    #   finds the smallest "rectangular" (in spherical coordinates) region on
    #   the sphere that encloses the circle.
    #
    # Returns array of two pair of coordinates ranges in decimal degrees.
    #
    # Method parameters: 'radius' => value in arcminutes

    my($self) = shift(@_);
    my(%args) = @_;
    return('Error: The radius is not a valid real number.')
      unless ($args{'radius'} =~ /^[-\+]?(\d+(\.\d*)?|\.\d+)$/);
    my($r) = $args{'radius'} / 60.0;
    my(@coords) = @{ $self->coordinates() };

    return(&Astro::Utils::sphrect($r,@coords));

} #eosub--Astro::Object::rect_ranges
#------------------------------------------------------------------------------
sub parse {
    # Private method.
    #
    # Parses the object's 'entry' member and populates the 'coordinates'
    #   and possibly the 'name' members as appropriate. Returns a string of
    #   non-zero length if there is an error.
    #
    # Method parameters: none

    my($self) = shift(@_);

    local($_) = $self->{'entry'} = &$Astro::Config::cleaner($self->{'entry'});

    return('Error: Please enter a valid object name or a pair of coordinates.')
      if ($self->{'entry'} eq '');

    my($error) = '';
    my($coord1,$coord2) = ('','');

    my($sign)         = '[+-]?';
    my($decimal_exp)  = '(\d\.\d*|\.\d+)[eEdD][+-]?\d{1,3}';
    my($decimal)      = '((\d{1,3}\.\d*|\.\d+)|'.$decimal_exp.')';
    my($delimiter)    = '\s*(,\s*|[+-])';
    my($sci_notation) = $sign . $decimal . $delimiter . $sign . $decimal;

    if (/[a-z]/i && !/^($sci_notation)$/) {
	# Alphas imply that an object name was specified unless it's
	# scientific notation.
	$self->{'name'} = $_;
    } else {		# Check for valid coordinate input.

	# Build pattern for valid coordinates.
	my($hr_min)       = '\d{1,2}\s+([0-5]?[0-9](\.\d*)?|\.\d+)';
	my($hr_min_sec)   = '\d{1,2}\s+([0-5]?[0-9])\s+([0-5]?[0-9](\.\d*)?|\.\d+)';
	my($deg_min)      = '\d{1,3}\s+([0-5]?[0-9](\.\d*)?|\.\d+)';
	my($deg_min_sec)  = '\d{1,3}\s+([0-5]?[0-9])\s+([0-5]?[0-9](\.\d*)?|\.\d+)';
	my($coordpattern) = $sign.'('.$decimal.'|';
	if ($self->{'coordsys'}{'short'} eq 'Eq') {
	    $coordpattern .= $hr_min.'|'.$hr_min_sec.')';
	} else {
	    $coordpattern .= $deg_min.'|'.$deg_min_sec.')';
	}
	my($pattern)      = $coordpattern . $delimiter . $coordpattern;

	my($neg) = 0;		# Flag indicating whether 1st coord is negative
	if (/^($pattern)$/) {	# User entered a valid set of coordinates.

	    if (/^-/) {		# 1st coord is negative
		s/^[-+]//;	# Trim leading sign to make splits easier.
		$neg = 1;	# Set flag to make 1st coord negative later.
	    }

	    if (/,/) {
		($coord1,$coord2) = split(/,/,$_);
	    } elsif (/^($sci_notation)$/) {
		s/[dD]/E/g;	# Change Fortran "D" notation to the more
				# commonly accepted "E" notation.
		if (/\s+\+/) {
		    ($coord1,$coord2) = split(/\s+\+/,$_);
		} elsif (/-/) {
		    ($coord1,$coord2) = split(/\s+-/,$_);
		    $coord2 = '-' . $coord2;
		}
	    } elsif (/\+/) {
		($coord1,$coord2) = split(/\+/,$_);
	    } elsif (/-/) {
		($coord1,$coord2) = split(/-/,$_);
		$coord2 = '-' . $coord2;
	    }

	    ($coord1 = '-' . $coord1) if $neg;

	    # Clean up coordinates before returning.
	    $coord1 = &$Astro::Config::cleaner($coord1);
	    $coord2 = &$Astro::Config::cleaner($coord2);

	    # Convert coordinates to decimal degress (if necessary).
	    if ($self->{'coordsys'}{'short'} eq 'Eq') {
		($coord1,$coord2) = &Astro::CoordConv::Equ2Decimal($coord1,
								   $coord2);
	    } elsif ($self->{'coordsys'}{'short'} eq 'G') {
		($coord1,$coord2) = &Astro::CoordConv::Gal2Decimal($coord1,
								   $coord2);
	    } elsif ($self->{'coordsys'}{'short'} eq 'Ec') {
		($coord1,$coord2) = &Astro::CoordConv::Ecl2Decimal($coord1,
								   $coord2);
	    } else {
		$coord1 = &Astro::CoordConv::DMS2Decimal($coord1);
		$coord2 = &Astro::CoordConv::DMS2Decimal($coord2);
	    }

	    if ($coord1 =~ /error/i) {
		$error = $coord1;
	    }
	    if ($coord2 =~ /error/i) {
		$error .= ($error ? $Astro::Config::eol : '') . $coord2;
	    }
	    if ($error) {
		$error .= ' Please enter appropriate coordinate values.';
	    } else {
		$self->{'coordinates'} = [ $coord1, $coord2 ];
	    }

	} else { # User entered invalid coordinates string.
	    $error = 'Error: Invalid format for coordinates!';
	}
    }

    if ($::debug) {
	print 'Astro::Object::parse: Error: ',$error,$Astro::Config::eol
	  if $error;
	print 'Astro::Object::parse: Coord1: ',$coord1,$Astro::Config::eol,
	      'Astro::Object::parse: Coord2: ',$coord2,$Astro::Config::eol,
	      'Astro::Object::parse: Object: ',$self->{'name'},
	       $Astro::Config::eol,$Astro::Config::eol;
    }

    return($error);

} #eosub--Astro::Object::parse
#------------------------------------------------------------------------------
sub convert {
    # Converts the Astro::Object to another coordinate system. Return value
    #   is either an error string or another Astro::Object (not the same as
    #   $self!).
    #
    # Method parameters: 'coordsys' => Astro::CoordSys object of destination
    #   coordinate system

    my($self) = shift(@_);
    my($input_coord_sys) = $self->{'coordsys'};
    my(%args) = @_;
    my($output_coord_sys) = $args{'coordsys'};
    local($_);

    if (!ref($input_coord_sys) || !ref($output_coord_sys)) {
	return('Error: Not a valid coordinate system object.');
    }

    print 'Astro::Object::convert: input_coord_sys = ',
          $input_coord_sys->name(),'; output_coord_sys = ',
          $output_coord_sys->name(),$Astro::Config::eol if $::debug;

    my($coord1,$coord2) = @{ $self->{'coordinates'} };

    print "Astro::Object::convert: input_coordinates = ($coord1,$coord2)",
          $Astro::Config::eol if $::debug;

    my($new_coord1,$new_coord2) = &$Astro::Config::converter($coord1,$coord2,
							     $input_coord_sys,
							     $output_coord_sys);
    my($error);
    if ($new_coord1 !~ /error/i) {
	print 'Astro::Object::convert: output_coordinates = ',
	      "($new_coord1,$new_coord2)",$Astro::Config::eol if $::debug;
	my(%constructor_args) = ('coordinates' => [ $new_coord1, $new_coord2 ],
				 'coordsys' => $output_coord_sys);
	$constructor_args{'name'} = $self->{'name'} if
	  (defined($self->{'name'}));
	return(Astro::Object->new(%constructor_args));
    } else {
	$error = $new_coord1;
	print 'Astro::Object::convert: ',$error,$Astro::Config::eol if $::debug;
    }
    return($error);

} #eosub--Astro::Object::convert
#------------------------------------------------------------------------------
sub resolve {
    # Private method.
    #
    # Resolves the 'name' member into coordinates using an
    #   Astro::Object::Resolver and populates the 'coordinates',
    #   'resolver_used', and possibly the 'resolver_cache_used' data
    #   members of $self.
    #
    # Method argument: an Astro::Object::Resolver object
    # Return value: a string describing the error encountered (if any)

    my($self) = shift(@_);
    my($error) = '';

    return("Error: Object is not a name, and thus doesn't need resolving.")
      if (! $self->is_name());

    return("Error: Object already has coordinates, and thus doesn't need " .
	   "to be resolved.") if (defined($self->{'coordinates'}));

    my(@coords) = ();
    my $cache_used = '';
    my $all_matches = [];
    my($resolver) = shift(@_);
    if (ref($resolver) && ref($resolver->next())
	&& $resolver->{'check_all_resolvers'}) {   # check multiple resolvers
	print 'Astro::Object::resolve: Checking multiple resolvers...',
	      $Astro::Config::eol if $::debug;
	my(@results) = ();
	my $first_successful_resolver = undef;
	while (ref($resolver)) {
	    @results = $resolver->resolve($self);
	    if (scalar(@results) == 3) {
		$first_successful_resolver ||= $resolver;
		my $match_cache_used = pop(@results);
		$cache_used ||= $match_cache_used;
		unless (scalar(@coords) == 2) {
		    @coords = @results;
		}
		my $object = Astro::Object->new('coordinates' =>
						   [ @results ],
						'name' => uc($self->{'name'}),
						'coordsys' =>
						   $self->{'coordsys'});
		$object->{'resolver_used'} = $resolver->name();
		$object->{'resolver_cache_used'} = $match_cache_used
		  if ($match_cache_used);
		push(@$all_matches,$object);
	    } elsif (scalar(@results) == 4) {
		$first_successful_resolver ||= $resolver;
		my $matches = pop(@results);
		print 'Astro::Object::resolve: number of matches = ',
		      scalar(@$matches),$Astro::Config::eol if $::debug;
		my $match_cache_used = pop(@results);
		$cache_used ||= $match_cache_used;
		unless (scalar(@coords) == 2) {
		    @coords = @results;
		}
		my $resolver_name = $resolver->name();
		my $match;
		foreach $match (@$matches) {
		    $match->{'resolver_used'} = $resolver_name;
		}
		push(@$all_matches,@{ $matches });
	    } elsif ((scalar(@coords) != 2) && scalar(@results)
		     && ($results[0] =~ /(error|warning)/i)) {
		$coords[0] = $results[0];
	    }
	    $resolver = $resolver->next();
	}
	$resolver = $first_successful_resolver;
    } else {
	while (ref($resolver)) {
	    @coords = $resolver->resolve($self);
	    if (scalar(@coords) == 3) {
		$cache_used = pop(@coords);
		last;
	    } elsif (scalar(@coords) == 4) {
		my $matches = pop(@coords);
		$cache_used = pop(@coords);
		push(@$all_matches,@{ $matches });
		last;
	    }
	    $resolver = $resolver->next();
	}
    }

    if (scalar(@coords) == 1) {
	$error = shift(@coords);
    } elsif (!scalar(@coords)) {
	$error = 'Error: Unknown problem encountered while name resolving.';
    } else {
	$self->{'resolver_used'} = $resolver->name();
	$self->{'resolver_cache_used'} = $cache_used if ($cache_used);
	$self->{'coordinates'} = [ @coords ];
	if (scalar(@$all_matches) > 1) {
	    my $match;
	    foreach $match (@$all_matches) {
		$match->set_attribute('offset_from_primary_match' =>
				        $self->angular_separation($match));
	    }
	    $self->{'resolver_matches'} = $all_matches;
	}
    }

    return($error);

} #eosub--Astro::Object::resolve
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
