#=============================================================================#
# Module Name:    Astro::Object::Resolver::CheckCaches
# Module Author:  Edward J. Sabol
# Creation Date:  1998/06
# Project:        HEASARC :: Astro
#=============================================================================#
# Module Description:
#   Object-oriented interface to the SIMBAD name resolver. Object inherits
#   from base class Astro::Object::Resolver.
#=============================================================================#
#
# $Id: CheckCaches.pm,v 1.3 2003/02/03 21:38:17 sabol Exp $
# $Log: CheckCaches.pm,v $
# Revision 1.3  2003/02/03 21:38:17  sabol
# Updated comments style.
#
# Revision 1.2  2001/05/11 20:29:40  sabol
# Added package-wide configuration variable @default_caches_to_check and
#   removed the old method of globbing for DBM files in the cache directory.
#   It appears globbing is not compatible with taint mode. Also, moved
#   wildcards error message from parent object to here.
#
# Revision 1.1  1999/08/11 11:50:14  sabol
# Now uses 'vars' compiler directive to specify package-wide variable.
#   Moved all 'use' directives after the package declaration.
#
# Revision 1.0  1998/06/17 18:42:50  sabol
# Created.
#
#=============================================================================#
# High Energy Astrophysics Science Archive Research Center (HEASARC),
# NASA's Goddard Space Flight Center, Greenbelt, MD 20771, USA.
#
# You are free to use this code as you wish, but unless you change
# it significantly, please leave the above name and address intact.
#=============================================================================#

package Astro::Object::Resolver::CheckCaches;

use strict;
use Astro::Config;
use Astro::CoordSys;

use vars qw(@ISA @default_caches_to_check);
@ISA = qw(Astro::Object::Resolver);
@default_caches_to_check = ('NED','SIMBAD');

#------------------------------------------------------------------------------
sub initialize {
    # Method arguments: Same as Astro::Object::Resolver::new(), except no
    #   'type' option (naturally) and the 'use_cache' and 'cache_read_only'
    #   options are ignored. It also supports one additional option:
    #       'caches': an optional reference to an array of cache file names to
    #                 check; if any file name does not include a path, then
    #                 $Astro::Object::Resolver::default_cache_directory is
    #                 prepended. If this option is not specified, it checks
    #                 all caches in the default_cache_directory.

    my($self) = shift(@_);
    my(%args) = @_;

    $self->{'name'} = 'CheckCaches';
    my($cache_dir) = (($::cgi =~ /W3Browse/i) ? "$::root/$::cgi/dbm" :
		      $Astro::Object::Resolver::default_cache_directory);
    my(@caches) = (ref($args{'caches'}) eq 'ARRAY'
		   ? @{ $args{'caches'} } : @default_caches_to_check);
    local($_);
    foreach (@caches) {
	($_ = $cache_dir . '/' . $_) unless (/\//);
    }
    print 'CheckCaches::initialize: caches are (',join(', ',@caches),')',
          $Astro::Config::eol if $::debug;

    $self->{'caches'} = [ @caches ];
    $self->{'cache_counter'} = 0;
    $self->{'use_cache'} = 0;
    $self->{'cache_read_only'} = 1;
    $self->{'coordsys'} = Astro::CoordSys->new('system' => 'Equatorial',
					       'equinox' => 2000);
    return($self);

} #eosub--Astro::Object::Resolver::CheckCaches::initialize
#------------------------------------------------------------------------------
sub do_resolve {

    my($self,$name) = @_;

    return('Error: Sorry, wildcards in object names to be resolved ' .
	   'by the HEASARC CheckCaches name resolver are not supported. ' .
	   'Please modify your input.') if ($name =~ /[\*\?]/);

    my($output) = 'Error: Invalid object name. Certain characters are not allowed.';
    if ($name =~ /^([^;<>&\"\`\\\|]+)$/) {
	$name = $1;
	print "CheckCaches: name: ",$name,$Astro::Config::eol if $::debug;
	$output = 'Error: Could not access any caches.';
	my($cache);
	while (($self->{'cache_counter'} < scalar($self->{'caches'}))
	       && ($cache = $self->{'caches'}->[$self->{'cache_counter'}])) {
	    print "CheckCaches: Checking $cache... (counter = ",
	          $self->{'cache_counter'},')',$Astro::Config::eol if $::debug;
	    $self->{'cache'} = $cache;
	    $output = $self->check_cache($name);
	    print 'CheckCaches: Result: ',$output,$Astro::Config::eol
	      if $::debug;
	    if (($output !~ /(error|warning)/i) && length($output)) {
		$self->{'cache_used_to_resolve'} = $cache;
		print "CheckCaches: returning (coord1, coord2) = ($output)",
		      $Astro::Config::eol if $::debug;
		last;
	    } else {
		$self->{'cache'} = undef;
		$self->{'cache_counter'}++;
	    }
	}
    }

    return($output);

} #eosub--Astro::Object::Resolver::CheckCaches::do_resolve
#------------------------------------------------------------------------------
#########
#  EOF  #
#########
1;
