###############################################################################
#
# MODULE: $Id: Ftools.pm,v 1.4 2003/07/16 23:58:32 sabol Exp $
#
#  AUTHOR: Edward J. Sabol <sabol@alderaan.gsfc.nasa.gov>
# CREATED: 1997/01/30
#
# DESCRIPTION: Sets up the environment variables for using FTOOLS software
#  in a Perl script. And defines some useful functions for using various
#  FTOOLS.
#
# HISTORY:
#  Thu Jan 30 17:52:13 1997 sabol created
#  Thu Feb  6 19:26:14 1997 sabol modified to use import so that one can
#                                 specify which version (release or develop)
#                                 of FTOOLS to use
#  Fri Feb  7 16:19:50 1997 sabol added fdumphdr (which is exported)
#  Fri Mar 28 11:19:15 1997 sabol "release" is now the default instead of
#                                 "develop"
#  Fri Mar 28 15:04:16 1997 sabol changed "BEGIN" to "sub import" to get rid
#                                 of a compiler warning and to better conform
#                                 with module creation guidelines
#  Fri Apr  4 14:43:40 1997 sabol fixed code structure to avoid compile-time
#                                 warning of usage of an uninitialized variable
#  Fri Jun 20 16:32:53 1997 sabol updated for FTOOLS v4.0 release
#  Mon Sep 29 11:32:37 1997 sabol added setting of FTOOLSOUTPUT as required by
#                                 the latest version of FTOOLS
#  Thu Oct  2 16:57:14 1997 sabol all fatal errors now prefixed with "Error: "
#  Tue Oct  7 11:52:00 1997 sabol tweaked environment variable settings; the
#                                 FTOOLS directories now go at the front of the
#                                 PATH
#  Tue Oct  7 12:04:23 1997 sabol added fdump_open() and fdump() functions;
#                                 added $Ftools::eol package variable for
#                                 specifying the end-of-line string to be used
#                                 for debugging
#  Tue Jul  7 17:06:50 1998 sabol added $Ftools::error_handler package variable
#                                 for specfying a routine to handle errors;
#                                 added "www" as a package argument for using
#                                 FTOOLS in the Web area; tweaked errors
#                                 slightly so they all start with the string
#                                 "Error"; $Ftools::eol will now default to
#                                 "<br>" if CGI::Error() is present;
#                                 increased module version number to 1.04
#  Tue Jul  7 17:57:09 1998 sabol added $Ftools::debug package variable
#                                 and replaced $::debug with it
#  Fri Jul 31 17:25:13 1998 sabol improved usage documentation below
#  Tue Dec  1 14:01:58 1998 sabol added nh() function
#  Wed Dec  2 13:37:53 1998 sabol slightly improved error detection in nh()
#  Wed Dec  2 17:56:36 1998 sabol "use Ftools 'www'" now sets the umask to 002
#  Wed Nov 10 18:11:43 1999 sabol really minor code changes
#  Thu Mar  9 14:00:58 2000 sabol updated environment settings for
#                                 compatibility with HEAsoft 5.0 release
#  Tue Feb 20 17:24:12 2001 sabol added support for optional radius parameter
#                                 to nh()
#  Thu Aug 22 17:36:44 2002 sabol detaint nh command-line arguments; additional
#                                 nh() debug output
#  Fri May 30 13:27:02 2003 sabol now sets $XANADU and sets path correctly on
#                                 Linux
#  Wed Jul 16 19:54:28 2003 sabol now sets $EXT correctly on both Alpha and
#                                 Linux; fdumphdr() can now take an optional
#                                 parameter used to specify the FITS extension
#
###############################################################################
# Note: Module defaults to using the version of FTOOLS specified by
#       the environment variable FTOOLS_VERSION. If that environment variable
#       is not set, it will default to "release".
#       * "use Ftools 'develop';" if you want or need to use the develop
#         version of FTOOLS, etc.
#       * "use Ftools 'www';" if you're using FTOOLS in the HEASARC Web
#         environment
###############################################################################

package Ftools;

require 5.002;
use strict;
use FileHandle;
use Carp;

require Exporter;
use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION $eol $error_handler $debug);
@ISA = qw(Exporter);
@EXPORT = qw(fdumphdr fdump_open fdump nh nh_parse);
@EXPORT_OK = ();
$VERSION = '1.05';

$debug = 0;
	# flag for specifying whether module routines should print
	#   debugging information

$eol = defined(&CGI::Error) ? '<br>' : "\n";
	# if the debugging output is intended for a Web browser, one
	#   should set $Ftools::eol to "<br>" instead

$error_handler = (defined(&CGI::Error) ? \&CGI::Error
		  : defined(&::Error) ? \&::Error
		  : \&croak);
        # reference to subroutine for handling (usually fatal)
        #   errors in Ftools.pm; by default, it is assigned a
        #   reference to either &CGI::Error or &::Error() or croak()

#------------------------------------------------------------------------------
sub import {

    my($pkg) = shift(@_);
    my($callpkg) = caller($Exporter::ExportLevel);
    &Exporter::export($pkg,$callpkg);

    my($version);
    if (scalar(@_)) {
	my($arg) = shift(@_);
	$version = (($arg =~ /www/i) ? 'www'
		    : ($arg =~ /develop/i) ? 'develop'
		    : 'release');
    } else {
	$version = (defined($ENV{'FTOOLS_VERSION'})
		    && ($ENV{'FTOOLS_VERSION'} =~ /^(release|develop)$/))
	  ? $ENV{'FTOOLS_VERSION'} : 'release';
    }

    # Set environment variables for FTOOLS usage.
    my($ftools_root);
    if ($version eq 'www') {
	$ftools_root = '/www/ftools';
	$ENV{'FTOOLS'} = $ftools_root;
	$ENV{'LOCPFILES'} = '/tmp/pfiles';
	umask 002;
    } else {
	$ftools_root = "/software/lheasoft/$version";
	$ENV{'FTOOLS'} = $ftools_root . ($^O eq 'linux' ? '/Linux_2.4_i686'
					 : '/OSF1_4.0_alpha');
	if (!defined($ENV{'LOCPFILES'})) {
	    $ENV{'LOCPFILES'} = $ENV{'HOME'} . '/pfiles';
	}
    }
    $ENV{'LHEASOFT'} = $ENV{'FTOOLS'};
    $ENV{'XANADU'} = $ENV{'FTOOLS'};
    $ENV{'FTOOLS_VERSION'} = $version;
    $ENV{'FTOOLSINPUT'} = 'stdin';
    $ENV{'FTOOLSOUTPUT'} = 'stdout';
    $ENV{'SYSPFILES'} = $ENV{'LHEASOFT'} . (-d ($ENV{'LHEASOFT'} . '/syspfiles')
					    ? '/syspfiles' : '/bin');
    $ENV{'PFILES'} = $ENV{'LOCPFILES'} . ';' . $ENV{'SYSPFILES'};
    $ENV{'LD_LIBRARY_PATH'} = $ENV{'LHEASOFT'} . '/lib' .
      (length($ENV{'LD_LIBRARY_PATH'})
       ? ':' . $ENV{'LD_LIBRARY_PATH'} : '');
    $ENV{'PFCLOBBER'} = '1';
    $ENV{'LHEA_HELP'} = $ftools_root . '/help';
    $ENV{'LHEA_DATA'} = $ftools_root . '/refdata';
    $ENV{'PGPLOT_FONT'} = $ENV{'LHEASOFT'} . '/lib/grfont.dat';
    $ENV{'PGPLOT_RGB'} = $ENV{'LHEASOFT'} . '/lib/rgb.txt';
    $ENV{'PGPLOT_DIR'} = $ENV{'LHEASOFT'} . '/lib';
    $ENV{'POW_LIBRARY'} = $ENV{'LHEASOFT'} . '/lib/pow';
    $ENV{'XRDEFAULTS'} = $ENV{'LHEASOFT'} . '/xrdefaults';
    $ENV{'EXT'} = ($^O eq 'linux' ? 'lnx' : 'osf');
    $ENV{'XANADU'} = $ftools_root;
    $ENV{'XANBIN'} = $ENV{'XANADU'} . '/bin';
    $ENV{'PREFTOOLS_PATH'} = $ENV{'PATH'};
    $ENV{'PATH'} = $ENV{'LHEASOFT'} . '/scripts:' . $ENV{'LHEASOFT'} . '/bin' .
      (($ENV{'PATH'} ne '') ? (':' . $ENV{'PATH'}) : '');

} #eosub--import
#------------------------------------------------------------------------------
sub fdumphdr {
    # Returns a hash created from parsing the FITS file header information.

    my($file,$ext) = @_;

    my(%header,$key,$tmp);

    &$error_handler("Error: Could not open $file! Either file does not exist or insufficient permissions.") unless (-e $file && -r _);
    if (length($ext) && ($ext !~ /^\+/)) { $ext = '+' . $ext; }
    my($command) = "(fdump $file$ext STDOUT - - page=no prdata=no 2>&1) |";
    print "Ftools::fdumphdr: command = $command$eol" if $debug;
    my($fdumpfh) = new FileHandle($command);
    defined($fdumpfh) || &$error_handler("Error: Could not execute fdump!");
    while (<$fdumpfh>) {
	/^(\S+)\s*=\s*\'?\s*([^\']+)\'?\s+\// && do { ($key,$tmp) = ($1,$2);
						      $tmp =~ s/\s*$//;
						      $header{$key} = $tmp;
						    }
    }
    undef $fdumpfh;

    return(%header);

} #eosub--fdumphdr
#------------------------------------------------------------------------------
sub fdump_open {
    # Executes fdump on a FITS file and returns the opened FileHandle.

    my(%args) = @_;

    my($file) = $args{'file'};
    my($columns) = ((ref($args{'columns'}) eq 'ARRAY')
		    ? join(',',@{ $args{'columns'} }) : $args{'columns'});

    if ($file =~ /^\s*([^\'\"\`\\;<>&\|\n]+)\s*$/) {
	$file = $1;
    } else {
	&$error_handler("Error: Illegal file name (\"$file\") passed as argument to Ftools::fdump_open()!");
    }

    if ($columns) {
	if ($columns =~ /^\s*([^\'\"\`\\;<>&\|\n]+)\s*$/) {
	    $columns = $1;
	} else {
	    &$error_handler("Error: Illegal column name list (\"$columns\") passed as argument to Ftools::fdump_open()!");
	}
    }

    my($fdumpfh);
    return $fdumpfh unless (-e $file && -r _);
    my($command) = "(fdump $file STDOUT $columns - page=no fldsep=\"\|\" prhead=no wrap=yes showcol=no showunit=no showrow=no pagewidth=256 showscale=no 2>&1) |";
    $fdumpfh = new FileHandle($command);
    return($fdumpfh);

} #eosub--fdump_open
#------------------------------------------------------------------------------
sub fdump {
    # Function  input: hash with the following key/value pairs:
    #   'file'         => name of the FITS file to be opened (required!)
    #   'columns'      => either a comma-delimited string containing the
    #                       columns to extract from the FITS file or a reference
    #                       to an array of column names (required!)
    #   'array'        => if you only want the values of one column, specify a
    #                       number indicating which column of values you want
    #                       parsed and returned as an array;
    #                     if you specify a ref to an array of column numbers,
    #                       then it will return an array of refs to arrays
    #                       containing the values of the columns you specified
    #                     (note: column numbering starts at 1.)
    #                     (optional)
    #   'hash'         => if you would like the fdump output to be parsed
    #                       into a hash data structure, include this key and
    #                       supply as the value a ref to another hash with the
    #                       following key/value pairs:
    #                         'key'   => either a column number or a ref to
    #                                      an array of column numbers
    #                                      (for a multidimensional hash)
    #                         'value' => either a column number or a ref to
    #                                      an array of column numbers
    #                                      indicating which column(s) in
    #                                      the query results contains the
    #                                      values for the to-be-created hash
    #                         (note: column numbering starts at 1.)
    #                       (note: if multiple values per key are parsed,
    #                       either due to a recurring key or specifying
    #                       multiple columns as the 'value', then the
    #                       value will be a ref to an array.)
    #                     (optional)
    #   'row_hash'     => if you would like the fdump output to be parsed
    #                       into an array of hash data structures, one per row,
    #                       where the keys to each hash are the column names,
    #                       include this key and supply as the value either:
    #                          (a) the string 'all', which means that all
    #                              columns should be included in the hashes,
    #                       or
    #                          (b) a reference to an array of column numbers
    #                              which will specify which columns to include
    #                              in the hashes
    #                              (note: column numbering starts at 1.)
    #   'column_hash'  => if you would like the fdump output to be parsed into
    #                       a hash of array references, one array per column,
    #                       where the keys to the hash are the column names,
    #                       include this key and supply as the value either:
    #                          (a) the string 'all', which means that all
    #                              columns should be included in the hash,
    #                       or
    #                          (b) a reference to an array of column numbers
    #                              which will specify which columns to include
    #                              in the hash
    #                              (note: column numbering starts at 1.)
    #
    #   Note: If you don't specify either 'hash', 'array', 'column_hash' or
    #   'row_hash', then the (mostly) unparsed output of fdump will be returned.
    #
    # Function output: if the 'hash', 'column_hash', or 'row_hash' option is
    #   chosen, the output is a reference to a hash data structure; otherwise,
    #   it is an array

    my(%args) = @_;

    my($file) = $args{'file'};
    my($columns) = ((ref($args{'columns'}) eq 'ARRAY')
		    ? join(',',@{ $args{'columns'} }) : $args{'columns'});
    my($fdumpfh) = &Ftools::fdump_open(%args);
    defined($fdumpfh) || &$error_handler('Error: Could not execute fdump!');
    my(@output) = ();

    if (defined($args{'array'})) {
	my(@results) = ();
	my(@columns,$line);

	if (ref($args{'array'})) {
	    my(@cols) = @{ $args{'array'} };
	    my(@values);
	    print "Ftools::fdump: specified columns (",
	          join(',',@cols),")$eol" if $debug;
	    grep(($_ += $[) && 0,@cols) if $[; # See Camel, 1st Ed., p. 221.
	    while (<$fdumpfh>) {
		next if (/^\s*$/);
		&$error_handler("Error from fdump: $_") unless (/\|$/);
		$line = '|' . $_;
		$line =~ s/\s*\|\s*/|/g;
		@columns = split(/\|/,$line);
		@values = @cols;
		map($_ = $columns[$_],@values);
		push(@results,((scalar(@values) > 1)
			       ? [ @values ] : shift(@values)));
	    }
	    if ($debug) {
		print "Ftools::fdump: array results:$eol";
		foreach (@results) {
		    if (ref($_)) {
			print '[ ',join(', ',@{ $_ })," ]$eol";
		    } else {
			print "$_$eol";
		    }
		}
	    }
	    @output = @results;

	} else {

	    my($column) = $args{'array'} + $[;
	    print "Ftools::fdump: specified column $args{'array'}$eol"
	      if $debug;
	    while (<$fdumpfh>) {
		next if (/^\s*$/);
		&$error_handler("Error from fdump: $_") unless (/\|$/);
		$line = '|' . $_;
		$line =~ s/\s*\|\s*/|/g;
		@columns = split(/\|/,$line);
		push(@results,$columns[$column]) if defined($columns[$column]);
	    }
	    print "Ftools::fdump: array results: ",join(',',@results),$eol
	      if $debug;
	    @output = @results;
	}

    } elsif (defined($args{'hash'})) {

	my($hash) = $args{'hash'};
	my(@keycols) = (ref($hash->{'key'})
			? @{ $hash->{'key'} } : ($hash->{'key'}));
	my(@valcols) = (ref($hash->{'value'})
			? @{ $hash->{'value'} } : ($hash->{'value'}));
	my($results) = {};
	my(@columns) = ();
	my(@keys,@values,$line);

	print "Ftools::fdump: hash specified value column(s) (",
	      join(',',@valcols),")$eol",
	      "Ftools::fdump: hash specified key column(s) (",
	      join(',',@keycols),")$eol" if $debug;
	grep(($_ += $[) && 0,@keycols) if $[; # See Camel, 1st Ed., p. 221.
	grep(($_ += $[) && 0,@valcols) if $[;
	while (<$fdumpfh>) {
	    next if (/^\s*$/);
	    &$error_handler("Error from fdump: $_") unless (/\|$/);
	    $line = '|' . $_;
	    $line =~ s/\s*\|\s*/|/g;
	    @columns = split(/\|/,$line);
	    @keys = @keycols;
	    map($_ = $columns[$_],@keys);
	    @values = @valcols;
	    map($_ = $columns[$_],@values);
	    print "columns = (",join(',',@columns[1 .. $#columns]),")$eol",
	          "keys = (",join(',',@keys),")$eol",
	          "value(s) = (",join(',',@values),")$eol" if $debug;
	    my($hashref) = $results;
	    my($lastkey) = pop(@keys);
	    foreach (@keys) {
		if (!defined($hashref->{$_})) {
		    $hashref->{$_} = {};
		}
		$hashref = $hashref->{$_};
	    }
	    if (defined($hashref->{$lastkey})) {
		if (ref($hashref->{$lastkey}) eq 'ARRAY') {
		    push(@{ $hashref->{$lastkey} },@values);
		} else {
		    $hashref->{$lastkey} = [ $hashref->{$lastkey}, @values ];
		}
	    } else {
		$hashref->{$lastkey} = ((scalar(@values) > 1)
					? [ @values ] : shift(@values));
	    }
	}

	if ($debug) {
	    print "Ftools::fdump: dumping hash results...$eol";
	    &::DumpHash($results);
	}

	return($results);

    } elsif (defined($args{'column_hash'})) {

	my(@cols) = ((ref($args{'column_hash'}) eq 'ARRAY')
		     ? @{ $args{'column_hash'} } : ($args{'column_hash'}));
	my($results) = {};

	# Masssage column header...
	$columns =~ s/\s*,\s*/,/g;
	$columns = ',' . $columns;
	my(@keys) = split(/,/,$columns);
	print "Ftools::fdump: column_hash specified column(s) (",
	      join(',',@cols),")$eol" if $debug;

	if (($args{'column_hash'} eq 'all') || (scalar(@cols) == 0)) {
	    @cols = (1 .. scalar(@keys) - 1);
	}
	grep(($_ += $[) && 0,@cols) if $[; # See Camel, 1st Ed., p. 221.
	print "column headings = (",join(',',@keys[1 .. $#keys]),")$eol"
	  if $debug;

	# Initialize array references...
	my($col);
	foreach $col (@cols) {
	    $results->{$keys[$col]} = [];
	}

	# Iterate over each line of fdump output...
	my($line,@columns);
	while (<$fdumpfh>) {
	    next if (/^\s*$/);
	    &$error_handler("Error from fdump: $_") unless (/\|$/);
	    $line = '|' . $_;
	    $line =~ s/\s*\|\s*/|/g;
	    @columns = split(/\|/,$line);
	    print "columns = (",join(',',@columns[1 .. $#columns]),")$eol"
	      if $debug;
	    foreach $col (@cols) {
		push(@{ $results->{$keys[$col]} },$columns[$col]);
	    }
	}

	if ($debug) {
	    print "Ftools::fdump: dumping column_hash results...$eol";
	    &::DumpHash($results);
	}

	return($results);

    } elsif (defined($args{'row_hash'})) {

	my(@cols) = ((ref($args{'row_hash'}) eq 'ARRAY')
		     ? @{ $args{'row_hash'} } : ($args{'row_hash'}));
	my(@results) = ();

	# Masssage column header...
	$columns =~ s/\s*,\s*/,/g;
	$columns = ',' . $columns;
	my(@keys) = split(/,/,$columns);
	print "Ftools::fdump: row_hash specified column(s) (",
	      join(',',@cols),")$eol" if $debug;

	if (($args{'row_hash'} eq 'all') || (scalar(@cols) == 0)) {
	    @cols = (1 .. scalar(@keys) - 1);
	}
	grep(($_ += $[) && 0,@cols) if $[; # See Camel, 1st Ed., p. 221.
	print "column headings = (",join(',',@keys[1 .. $#keys]),")$eol"
	  if $debug;

	my($line,$col,@columns);
	while (<$fdumpfh>) {
	    next if (/^\s*$/);
	    &$error_handler("Error from fdump: $_") unless (/\|$/);
	    $line = '|' . $_;
	    $line =~ s/\s*\|\s*/|/g;
	    @columns = split(/\|/,$line);
	    print "columns = (",join(',',@columns[1 .. $#columns]),")$eol"
	      if $debug;
	    my($hash) = {};
	    foreach $col (@cols) {
		$hash->{$keys[$col]} = $columns[$col];
	    }
	    push(@results,$hash);
	}

	if ($debug) {
	    print "Ftools::fdump: dumping row_hash results...";
	    my($row);
	    for ($row = 0; $row <= $#results; $row++) {
		print $eol,'row ',$row + 1,':',$eol;
		&::DumpHash($results[$row]);
	    }
	}

	return(@results);
    } else {
	while (<$fdumpfh>) {
	    next if (/^\s*$/);
	    &$error_handler("Error from fdump: $_") unless (/\|$/);
	    push(@output,$_);
	}
	if ($debug) {
	    print "Ftools::fdump: dumping unparsed results...$eol";
	    foreach (@output) {
		s/\n/$eol/ if ($eol ne "\n");
		print;
	    }
	}
    }

    return(@output);

} #eosub--fdump
#------------------------------------------------------------------------------
sub nh {
    # Function  input: hash with the following key/value pair:
    #   'position'     => an Astro::Object object
    # Function output: an array of output from the `nh' ftool binary

    my(%args) = @_;

    ref($args{'position'})
      || &$error_handler('Error: Not a valid astronomical position.');
    my($object) = $args{'position'};

    my($radius) = '';
    if (defined($args{'radius'})) {
	&$error_handler('Error: Not a valid astronomical search radius.')
	  unless (($args{'radius'} > 0) && ($args{'radius'} < 180));
	$radius = 'disio=' . $args{'radius'};
    }

    my($coordsys) = $object->coordsys();
    &$error_handler('Error: The coordinate system must be equatorial.')
      unless ($coordsys->name() eq 'Equatorial');
    my($equinox) = $coordsys->equinox();
    ($equinox =~ /^([0-9\.]+)/) && ($equinox = $1);	# detaint
    my(@coordinates) = @{ $object->coordinates() };
    ($coordinates[0] =~ /^([-0-9\.eE]+)/) && ($coordinates[0] = $1); # detaint
    ($coordinates[1] =~ /^([-0-9\.eE]+)/) && ($coordinates[1] = $1); # detaint

    my($command) = "nh $equinox @coordinates $radius 2>&1 |";
    print "Ftools::nh: command = $command$eol" if $debug;
    my($nhfh) = new FileHandle $command;
    defined($nhfh) || &$error_handler('Error: Could not execute nh!');
    my(@output) = <$nhfh>;
    &$error_handler('Error: Could not execute nh!')
      unless (scalar(@output));

    return(@output);

} #eosub--nh
#------------------------------------------------------------------------------
sub nh_parse {
    # Function  input: hash with the following key/value pair:
    #   'position'     => an Astro::Object object
    # Function output: a reference to a hash of data parsed from the output
    #                  of the `nh' ftool binary

    my(@output) = &Ftools::nh(@_);
    my($hash) = ($debug ? { 'output' => [ @output ] } : {});
    my($mode) = 0;
    my($line);
    foreach $line (@output) {
	if (! $mode) {
	    ($line =~ /LII\s*,\s*BII\s*(\S+)\s+(\S+)/)
	      && do { $hash->{'galactic coordinates'} = [ $1, $2 ]; next; };
	    ($line =~ /Requested position at X and Y pixel\s+(\S+)\s+(\S+)/)
	      && do { $hash->{'pixel coordinates'} = [ $1, $2 ]; next; };
	    ($line =~ /Search Nh in\s+(\d+)\s+X\s+(\d)\s+box/)
	      && do { $hash->{'box size (pixels)'} = [ $1, $2 ]; next; };
	    ($line =~ /Each pixel is\s+(\S+)\s+deg\s+(\S+)\s+deg/)
	      && do { $hash->{'pixel resolution (deg)'} = [ $1, $2 ]; next; };
	    ($line =~ /\s+RA\s+DEC\s+Dist\s+Nh/i) && do { $mode = 1; next; };
	    ($line =~ /^\s*(Error.*)/i) && do { $hash->{'error'} = $1; next; };
	    ($line =~ /^\s*Weighted average nH \(cm\*\*-2\)\s+(\S+)/i)
	      && do { $hash->{'weighted average nH (cm**-2)'} = $1; next; };
	} else {
	    if ($line =~ /^\s*Average nH \(cm\*\*-2\)\s+(\S+)/i) {
		$hash->{'average nH (cm**-2)'} = $1;
		$mode = 0;
	    } else {
		$line =~ s/^\s+//;
		my(@values) = split(/\s+/,$line);
		my($row_hash) = { 'coordinates' => [ $values[0], $values[1] ],
				  'distance' => $values[2],
				  'nH' => $values[3] };
		if (exists($hash->{'box values'})) {
		    push(@{ $hash->{'box values'} },$row_hash);
		} else {
		    $hash->{'box values'} = [ $row_hash ];
		}
	    }
	}
    }

    return($hash);

} #eosub--nh_parse
#------------------------------------------------------------------------------
1;
